/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: a52.c 11 2004-07-15 13:46:20Z picard $
 *
 * BetaPlayer A52 plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "..\common\common.h"
#include "a52.h"
#include "liba52\a52.h"

#define A52_INPUT		0x100
#define A52_OUTPUT		0x101

typedef struct a52
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;
	pin Output;
	merge Merge;

	a52_state_t* State;
	sample_t* Samples;
	int FrameSize;
	int SubFrame;

	// output data
	packet Packet;
	tick_t RefTime;
	bool_t Pending;
	int Total;

} a52;

const datadef A52Params[] = 
{
	{ A52_ID,		TYPE_CLASS },
	{ A52_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ A52_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },

	DATADEF_END
};

static int SetFormat( a52* p, const audio* Format )
{
	if (p->State)
	{
	    a52_free(p->State);
		p->State = NULL;
	}

	memset(&p->InputFormat,0,sizeof(p->InputFormat));
	MergeDone(&p->Merge);

	if (Format)
	{
		if (Format->Format != AUDIOFMT_A52)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		p->OutputFormat.Flags = PCM_PLANES;
		p->OutputFormat.Format = AUDIOFMT_PCM;
		p->OutputFormat.Bits = sizeof(sample_t)*8;
		p->OutputFormat.FracBits = 30;
		p->OutputFormat.SampleRate = p->InputFormat.SampleRate;
		p->OutputFormat.Channels = 2;

		p->Pending = 0;
		p->RefTime = 0;
		p->Total = 0;
		p->SubFrame = 6;
		p->FrameSize = 0;
		MergeInit(&p->Merge);

		p->State = a52_init(0);
		p->Samples = a52_samples(p->State);
	}

	return ERR_NONE;
}

bool_t DecodeFrame( a52* p )
{
	for (;;)
	{
		if (p->SubFrame >= 6)
		{
			static const int Channels[8] = { 2, 1, 2, 3, 3, 4, 4, 5 };
			int Flags;
			int SampleRate;
			int BitRate;
			level_t Level;

			// find valid frame header

			while (!p->FrameSize)
			{
				if (p->Merge.WritePos-p->Merge.ReadPos < 8)
				{
					MergePack(&p->Merge,0);
					return 0;
				}

                p->FrameSize = a52_syncinfo(p->Merge.Data+p->Merge.ReadPos,&Flags,&SampleRate,&BitRate);
				
				if (p->FrameSize)
				{
					p->InputFormat.SampleRate = SampleRate;
                    p->InputFormat.Channels = Channels[Flags & 7];
                    if (Flags & A52_LFE)
						p->InputFormat.Channels++;
					p->InputFormat.ByteRate = BitRate/8;
					break;
				}

				p->Merge.ReadPos++;
			}

			if (p->Merge.WritePos-p->Merge.ReadPos < p->FrameSize)
			{
				MergePack(&p->Merge,0);
				return 0;
			}

            Flags = A52_STEREO;
            Level = 1<<26;
            if (!a52_frame(p->State, p->Merge.Data+p->Merge.ReadPos, &Flags, &Level, 384)) 
			{
				// valid frame
				p->SubFrame = 0;
				p->Merge.ReadPos += p->FrameSize;
			}
			p->FrameSize = 0;
		}
		else
		{
			if (a52_block(p->State))
				p->SubFrame = 6; // skip to next frame
			else
			{	
				p->Packet.RefTime = p->RefTime;
				p->Packet.Length = 256 * sizeof(sample_t);
				p->Packet.Data[0] = p->Samples;
				p->Packet.Data[1] = p->Samples+256;
				p->Total += 256;
				p->RefTime = -1;
				p->SubFrame++;
				return 1;
			}
		}
	}
	return 0;
}

static int Decode( a52* p, const packet* Packet )
{
	int Result;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;

	if (p->Pending)
	{
		do
		{
			if (p->Output.Node->Set(p->Output.Node,p->Output.No,
									&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
				return ERR_BUFFER_FULL;
		}
		while (DecodeFrame(p));

		p->Pending = 0;
	}

	// set new reftime
	if (Packet->RefTime >= 0)
		p->RefTime = Packet->RefTime;

	// add new packet to mergebuffer
	MergeWrite(&p->Merge,Packet->Data[0],Packet->Length);

	if (DecodeFrame(p))
	{
		Result = ERR_NONE;
		do
		{
			int i = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));
			if (i == ERR_BUFFER_FULL)
			{
				p->Pending = 1;
				//return last result
				break;
			}

			Result = i;
		}
		while (DecodeFrame(p));
	}
	else
		Result = ERR_NEED_MORE_DATA; // if no frame decoded: need more packet

	return Result;
}

static bool_t IsEmpty( a52* p )
{
	packet Zero;

	if (!p->InputFormat.Format) 
		return 1;

	Zero.CurrTime = -1;
	Zero.DropLevel = p->Packet.DropLevel;
	Zero.Length = 0;
	Zero.Data[0] = NULL;
	Zero.RefTime = -1;

	return Decode(p,&Zero) == ERR_NEED_MORE_DATA;
}

static int Enum( a52* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,A52Params,NULL);
}

static int Get( a52* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(A52_ID,int); break;
	case A52_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case A52_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case A52_OUTPUT: GETVALUE(p->Output,pin); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_EMPTY:GETVALUE(IsEmpty(p),bool_t); break;
	}
	return Result;
}

static int Set( a52* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case A52_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case A52_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case A52_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_EMPTY:
		MergeClear(&p->Merge);
		p->Pending = 0;
		p->FrameSize = 0;
		p->SubFrame = 6;
		break;
	}
	return Result;
}
static a52 A52;

void A52_Init()
{
	NodeFill(&A52.VMT,sizeof(a52),Enum,Get,Set);
	NodeRegister(&A52.VMT,PRI_DEFAULT);
}

void A52_Done()
{
	NodeUnRegister(&A52.VMT);
}



