/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: avi.c 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer AVI plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "avi.h"

#define INDEXSTEP	1024

#define AVI_INDEX_OF_INDEXES	0x00
#define AVI_INDEX_OF_CHUNKS		0x01
#define AVI_INDEX_KEY			0x80000000

typedef struct aviindex
{
	int32_t Id;
	int8_t Flags[4];
	int32_t Pos;
	int32_t Length;
} aviindex;

typedef struct avisuperindex
{
	int Offset;
	int Size;
	int StartTime;
	int EndTime;
	int LastData;

} avisuperindex;

typedef struct avistdindex
{
	int FourCC;
	int Size;
	int16_t LongsPerEntry;
	int8_t IndexSubType;
	int8_t IndexType;
	int Count;
	int ChunkId;
	int OffsetLow;
	int OffsetHigh;
	int Reserved;

} avistdindex;

typedef struct avistream
{
	format_stream Stream;
	
	int* Index;			// mediatime
	bool_t ByteRate;
	int RefStart;
	int MediaTime;
	int NewMediaTime;	// after seek (if seek successed)
	fraction64 MediaRate; // byterate or packetrate
	int MaxPacket;
	int SuperIndexCount;
	avisuperindex* SuperIndex; //opendml index

} avistream;

typedef struct avi
{
	format_base Format;

	int FramePeriod;
	int ByteRate;
	bool_t OutOfSync;
	bool_t ErrorIndex;
	bool_t OpenDML;
	bool_t CheckZero;

	int DataPos;
	int DataEnd;

	// stream header and format loading
	int HeadNo;
	bool_t HeadType;

	int VideoHandler;
	int VideoFrames;
	fraction VideoFPS;
	int VideoStart;

	fraction AudioRate;
	int AudioLength;
	int AudioStart;
	int AudioSampleSize;
	
/*	int AudioFormat;
	int AudioChannels;
	int AudioSampleRate;
	int AudioByteRate;
	int AudioBits;
*/
	int IndexAdjustPos;
	int IndexCount;
	int IndexPos;
	int IndexReaded;
	int IndexBufferNo;
	aviindex IndexBuffer[INDEXSTEP];

	char NameChar[256];

} avi;

#define ALIGN2(x) (((x)+1)&~1)
#define ISDIGIT(x) ((x)>='0' && (x)<='9')
#define STREAMNO(x) ((((x) & 255)-'0')*10 + ((((x) >> 8) & 255)-'0'))
#define ISDATA(x) (ISDIGIT((x) & 255) && ISDIGIT(((x) >> 8) & 255))
#define ISIDX(x) (ISDIGIT(((x) >> 16) & 255) && ISDIGIT(((x) >> 24) & 255) && ((x) & 255)=='i' && (((x) >> 8) & 255)=='x')
#define IDXNO(x) (((((x) >> 16) & 255)-'0')*10 + ((((x) >> 24) & 255)-'0'))

static int Init(avi* p)
{
	p->CheckZero = 0;
	p->OpenDML = 1;
	p->IndexAdjustPos = 4;
	p->IndexPos = -1;
	p->IndexCount = 0;
	p->IndexReaded = 0;
	p->IndexBufferNo = -1;
	p->DataPos = -1;
	p->DataEnd = -1;
	p->OutOfSync = 0;
	p->ErrorIndex = 0;

	memset(p->NameChar,0,sizeof(p->NameChar));
	memset(p->NameChar+'a',1,'z'-'a');
	memset(p->NameChar+'A',1,'Z'-'A');
	memset(p->NameChar+'0',1,'9'-'0');
	p->NameChar['_']=1;
	p->NameChar[' ']=1;

	return ERR_NONE;
}

static int Probe(avi* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

    if (Data[0]=='R' &&	Data[1]=='I' &&	Data[2]=='F' &&	Data[3]=='F' &&
		Data[8]=='A' &&	Data[9]=='V' &&	Data[10]=='I'&&	Data[11]==' ')
		return ERR_NONE;

	if (CheckExts(URL,p->Format.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static INLINE int ChunkTime( avistream* Stream, int DataLength )
{
	if (Stream->ByteRate)
		return DataLength; //byterate

	// packetrate
	if (DataLength > Stream->MaxPacket) 
		return (DataLength + Stream->MaxPacket - 1) / Stream->MaxPacket;
	return 1;
}

static void CheckZeroChunks();
static int ReadPacket(avi* p, format_packet* Packet)
{
	avistream* Stream;
	int No,Ch,StartPos;
	int SubType,Type,DataLength;
	tick_t Duration;
	int LongsPerEntry;
	int IndexType;
	int Count;
	int BaseOffset;

	// find a valid chunk header first

	//DEBUG_MSG1(DEBUG_FORMAT,T("AVI Packet %08x"),p->Format.FilePos);

	StartPos = p->Format.FilePos;

	if (!p->OpenDML && p->DataEnd >= 0 && p->Format.FilePos >= p->DataEnd)
		return ERR_END_OF_FILE;

	do
	{
		bool_t PossibleAlign = p->Format.FilePos & 1;
		Type = 0;
		No = 0;
		while (No<32)
		{
			Ch = Format_Read8(&p->Format);

			if (Ch>=0 && p->NameChar[Ch])
			{
				Type |= Ch << No;
				No += 8;

				if (No==32 && PossibleAlign)
				{
					int Stream = STREAMNO(Type);
					if (!ISDATA(Type) || ISDIGIT((Type>>16) & 255) || ISDIGIT((Type>>24) & 255) || 
						Stream <0 || Stream>=p->Format.StreamCount)
					{
						No -= 8;
						Type >>= 8;
					}
				}
			}
			else
			{
				if (!PossibleAlign)
				{
					p->OutOfSync = 1;
					if ((p->Format.FilePos-StartPos)>=CHUNKSIZE)
						return ERR_DATA_NOT_FOUND; // this will force the process loop to stop, so input thread can be woken up
					if (Format_Eof(&p->Format))
						return ERR_END_OF_FILE;
				}
				No = 0;
				Type = 0;
				PossibleAlign = 0;
			}
		}

		if (!p->OpenDML && p->DataEnd >= 0 && p->Format.FilePos >= p->DataEnd)
			return ERR_END_OF_FILE;

		DataLength = Format_ReadLE32(&p->Format);

		if (DataLength<0 || (p->Format.FilePos+DataLength<0) || (p->DataPos>=0 && p->Format.FileSize>=0 && p->Format.FilePos+DataLength>p->Format.FileSize))
		{
			DataLength = -1;
			p->OutOfSync = 1;
			continue;
		}

		if (ISDATA(Type))
		{
			No = STREAMNO(Type);
			if (No >= 0 && No < p->Format.StreamCount)
			{
				//DEBUG_MSG3(DEBUG_FORMAT,"AVI Packet stream:%d pos:%08x length:%d",No,p->Format.FilePos,DataLength);

				if (p->OutOfSync)
				{
					if (p->Format.ProcessTime >= 0)
					{
						int i;
						tick_t RefTime;
	
						// this won't solve the out of sync problem
						// but at least the streams won't be dropped

						for (i=0;i<p->Format.StreamCount;++i)
						{
							Stream = (avistream*) p->Format.Streams[i];
							RefTime = Stream->RefStart+Scale64(Stream->MediaTime,Stream->MediaRate.Num,Stream->MediaRate.Den);
							if (RefTime < p->Format.ProcessTime - TICKSPERSEC)
								Stream->MediaTime = Scale64(p->Format.ProcessTime-Stream->RefStart,Stream->MediaRate.Den,Stream->MediaRate.Num);
						}
					}
					p->OutOfSync = 0;
				}

				Stream = (avistream*) p->Format.Streams[No];
				Packet->Stream = &Stream->Stream;
				Packet->Data = Format_ReadAsRef(&p->Format,DataLength);
				Packet->RefTime = Stream->RefStart+Scale64(Stream->MediaTime,Stream->MediaRate.Num,Stream->MediaRate.Den);

				DEBUG_MSG5(DEBUG_FORMAT,T("AVI stream:%d time:%d mediatime:%d pos:%08x length:%08x"),No,Packet->RefTime,Stream->MediaTime,p->Format.FilePos,DataLength);

				Stream->MediaTime += ChunkTime(Stream,DataLength);
				return ERR_NONE;
			}
		}

		if (ISIDX(Type))
		{
			No = IDXNO(Type);
			if (No >= 0 && No < p->Format.StreamCount)
				break; // opendml index
		}

		switch (Type)
		{
		case FOURCCLE('R','I','F','F'):
			if (DataLength > 256 && (p->Format.FileSize < 0 || p->Format.FileSize > DataLength + 8))
				p->Format.FileSize = DataLength + 8;
			p->HeadNo = -1;
			Format_ReadLE32(&p->Format); // 'AVI ' or 'AVIX'
			return ERR_NONE;

		case FOURCCLE('a','v','i','h'):
			p->FramePeriod = Format_ReadLE32(&p->Format);
			p->ByteRate = Format_ReadLE32(&p->Format);
			DataLength -= 8;
			break;

		case FOURCCLE('i','n','d','x'):
			// opendml index

			LongsPerEntry = Format_ReadLE16(&p->Format);
			Format_Read8(&p->Format); //subtype
			IndexType = Format_Read8(&p->Format);
			Count = Format_ReadLE32(&p->Format);
			Type = Format_ReadLE32(&p->Format);
			BaseOffset = (int)Format_ReadLE64(&p->Format);
			Format_Skip(&p->Format,4);
			DataLength -= 6*4;

			if (ISDATA(Type))
			{
				No = STREAMNO(Type);
				if (No >= 0 && No < p->Format.StreamCount)
				{
					avisuperindex* s;

					Stream = (avistream*) p->Format.Streams[No];

					if (IndexType == AVI_INDEX_OF_CHUNKS && LongsPerEntry==2)
					{
						p->OpenDML = 1;
						p->Format.FileSize = -1;

						// create fake index of indexes

						Stream->SuperIndexCount = 1;
						Stream->SuperIndex = s = (avisuperindex*) Alloc(sizeof(avisuperindex));

						s->Offset = p->Format.FilePos - 6*4 - 8;
						s->Size = DataLength + 6*4 + 8;
						s->StartTime = 0;
						s->EndTime = 0;
						s->LastData = 0;

						for (No=0;No<Count;++No)
						{
							int Data = BaseOffset + Format_ReadLE32(&p->Format) - 8;
							int Size = Format_ReadLE32(&p->Format) & ~AVI_INDEX_KEY;

							if (s->LastData < Data)
								s->LastData = Data;

							s->EndTime += ChunkTime(Stream,Size);
						}
					}
					if (IndexType == AVI_INDEX_OF_INDEXES && LongsPerEntry==4)
					{
						int Time;

						p->OpenDML = 1;
						p->Format.FileSize = -1;

						// load index of indexes

						Stream->SuperIndexCount = Count;
						Stream->SuperIndex = s = (avisuperindex*) Alloc(sizeof(avisuperindex)*Count);

						Time = 0;
						for (No=0;No<Count;++No,++s)
						{
							s->Offset = (int)Format_ReadLE64(&p->Format);
							s->Size = Format_ReadLE32(&p->Format);
							s->StartTime = Time;
							Time += Format_ReadLE32(&p->Format);
  							s->EndTime = Time;
							s->LastData = -1;
							DataLength -= 16;
						}
					}
				}
			}

			break;

		case FOURCCLE('s','t','r','h'):

			p->HeadType = Format_ReadLE32(&p->Format);
			DataLength -= 4;

			if (p->HeadType == FOURCCLE('v','i','d','s'))
			{
				p->VideoHandler = Format_ReadLE32(&p->Format);

				Format_ReadLE32(&p->Format); // flags
				Format_ReadLE16(&p->Format); // priority
				Format_ReadLE16(&p->Format); // language
				Format_ReadLE32(&p->Format); 

				p->VideoFPS.Den = Format_ReadLE32(&p->Format); 
				p->VideoFPS.Num = Format_ReadLE32(&p->Format); 

				if (!p->VideoFPS.Den || !p->VideoFPS.Num)
				{
					p->VideoFPS.Den = 1;
					p->VideoFPS.Num = 25;
				}
				if (p->VideoFPS.Den==1 && p->VideoFPS.Num >= 12000) //not important
					p->VideoFPS.Den = 1000;

				p->VideoStart = Format_ReadLE32(&p->Format);
				p->VideoFrames = Format_ReadLE32(&p->Format);

				DataLength -= 32;
			}
			else
			if (p->HeadType == FOURCCLE('a','u','d','s'))
			{
				Format_ReadLE32(&p->Format); // format
				Format_ReadLE32(&p->Format); // flags
				Format_ReadLE16(&p->Format); // priority
				Format_ReadLE16(&p->Format); // language
				Format_ReadLE32(&p->Format); 
				p->AudioRate.Den = Format_ReadLE32(&p->Format); 
				p->AudioRate.Num = Format_ReadLE32(&p->Format); 
				p->AudioStart = Format_ReadLE32(&p->Format);
				p->AudioLength = Format_ReadLE32(&p->Format);

				Format_ReadLE32(&p->Format); // initialframes
				Format_ReadLE32(&p->Format); // suggested buffersize

				p->AudioSampleSize = Format_ReadLE32(&p->Format);

				DataLength -= 44;
			}
			break;

		case FOURCCLE('s','t','r','f'):
			if (p->HeadType == FOURCCLE('v','i','d','s'))
			{
				Stream = (avistream*) Format_AddStream(&p->Format,sizeof(avistream));
				if (Stream)
				{
					DataLength = Format_BitmapInfo(&p->Format,&Stream->Stream,DataLength);

					if (((video*)Stream->Stream.Format)->Pixel.FourCC == 4)
						((video*)Stream->Stream.Format)->Pixel.FourCC = UpperFourCC(p->VideoHandler);

					Stream->Index = NULL;
					Stream->SuperIndex = NULL;
					Stream->ByteRate = 0;
					Stream->Stream.Fragmented = 0;

					Stream->MediaRate.Den = p->VideoFPS.Num;
					Stream->MediaRate.Num = (int64_t)p->VideoFPS.Den * TICKSPERSEC;
					Stream->MaxPacket = MAX_INT;
					Stream->RefStart = Scale64(p->VideoStart,Stream->MediaRate.Num,Stream->MediaRate.Den);

					((video*)Stream->Stream.Format)->FrameRate = p->VideoFPS;

					if (p->VideoFrames)
					{
						Duration = Scale64(p->VideoFrames,Stream->MediaRate.Num,Stream->MediaRate.Den);
						if (p->Format.Duration < Duration)
							p->Format.Duration = Duration;
					}

					Format_PrepairStream(&p->Format,&Stream->Stream);
				}
				Format_Skip(&p->Format,DataLength);
				return ERR_NONE;
			}
			else
			if (p->HeadType == FOURCCLE('a','u','d','s'))
			{
				Stream = (avistream*) Format_AddStream(&p->Format,sizeof(avistream));
				if (Stream)
				{
					DataLength = Format_WaveFormat(&p->Format,&Stream->Stream,DataLength);

					Stream->Index = NULL;
					Stream->SuperIndex = NULL;
					Stream->ByteRate = p->AudioSampleSize != 0;
					Stream->Stream.Fragmented = Stream->ByteRate;
					
					if (!Stream->ByteRate)
					{
						// VBR (packetrate)
						Stream->MediaRate.Den = p->AudioRate.Num;
						Stream->MediaRate.Num = (int64_t)p->AudioRate.Den * TICKSPERSEC;
						Stream->MaxPacket = p->AudioRate.Den;
					}
					else
					{
						// CBR (byterate)
						
						// some digital cameras don't fill AudioSampleSize correctly only BlockAlign (with ADPCM)
						if (p->AudioSampleSize < ((audio*)Stream->Stream.Format)->BlockAlign)
							p->AudioSampleSize = ((audio*)Stream->Stream.Format)->BlockAlign;

						if (((audio*)Stream->Stream.Format)->Format == 0x55 /*MP3*/ && 
							p->AudioRate.Num == 4977)
						{
							Stream->MediaRate.Den = (p->AudioRate.Num*16-7)*p->AudioSampleSize;
							Stream->MediaRate.Num = (int64_t)p->AudioRate.Den*TICKSPERSEC*16;
						}
						else
						{
							Stream->MediaRate.Den = p->AudioRate.Num*p->AudioSampleSize;
							Stream->MediaRate.Num = (int64_t)p->AudioRate.Den*TICKSPERSEC;
						}
					}

					Stream->RefStart = Scale64(p->AudioStart,(int64_t)p->AudioRate.Den*TICKSPERSEC,p->AudioRate.Num);

					if (p->AudioLength)
					{
						Duration = Scale64(p->AudioLength,(int64_t)p->AudioRate.Den*TICKSPERSEC,p->AudioRate.Num);
						if (Duration>0 && p->Format.Duration < Duration)
							p->Format.Duration = Duration;
					}

					Format_PrepairStream(&p->Format,&Stream->Stream);
				}
				Format_Skip(&p->Format,DataLength);
				return ERR_NONE;
			}
			break;

		case FOURCCLE('L','I','S','T'):
			SubType = Format_ReadLE32(&p->Format);
			DataLength -= 4;

			switch (SubType)
			{
			case FOURCCLE('s','t','r','l'):
				p->HeadNo++;
				return ERR_NONE;

			case FOURCCLE('h','d','r','l'):
				return ERR_NONE;

			case FOURCCLE('m','o','v','i'):
				p->DataPos = p->Format.FilePos;
				p->DataEnd = DataLength + p->Format.FilePos;
				p->Format.HeaderLoaded = 1;

				for (No=0;No<p->Format.StreamCount;++No)
					((avistream*)p->Format.Streams[No])->MediaTime = 0;

				if (p->CheckZero)
					CheckZeroChunks();

				return ERR_NONE;

			case FOURCCLE('o','d','m','l'):
				p->CheckZero = 1;
				break;

			default:
				return ERR_NONE;
			}
			break;

		case FOURCCLE('s','t','r','n'):
		case FOURCCLE('J','U','N','K'):
			break;

		default: 
			if (p->OutOfSync || DataLength > 256*1024) // unknown
				DataLength = -1;
		}
	}
	while (DataLength<0);

	Format_Skip(&p->Format,DataLength);

	return ERR_NONE;
}

static int ErrorIndex(avi* p)
{
	if (!p->ErrorIndex)
	{
		p->ErrorIndex = 1;
		NodeError((node*)p,AVI_ID,AVI_NOINDEX);
	}
	return ERR_NOT_SUPPORTED;
}

static int LoadIndexBuffer(avi* p,int No)
{
	if (p->IndexBufferNo != No)
	{
		int Bytes = (p->IndexCount - No*INDEXSTEP)*sizeof(aviindex);
		if (Bytes > INDEXSTEP*sizeof(aviindex))
			Bytes = INDEXSTEP*sizeof(aviindex);

		if (p->Format.Input->Seek(p->Format.Input,p->IndexPos+No*INDEXSTEP*sizeof(aviindex),SEEK_SET)<0)
			return ErrorIndex(p);

		if (p->Format.Input->Read(p->Format.Input,p->IndexBuffer,Bytes) != Bytes)
			return ErrorIndex(p);

		p->IndexBufferNo = No;

		if (No == 0)
		{
			if (INT32LE(p->IndexBuffer[0].Pos) == p->DataPos)
				p->IndexAdjustPos = p->DataPos;
			else
				p->IndexAdjustPos = 4;
		}
	}
	return ERR_NONE;
}

static void CheckZeroChunks()
{
}

static int ReadIndex(avi* p,avistream* Stream,avisuperindex* Index,int MediaTime,int DstPos)
{
	bool_t NeedKey = Stream->Stream.FormatType == PACKET_VIDEO;
	avistdindex StdIndex;
	int No;
	int FoundPos = -1;
	int32_t* i; 

	if (DstPos>=0 && Index->LastData>=0 && DstPos > Index->LastData)
		return -1;

	if (MediaTime>=0 && Index->EndTime <= MediaTime)
		return -1;

	if (p->Format.Input->Seek(p->Format.Input,Index->Offset,SEEK_SET) < 0)
		return -1;

	if (p->Format.Input->Read(p->Format.Input,&StdIndex,sizeof(StdIndex)) != sizeof(StdIndex))
		return -1;

	if (StdIndex.IndexType != AVI_INDEX_OF_CHUNKS ||
		StdIndex.LongsPerEntry != 2)
		return -1;

	i = (int32_t*) Alloc(8*StdIndex.Count);
	if (!i)
		return -1;

	if (p->Format.Input->Read(p->Format.Input,i,8*StdIndex.Count) != 8*StdIndex.Count)
	{
		Free(i);
		return -1;
	}

	if (Index->LastData<0)
	{
		Index->LastData = 0;
		for (No=0;No<StdIndex.Count;++No)
		{
			int Data = StdIndex.OffsetLow + i[No*2] - 8;
			if (Index->LastData < Data)
				Index->LastData = Data;
		}
	}
	
	Stream->NewMediaTime = Index->StartTime;

	for (No=0;No<StdIndex.Count;++No)
	{
		int Data = StdIndex.OffsetLow + i[No*2] - 8;
		int Size = i[No*2+1];

		if (MediaTime>=0 && Stream->NewMediaTime >= MediaTime && (!NeedKey || !(Size & AVI_INDEX_KEY))) 
		{
			FoundPos = Data; // found sample after mediatime
			break;
		}

		if (DstPos>=0 && Data >= DstPos) 
		{
			FoundPos = Data; // found a sample after the destination position
			break; 
		}

		Stream->NewMediaTime += ChunkTime(Stream,Size & ~AVI_INDEX_KEY);
	}

	Free(i);
	return FoundPos;
}

static int Seek(avi* p, tick_t Time, int FilePos )
{
	avistream* Stream;
	avistream* s;
	int i,j,No;
	int Result;

	if (p->DataPos < 0)
	{
		// no avi header yet, only seek to the beginning supported

		if (Time>0 || FilePos>0)
			return ERR_NOT_SUPPORTED;

		p->OutOfSync = 0;
		return Format_Seek(&p->Format,0,SEEK_SET);
	}

	// seek to the beginning? (no index needed for that)

	Stream = (avistream*) Format_DefSyncStream(&p->Format);

	if (Time <= 0 && FilePos <= 0)
	{
		if (Format_Seek(&p->Format,p->DataPos,SEEK_SET) != ERR_NONE)
			return ERR_NOT_SUPPORTED;

		for (j=0;j<p->Format.StreamCount;++j)
		{
			s = (avistream*) p->Format.Streams[j];
			s->MediaTime = 0;
		}

		p->OutOfSync = 0;
		return ERR_NONE;
	}

	//if (Format_Seek(&p->Format,p->DataPos+Scale(Time,p->DataEnd-p->DataPos,p->Format.Duration),SEEK_SET) != ERR_NONE)
	//	return ERR_NOT_SUPPORTED;
	//return ERR_NONE;

	if (Stream && Stream->SuperIndex)
	{
		// using OpenDML index

		int No;
		int Pos = -1;
		int OldFilePos = p->Format.Input->Seek(p->Format.Input,0,SEEK_CUR);
		int MediaTime = Scale64(Time,Stream->MediaRate.Den,Stream->MediaRate.Num);

		// find position by MediaTime
		for (No=0;No<Stream->SuperIndexCount;++No)
		{
			Pos = ReadIndex(p,Stream,Stream->SuperIndex+No,MediaTime,-1);
			if (Pos >= 0)
				break;
		}

		if (Pos<0)
		{
			Stream->NewMediaTime = Stream->SuperIndex[Stream->SuperIndexCount-1].EndTime;
			Pos = Stream->SuperIndex[Stream->SuperIndexCount-1].LastData;

			if (Pos<0)
			{
				p->Format.Input->Seek(p->Format.Input,OldFilePos,SEEK_SET);
				return ErrorIndex(p);
			}
		}

		// calculate NewMediaTime for all other streams
		for (j=0;j<p->Format.StreamCount;++j)
		{
			s = (avistream*) p->Format.Streams[j];
			if (s != Stream)
			{
				// find position by Pos
				for (No=0;No<s->SuperIndexCount;++No)
					if (ReadIndex(p,s,s->SuperIndex+No,-1,Pos) >= 0)
						break;

				if (No==s->SuperIndexCount)
					s->NewMediaTime = s->SuperIndex[s->SuperIndexCount-1].EndTime;
			}
		}

		if (Format_Seek(&p->Format,Pos,SEEK_SET) == ERR_NONE)
		{
			// set new positions
			for (j=0;j<p->Format.StreamCount;++j)
			{
				s = (avistream*) p->Format.Streams[j];
				s->MediaTime = s->NewMediaTime;
			}

			p->OutOfSync = 0;
			return ERR_NONE;
		}

		// restore old position
		p->Format.Input->Seek(p->Format.Input,OldFilePos,SEEK_SET);
		return ERR_NOT_SUPPORTED;
	}

	// do we have the index block?

	if (p->IndexPos < 0)
	{
		int Pos = p->DataEnd;
		// try to find index block

		if (p->Format.Input->Seek(p->Format.Input,Pos,SEEK_SET) < 0)
			return ERR_NOT_SUPPORTED;

		for (;;)
		{
			int32_t Head[2];

			if (Pos & 1)
			{
				if (p->Format.Input->Read(p->Format.Input,Head,sizeof(Head)) != sizeof(Head))
					return ErrorIndex(p);
				Pos += sizeof(Head);

				if (Head[0] != FOURCC('i','d','x','1'))
				{
					memmove((char*)Head,(char*)Head+1,sizeof(Head)-1);
					if (p->Format.Input->Read(p->Format.Input,(char*)Head+sizeof(Head)-1,1) != 1)
						return ErrorIndex(p);
					++Pos;
				}
			}
			else
			{
				if (p->Format.Input->Read(p->Format.Input,Head,sizeof(Head)) != sizeof(Head))
					return ErrorIndex(p);
				Pos += sizeof(Head);
			}

			if (Head[0] == FOURCC('i','d','x','1'))
			{
				p->IndexCount = Head[1] / sizeof(aviindex);

				i = ((p->IndexCount + INDEXSTEP - 1) / INDEXSTEP) + 1;

				for (No=0;No<p->Format.StreamCount;++No)
				{
					s = (avistream*) p->Format.Streams[No];
					s->Index = (int*) ReAlloc(s->Index,sizeof(int)*i);
					if (!s->Index)
						return ERR_OUT_OF_MEMORY;

					s->Index[0] = 0;
				}

				p->IndexPos = Pos;
				p->IndexReaded = 0;
				break;
			}

			Pos += Head[1];
			if (p->Format.Input->Seek(p->Format.Input,Pos,SEEK_SET) != Pos)
				return ErrorIndex(p);
		}
	}

	if (Stream)
	{
		int Next;
		int MediaTime;
		int IndexLeft = p->IndexCount;
		int IndexCount;
		int Pos;
		int OldFilePos = p->Format.Input->Seek(p->Format.Input,0,SEEK_CUR);
		bool_t NeedKey = Stream->Stream.FormatType == PACKET_VIDEO;

		MediaTime = Scale64(Time,Stream->MediaRate.Den,Stream->MediaRate.Num);

		// nice to have: support seek by file position

		// find the index page

		for (No=0;IndexLeft>0;++No,IndexLeft-=INDEXSTEP)
		{
			IndexCount = IndexLeft;
			if (IndexCount > INDEXSTEP)
				IndexCount = INDEXSTEP;

			if (p->IndexReaded <= No)
			{
				// load page and calc sum mediatime

				Result = LoadIndexBuffer(p,No);
				if (Result != ERR_NONE)
				{
					p->Format.Input->Seek(p->Format.Input,OldFilePos,SEEK_SET);
					return Result;
				}

				for (j=0;j<p->Format.StreamCount;++j)
				{
					s = (avistream*) p->Format.Streams[j];
					s->Index[No+1] = s->Index[No];
				}

				for (i=0;i<IndexCount;++i)
				{
					j = STREAMNO(p->IndexBuffer[i].Id);
					if (j>=0 && j<p->Format.StreamCount)
					{
						s = (avistream*) p->Format.Streams[j];
						s->Index[No+1] += ChunkTime(s,INT32LE(p->IndexBuffer[i].Length));
					}
				}

				p->IndexReaded = No+1;
			}

			//DEBUG_MSG4(DEBUG_FORMAT,"AVI IndexBlock no:%d start:%d end:%d (target:%d)",No,Stream->Index[No],Stream->Index[No+1],MediaTime);

			if (Stream->Index[No] <= MediaTime && Stream->Index[No+1] > MediaTime)
			{
				Result = LoadIndexBuffer(p,No);
				if (Result != ERR_NONE)
					return Result;

				for (j=0;j<p->Format.StreamCount;++j)
				{
					s = (avistream*) p->Format.Streams[j];
					s->NewMediaTime = s->Index[No];
				}

				// find sub page position

				for (i=0;i<IndexCount;++i)
				{
					j = STREAMNO(p->IndexBuffer[i].Id);
					if (j>=0 && j<p->Format.StreamCount)
					{
						s = (avistream*) p->Format.Streams[j];

						Next = s->NewMediaTime + ChunkTime(s,p->IndexBuffer[i].Length);
	
						//DEBUG_MSG3(DEBUG_FORMAT,"AVI Index stream:%d pos:%d flag:%d",j,INT32LE(p->IndexBuffer[i].Pos),p->IndexBuffer[i].Flags[0]);

						if (s==Stream && MediaTime < Next)
						{
							Pos = INT32LE(p->IndexBuffer[i].Pos) - p->IndexAdjustPos;

							// only accept keyframes with valid position
							if (Pos >= 0 && Pos < p->DataEnd - p->DataPos &&
								(!NeedKey || (p->IndexBuffer[i].Flags[0] & 0x10)))
								break;
							MediaTime = Next;
						}

						s->NewMediaTime = Next;
					}
				}
				
				// found? 
				if (i<IndexCount)
					break;

				// continue on next page
				MediaTime = Stream->Index[No+1];
			}
		}

		if (IndexLeft <= 0)
		{
			// go to end
			for (j=0;j<p->Format.StreamCount;++j)
			{
				s = (avistream*) p->Format.Streams[j];
				s->NewMediaTime = s->Index[p->IndexReaded];
			}

			Pos = p->DataEnd - p->DataPos;
		}

		if (Format_Seek(&p->Format,p->DataPos+Pos,SEEK_SET) == ERR_NONE)
		{
			// set new positions
			for (j=0;j<p->Format.StreamCount;++j)
			{
				s = (avistream*) p->Format.Streams[j];
				s->MediaTime = s->NewMediaTime;
			}

			p->OutOfSync = 0;
			return ERR_NONE;
		}

		// restore old position
		p->Format.Input->Seek(p->Format.Input,OldFilePos,SEEK_SET);
	}
	return ERR_NOT_SUPPORTED;
}

void FreeStream(avi* p, avistream* Stream)
{
	Free(Stream->Index);
	Free(Stream->SuperIndex);
}

static avi AVI;

void AVI_Init()
{
	AVI.Format = FormatBase;
	AVI.Format.VMT.Probe = Probe;
	AVI.Format.Init = Init;
	AVI.Format.Seek = Seek;
	AVI.Format.ReadPacket = ReadPacket;
	AVI.Format.FreeStream = FreeStream;
	AVI.Format.Id = AVI_ID;
	AVI.Format.Exts = T("avi:V;divx:V");

	NodeRegister((node*)&AVI.Format.VMT,PRI_DEFAULT);
}

void AVI_Done()
{
	NodeUnRegister((node*)&AVI.Format.VMT);
}