/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: wav.c 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer AVI plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "wav.h"

typedef struct wav_t
{
	format_base Format;

	int FileSize;
	int DataPos;
	int DataEnd;
	int Samples;

} wav;

#define ALIGN2(x) (((x)+1)&~1)
#define ALIGN8(x) (((x)+7)&~7)

static int Init(wav* p)
{
	p->DataPos = -1;
	p->DataEnd = -1;
	p->FileSize = 0;
	p->Samples = -1;

	return ERR_NONE;
}

static int Probe(wav* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

    if (((Data[0]=='R' && Data[1]=='I' && Data[2]=='F' && Data[3]=='F') ||
		(Data[0]=='R' && Data[1]=='M' && Data[2]=='P' && Data[3]=='3') ||
		(Data[0]=='S' && Data[1]=='D' && Data[2]=='S' && Data[3]=='S')) &&
		Data[8]=='W' &&	Data[9]=='A' &&	Data[10]=='V'&&	Data[11]=='E')
		return ERR_NONE;

	if (CheckExts(URL,p->Format.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static int ReadPacket(wav* p, format_packet* Packet)
{
	format_stream* Stream;
	int Type,Length,DataLength;

	//DEBUG_MSG1("WAV ReadPacket %08x",p->Format.FilePos);

	if (p->DataPos < 0 || p->Format.FilePos < p->DataPos)
	{
		Type = Format_ReadLE32(&p->Format);
		DataLength = Format_ReadLE32(&p->Format);
		Length = ALIGN2(DataLength);

		switch (Type)
		{
		case FOURCCLE('R','M','P','3'):
		case FOURCCLE('R','I','F','F'):
		case FOURCCLE('S','D','S','S'):
			p->FileSize = Length + 8;
			Format_ReadLE32(&p->Format); // 'WAVE'
			return ERR_NONE;

		case FOURCCLE('f','m','t',' '):

			Stream = Format_AddStream(&p->Format,sizeof(format_stream));
			if (Stream)
			{
				Length = Format_WaveFormat(&p->Format,Stream,Length);
				Stream->Fragmented = 0;
				Format_PrepairStream(&p->Format,Stream);
			}

			Format_Skip(&p->Format,Length);
			return ERR_NONE;

		case FOURCCLE('f','a','c','t'):
			p->Samples = Format_ReadLE32(&p->Format);
			Length -= 4;
			break;

		case FOURCCLE('d','a','t','a'):
			p->DataPos = p->Format.FilePos;
			p->DataEnd = Length + p->Format.FilePos;
			p->Format.HeaderLoaded = 1;

			if (p->Format.StreamCount)
			{
				audio* Format;
				tick_t Duration;

				Stream = p->Format.Streams[0];
				Format = (audio*)Stream->Format;

				if (p->Samples >= 0)
					Duration = Scale(p->Samples,TICKSPERSEC,Format->SampleRate);
				else
				if (Format->ByteRate)
					Duration = Scale(DataLength,TICKSPERSEC,Format->ByteRate);
				else
					Duration = 0;

				if (Duration>0 && p->Format.Duration < Duration)
					p->Format.Duration = Duration;
			}
			return ERR_NONE;
		}

		if (p->Format.FilePos + Length < p->FileSize)
			Format_Skip(&p->Format,Length);
	}
	else
	{
		if (p->Format.FilePos >= p->DataEnd)
			return ERR_END_OF_FILE;

		if (p->Format.StreamCount)
		{
			int Length;
			Stream = p->Format.Streams[0];

			if (p->Format.FilePos == p->DataPos)
				Packet->RefTime = 0;
			else
			if (((audio*)Stream->Format)->ByteRate)
				Packet->RefTime = Scale(p->Format.FilePos - p->DataPos,TICKSPERSEC,((audio*)Stream->Format)->ByteRate);
			else
				Packet->RefTime = -1;

			Length = p->DataEnd - p->Format.FilePos;
			if (Length > CHUNKSIZE) Length = CHUNKSIZE;
			Packet->Data = Format_ReadAsRef(&p->Format,-Length);
			Packet->Stream = Stream;
		}
	}
	return ERR_NONE;
}

static int Seek(wav* p, tick_t Time, int FilePos)
{
	format_stream* Stream;
	audio* Audio;

	if (p->DataPos < 0 || !p->Format.StreamCount)
	{
		// no avi header yet, only seek to the beginning supported

		if (Time > 0 || FilePos > 0)
			return ERR_NOT_SUPPORTED;

		return Format_Seek(&p->Format,0,SEEK_SET);
	}

	Stream = p->Format.Streams[0];

	if (FilePos < 0)
	{
		if (Time > 0 && !((audio*)Stream->Format)->ByteRate)
			return ERR_NOT_SUPPORTED;

		FilePos = Scale(Time,((audio*)Stream->Format)->ByteRate,TICKSPERSEC);
	}

	Audio = (audio*)Stream->Format; 
	if (Audio->BlockAlign > 1)
		FilePos = (FilePos / Audio->BlockAlign) * Audio->BlockAlign;

	if (FilePos<0) FilePos=0;
	if (FilePos>p->DataEnd-p->DataPos) FilePos=p->DataEnd-p->DataPos;

	return Format_Seek(&p->Format,p->DataPos+FilePos,SEEK_SET);
}

static wav WAV;

void WAV_Init()
{
	WAV.Format = FormatBase;
	WAV.Format.VMT.Probe = Probe;
	WAV.Format.Init = Init;
	WAV.Format.Seek = Seek;
	WAV.Format.ReadPacket = ReadPacket;
	WAV.Format.Id = WAV_ID;
	WAV.Format.Exts = T("wav:A;rmp:A");

	NodeRegister((node*)&WAV.Format.VMT,PRI_DEFAULT);
}

void WAV_Done()
{
	NodeUnRegister((node*)&WAV.Format.VMT);
}