/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: format_base.h 135 2004-12-08 07:37:19Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#ifndef __FORMAT_BASE_H
#define __FORMAT_BASE_H

// base format implementation

#define MAXSTREAM	256
#define CHUNKSIZE	16384
#define SAFETAIL	256
#define MINBUFFER	(256*1024)
#define SHOWAHEAD	(TICKSPERSEC/25)
#define HEADERLOAD	(1024*1024)
#define HEADERLOAD2	(MINBUFFER/2)
#define MEMORYLIMIT	(512*1024)

#define VIDEO_BURST		1
#define DEFAULT_BURST	6

typedef struct format_buffer
{
	int Id;
	int Length;
	int RefCount;
	struct format_buffer* Next;
	uint8_t Buffer[CHUNKSIZE];

} format_buffer;

typedef struct format_ref
{
	format_buffer* Buffer;
	int Begin;
	int Length;
	struct format_ref* Next;

} format_ref;

struct format_stream;
typedef struct format_packet
{
	format_ref* Data;
	tick_t RefTime;
	struct format_stream* Stream;
	struct format_packet* Next;
	bool_t Key;
	//int FilePos;
} format_packet;

typedef struct format_stream
{
	int No;
	int Id;
	pin Pin;
	pin Comment;
	int FormatType;
	void* Format;
	int FormatLength;
	void* Extra;
	int ExtraLength;
	bool_t Fragmented; //don't bother to merge buffers
	bool_t ForceMerge; 
	bool_t DisableDrop; // no dropping allowed
	int PacketBurst;

	// buffer for merging fragmented packets 
	// to a single continous memory
	merge Merge;

	bool_t Pending; // 1:full packet 2:only first reference
	tick_t LastTime; // last known refrence time (-1 out of sync)
	packet Packet;  // pending packet

	// buffered packets waiting for processing
	format_packet* PacketFirst;
	format_packet* PacketLast;

} format_stream;


typedef int (*fmtfunc)(void* This);
typedef void (*fmtvoid)(void* This);
typedef int (*fmtseek)(void* This,tick_t,int);
typedef int (*fmtreadpacket)(void* This, format_packet* Packet);
typedef void (*fmtstream)(void* This, void* Stream);
typedef int (*fmtstreamprocess)(void* This, void* Stream);
typedef void (*fmtbackup)(void* This, int Begin);

typedef struct format_base
{
	format VMT;

	fmtfunc Init;
	fmtvoid Done;
	fmtseek Seek;
	fmtfunc FillQueue;
	fmtreadpacket ReadPacket;
	fmtstream FreeStream;
	fmtstream ReleaseStream;
	fmtvoid AfterSeek;
	fmtstreamprocess Process;
	fmtstream Sended;
	fmtbackup BackupPacketState;

	int Id;
	const tchar_t* Exts;
	bool_t TimeStamps; // CalcDuration will try to load last packets
	bool_t NeedDuration;
	tick_t Duration;
	tick_t StartTime;
	int FileSize;
	stream* Input;
	tick_t DropTolerance;
	tick_t SkipTolerance;
	tick_t AVOffset;
	pin UpdateStreams;

	int SeekByPacket_DataStart;
	int SeekByPacket_BlockAlign;

	int StreamCount;
	int ActiveStreams;
	format_stream* Streams[MAXSTREAM];

	bool_t SyncMode;
	format_stream* SyncStream;
	tick_t SyncTime;

	int BufferId;
	format_packet* FreePackets;
	format_buffer* FreeBuffers; 
	format_ref*	   FreeRefs;

	bool_t NoMoreInput;

	void* BufferLock; // buffer handling must be multithread safe
	int BufferUsed; // number of buffers allocated (referenced)
	int BufferSize;

	int BufferAvailable; // number of available buffers in BufferFirst..BufferLast chain
	format_buffer* BufferFirst; 
	format_buffer* BufferLast; 

	format_buffer* ReadBuffer; 
	int ReadPos; 
	int FilePos;

	tick_t ProcessTime;
	bool_t Bench;
	bool_t HeaderLoaded;
	bool_t InSeek;

	void* InputLock;

} format_base;

DLL extern format_base FormatBase;

DLL int Format_Send(format_base* p,format_stream* Stream);
DLL format_stream* Format_AddStream(format_base*, int Length ); 
DLL void Format_RemoveStream(format_base* ); 
DLL void Format_PrepairStream(format_base*, format_stream* );
DLL format_stream* Format_DefSyncStream(format_base*);
DLL int Format_SeekByPacket(format_base*,tick_t Time,int FilePos);

DLL void* Format_StreamFormat(format_stream*, int Type, int Length ); 
DLL void* Format_StreamExtra(format_stream*, int Length ); 
DLL int Format_WaveFormat(format_base*,format_stream*,int Length);
DLL int Format_BitmapInfo(format_base*,format_stream*,int Length);
DLL int Format_WaveFormatMem(format_stream*,void*,int Length);
DLL int Format_BitmapInfoMem(format_stream*,void*,int Length);
DLL void Format_DefaultVideo( video* );
DLL void Format_DefaultAudio( audio* );

DLL format_buffer* Format_BufferRemove( format_base* );
DLL void Format_BufferRelease( format_base*, format_buffer* );

DLL int Format_Seek(format_base*,int Pos,int Origin);
DLL void Format_AfterSeek(format_base* p);

DLL void Format_ReleaseBuffers(format_base* p);
DLL bool_t Format_Eof(format_base*);
DLL int Format_FileSeek(format_base*,int Pos,int Origin); // keeps stream buffers
DLL void Format_Skip(format_base*,int);
DLL int Format_Read(format_base*,void*,int);
DLL int Format_Read8(format_base*);
DLL int Format_ReadBE16(format_base*);
DLL int Format_ReadLE16(format_base*);
DLL int Format_ReadBE32(format_base*);
DLL int Format_ReadLE32(format_base*);
DLL int64_t Format_ReadBE64(format_base*);
DLL int64_t Format_ReadLE64(format_base*);

DLL format_ref* Format_ReadAsRef(format_base*, int Length);
DLL format_ref* Format_DupRef(format_base*, format_ref*, int Offset, int Length );
DLL void Format_ReleaseRef( format_base*, format_ref* );

#endif
