/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: str.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#define MAXSTRING	1024
#define MAXBUF		64

#define BUFSIZE		4096

typedef struct stringtable
{
	const tchar_t* Empty;

	int BufferLeft;
	tchar_t** BufferEnd;
	tchar_t* Buffer[MAXBUF];

	int TableCount;
	stringdef Table[MAXSTRING];

	int TableDefCount;
	stringdef TableDef[MAXSTRING];

} stringtable;

int StringLang = LANG_DEFAULT; // current language is filled by platform
int StringCodePage = 1; //oem
static stringtable Table;

void String_Init()
{
	stringtable* p = &Table;

	p->Empty = T("");
	p->BufferEnd = p->Buffer;
	p->BufferLeft = 0;
	p->TableCount = 0;
	p->TableDefCount = 0;
}

void String_Done()
{
	stringtable* p = &Table;
	tchar_t** i;

	for (i=p->Buffer;i!=p->BufferEnd;++i)
		Free(*i);
}

int StringFind( stringdef* Table, int Count, int Class, int Id, bool_t* Found )
{
	int Mid = 0;
	int Lower = 0;
	int Upper = Count-1;
	stringdef* p;

	while ( Upper >= Lower ) 
	{
		Mid = ( Upper + Lower ) >> 1;

		p = Table+Mid;

		if ( Class < p->Class || (Class == p->Class && Id < p->Id))  		
			Upper = Mid-1;	
		else if ( Class > p->Class || Id > p->Id)  		
			Lower = Mid+1;	
		else 
		{			        
			*Found = 1;
			return Mid;
		}
	}

	*Found = 0;
	if ( Upper == Mid - 1 )
		return Mid;		
	else                 
		return Lower;    
}

bool_t StringAdd( int Lang, int Class, int Id, const tchar_t* s )
{
	stringtable* p = &Table;
	tchar_t* d;
	int Pos;
	int Size;
	int* Count;
	stringdef* Table;
	bool_t Found;

	if (Lang == LANG_DEFAULT)
	{
		Count = &p->TableDefCount;
		Table = p->TableDef;
	}
	else
	if (Lang == StringLang)
	{
		Count = &p->TableCount;
		Table = p->Table;
	}
	else
		return 0;

	if (*Count == MAXSTRING)
		return 0;

	Size = tcslen(s)+1;
	if (Size > BUFSIZE)
		Size = BUFSIZE;

	if (Size > p->BufferLeft && p->BufferEnd != &p->Buffer[MAXBUF])
	{
		*p->BufferEnd = (tchar_t*) Alloc(BUFSIZE*sizeof(tchar_t));
		if (*p->BufferEnd)
		{
			p->BufferLeft = BUFSIZE;
			++p->BufferEnd;
		}
	}

	if (Size > p->BufferLeft)
		return 0;

	d = p->BufferEnd[-1]+(BUFSIZE-p->BufferLeft);
	TcsNCpy(d,s,Size);
	p->BufferLeft -= Size;

	Pos = StringFind(Table,*Count,Class,Id,&Found);
	Table += Pos;

	if (!Found)
	{
		memmove(Table+1,Table,sizeof(stringdef)*(*Count-Pos));
		++(*Count);

		Table->Class = Class;
		Table->Id = Id;
	}

	Table->String = d;
	return 1;
}

bool_t StringAddMore( int Lang, stringdef* Def )
{
	if (Lang != StringLang && Lang != LANG_DEFAULT)
		return 0;

	for (;Def->Class >= 0;++Def)
		StringAdd(Lang,Def->Class,Def->Id,Def->String);

	return 1;
}

bool_t StringAddPrint( int Lang,int Class,int No,const tchar_t* Msg, ... )
{
	tchar_t s[256];
	vstprintf( s, Msg, (VA_LIST)(&Msg+1) );
	return StringAdd(Lang,Class,No,s);
}

const stringdef* StringEnum( int Class, int No )
{
	stringtable* p = &Table;
	stringdef* i;
	int LastId = -1;
	for (i=p->TableDef;i!=p->TableDef+p->TableDefCount;++i)
	{
		if (i->Class==Class && i->Id!=LastId && No--==0)
			return i;
		LastId = i->Id;
	}
	return NULL;
}

const tchar_t* String( int Class, int Id )
{
	stringtable* p = &Table;
	bool_t Found;
	int Pos;

	//DEBUG_MSG2(-1,T("String %d %d"),Class,Id);

	Pos = StringFind(p->Table,p->TableCount,Class,Id,&Found);
	if (Found)
		return p->Table[Pos].String;

	Pos = StringFind(p->TableDef,p->TableDefCount,Class,Id,&Found);
	if (Found)
		return p->TableDef[Pos].String;

	return p->Empty;
}

const tchar_t* StringDef( int Class, int Id )
{
	stringtable* p = &Table;
	bool_t Found;
	int Pos;

	Pos = StringFind(p->TableDef,p->TableDefCount,Class,Id,&Found);
	if (Found)
		return p->TableDef[Pos].String;

	Pos = StringFind(p->Table,p->TableCount,Class,Id,&Found);
	if (Found)
		return p->Table[Pos].String;

	return p->Empty;
}

#ifdef UNICODE
void TcsToStr(char* Out,const tchar_t* In,int OutLen)
{
	if (OutLen>0)
	{
		OutLen = wcstombs(Out,In,--OutLen);
		if (OutLen<0) OutLen=0;
		Out[OutLen] = 0;
	}
}
void TcsNCpy(tchar_t* Out,const tchar_t* In,int OutLen)
{
	if (OutLen>0)
	{
		int n = wcslen(In);
		wcsncpy(Out,In,OutLen);
		if (n>=OutLen) 
			Out[OutLen-1] = 0;
	}
}
void WcsToTcs(tchar_t* Out,const uint16_t* In,int OutLen)
{
	TcsNCpy(Out,In,OutLen);
}
#else
void TcsToStr(char* Out,const tchar_t* In,int OutLen)
{
	TcsNCpy(Out,In,OutLen);
}
void TcsNCpy(tchar_t* Out,const tchar_t* In,int OutLen)
{
	if (OutLen>0)
	{
		int n = strlen(In);
		strncpy(Out,In,OutLen);
		if (n>=OutLen) 
			Out[OutLen-1] = 0;
	}
}
void WcsToTcs(tchar_t* Out,const uint16_t* In,int OutLen)
{
	if (OutLen>0)
	{
		OutLen = wcstombs(Out,In,--OutLen);
		if (OutLen<0) OutLen=0;
		Out[OutLen] = 0;
	}
}
#endif
