/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: node_win32.c 145 2004-12-12 18:18:28Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../stdafx.h"

#if defined(_WIN32)

#ifndef STRICT
#define STRICT
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "../zlib/zlib.h"

#if defined(_WIN32_WCE)
#define TWIN(a) L ## a
#else
#define TWIN(a) a
#endif

static tchar_t RegBase[64];
static const tchar_t* CmdLine = NULL; 
static const tchar_t* ProgramId = NULL;
static const tchar_t* Version = NULL;

#define NULL_ID			FOURCC('N','U','L','L')

#if defined(_WIN32_WCE)
#define ACCESS_LOAD		0
#define ACCESS_SAVE		0 
#else
#define ACCESS_LOAD		KEY_READ
#define ACCESS_SAVE		KEY_READ | KEY_WRITE | KEY_SET_VALUE 
#endif

bool_t NodeBase( node* p, tchar_t* Base )
{
	char Id[4];

	if (p->Get(p,NODE_ID,Id,sizeof(Id)) != ERR_NONE)
		return 0;

	stprintf(Base,T("%s\\%c%c%c%c"),RegBase,Id[0],Id[1],Id[2],Id[3]);
	return 1;
}

bool_t NodeLoadValue( int Id, void* Data, int Size, int* Type )
{
	bool_t Result = 0;
	HKEY Key;
	tchar_t Name[32];
	stprintf(Name,T("%d"),Id);

	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, RegBase, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		DWORD RegSize = Size;
		DWORD RegType;

		if (RegQueryValueEx(Key, Name, 0, &RegType, (PBYTE)Data, &RegSize) == ERROR_SUCCESS)
			Result = 1;

		if (Type)
			switch (RegType)
			{
			case REG_SZ: *Type = TYPE_STRING; break;
			case REG_DWORD: *Type = TYPE_INT; break;
			default: *Type = 0; break;
			}

		RegCloseKey( Key );
	}

	return Result;
}

void NodeSaveValue( int Id, const void* Data, int Size, int Type )
{
	DWORD Disp;
	HKEY Key;
	tchar_t Name[32];
	stprintf(Name,T("%d"),Id);

	if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, RegBase, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		if (Size)
		{
			DWORD RegType;
			switch (Type)
			{
			case TYPE_STRING: RegType = REG_SZ; break;
			case TYPE_INT: RegType = REG_DWORD; break;
			default: RegType = REG_BINARY; break;
			}

			RegSetValueEx( Key, Name, 0, RegType, (PBYTE)Data, Size );
		}
		else
			RegDeleteValue(Key,Name);
		RegCloseKey( Key );
	}
}

void NodeLoadSetup( node* p )
{
	HKEY Key;
	tchar_t Base[256];
	tchar_t Name[64];

	if (!p || !NodeBase(p,Base))
		return;
	
	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, Base, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		uint8_t Buffer[512];
		DWORD RegSize;
		DWORD RegType;
		node* Ptr;

		int No;
		datadef Type;

		for (No=0;p->Enum(p,No,&Type)==ERR_NONE;++No)
			if ((Type.Flags & DF_SETUP) && !(Type.Flags & (DF_OWNREG|DF_RDONLY)))
			{
				RegSize = sizeof(Buffer);
				stprintf(Name,T("%d"),Type.No);
				if (RegQueryValueEx(Key, Name, 0, &RegType, Buffer, &RegSize) == ERROR_SUCCESS)
				{
					switch (Type.Type)
					{
					case TYPE_BOOL:
					case TYPE_INT:
					case TYPE_HOTKEY:
					case TYPE_FRACTION:
					case TYPE_STRING:
					case TYPE_RECT:
					case TYPE_POINT:
					case TYPE_RGB:
					case TYPE_EQUALIZER:
					case TYPE_TICK:
						p->Set(p,Type.No,Buffer,RegSize);
						break;

					case TYPE_NODE:
						if (RegType == REG_DWORD)
						{
							if (*(int*)Buffer == NULL_ID)
							{
								Ptr = NULL;
								p->Set(p,Type.No,&Ptr,sizeof(Ptr));
							}
							else
							{
								Ptr = NodeFind(*(int*)Buffer);
								if (Ptr)
									p->Set(p,Type.No,&Ptr,sizeof(Ptr));
							}
						}
						break;
					}
				}
			}

		RegCloseKey( Key );
	}
}

void NodeSaveSetup( node* p )
{
	DWORD Disp;
	HKEY Key;
	tchar_t Base[256];
	tchar_t Name[64];

	if (!p || !NodeBase(p,Base))
		return;

	if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		uint8_t Buffer[512];
		DWORD RegSize;
		DWORD RegType;
		node* Ptr;
		int Size;
		int No;
		datadef Type;
		struct list 
		{
			tchar_t Name[64];
			struct list *Next;
		} *Chain = NULL,*List;

		RegSize = sizeof(Name)/sizeof(tchar_t);
		for (No=0;RegEnumValue(Key,No,Name,&RegSize,NULL,&RegType,NULL,NULL)==ERROR_SUCCESS;++No)
		{
			List = (struct list*)Alloc(sizeof(struct list));
			if (!List) break;
			tcscpy(List->Name,Name);
			List->Next = Chain;
			Chain = List;
			RegSize = sizeof(Name)/sizeof(tchar_t);;
		}

		while (Chain)
		{
			List = Chain;
			Chain = List->Next;
			RegDeleteValue(Key,List->Name);
			Free(List);
		}
		
		for (No=0;p->Enum(p,No,&Type)==ERR_NONE;++No)
			if ((Type.Flags & DF_SETUP) && !(Type.Flags & (DF_OWNREG|DF_RDONLY)))
			{
				RegType = RegSize = Size = 0;

				switch (Type.Type)
				{
				case TYPE_BOOL:
					RegType = REG_DWORD;
					RegSize = sizeof(DWORD);
					Size = sizeof(bool_t);
					break;
				case TYPE_TICK:
				case TYPE_INT:
				case TYPE_HOTKEY:
					RegType = REG_DWORD;
					RegSize = Size = sizeof(int);
					break;
				case TYPE_FRACTION:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(fraction);
					break;
				case TYPE_STRING:
					RegType = REG_SZ;
					RegSize = Size = sizeof(Buffer);
					break;
				case TYPE_RECT:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(rect);
					break;
				case TYPE_POINT:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(point);
					break;
				case TYPE_RGB:
					RegType = REG_DWORD;
					RegSize = sizeof(DWORD);
					Size = sizeof(rgb);
					break;
				case TYPE_EQUALIZER:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(equalizer);
					break;
				case TYPE_NODE:
					if (p->Get(p,Type.No,&Ptr,sizeof(Ptr)) == ERR_NONE)
					{
						RegType = REG_DWORD;
						RegSize = sizeof(DWORD);

						if (Ptr)
						{
							if (Ptr->Get(Ptr,NODE_ID,Buffer,sizeof(int)) != ERR_NONE)
								RegSize = 0;
						}
						else
							*(int*)Buffer = NULL_ID;
					}
					break;
				}

				if (Size)
				{
					int Result;
					memset(Buffer,0,RegSize);
					Result = p->Get(p,Type.No,Buffer,Size);
					if (Result != ERR_NONE && Result != ERR_NOT_SUPPORTED) // not supported settings should be still saved
						continue;
				}

				if (RegSize)
				{
					stprintf(Name,T("%d"),Type.No);
					RegSetValueEx( Key, Name, 0, RegType, Buffer, RegSize );
				}
			}

		RegCloseKey( Key );
	}
}

typedef struct lib
{
	HMODULE Dll;
	FARPROC Func;
	uint8_t* Min;
	uint8_t* Max;
} lib;

static int NodeLibCount = 0;
static lib NodeLib[256];

const tchar_t* NodeProgramId() { return ProgramId; }
const tchar_t* NodeVersion() { return Version; }
const tchar_t* NodeCmdLine() { return CmdLine; }

void Node_Init( const tchar_t* NewProgramId, const tchar_t* NewVersion, funcany Any, const tchar_t* NewCmdLine )
{
	lib* Lib;

	ProgramId = NewProgramId;
	Version = NewVersion;
	CmdLine = NewCmdLine ? NewCmdLine : T("");

	tcscpy(RegBase,T("SOFTWARE\\"));
	tcscat(RegBase,NewProgramId);

	String_Init();
	Platform_Init();
	Advanced_Init();
	Blit_Init();
	PCM_Init();

	Lib = &NodeLib[0];
	Lib->Dll = NULL;
	Lib->Func = (FARPROC)Any;
	Lib->Min = Lib->Max = NULL;

	NodeLibCount = 1;
}

static void FilterString(tchar_t* i)
{
	tchar_t* j=i;
	for (;*i;++i,++j)
	{
		if (i[0]=='\\' && i[1]=='n')
		{
			*j=10;
			++i;
		}
		else
			*j=*i;
	}
	*j=0;
}

void AsciiToTcs(tchar_t* Out,const char* In,int OutLen)
{
#ifdef UNICODE
	if (!MultiByteToWideChar(CP_ACP,0,In,-1,Out,OutLen))
		StrToTcs(Out,In,OutLen);
#else
	StrToTcs(Out,In,OutLen);
#endif
}

void TcsToAscii(char* Out,const tchar_t* In,int OutLen)
{
#ifdef UNICODE
	if (!WideCharToMultiByte(CP_ACP,0,In,-1,Out,OutLen,0,0))
		TcsToStr(Out,In,OutLen);
#else
	TcsToStr(Out,In,OutLen);
#endif
}

#ifndef CP_UTF8
#define CP_UTF8 65001
#endif

void UTF8ToTcs(tchar_t* Out,const char* In,int OutLen)
{
#ifdef UNICODE
	if (!MultiByteToWideChar(CP_UTF8,0,In,-1,Out,OutLen))
		StrToTcs(Out,In,OutLen);
#else
	WCHAR Temp[512];
	if (!MultiByteToWideChar(CP_UTF8,0,In,-1,Temp,512) ||
		!WideCharToMultiByte(CP_ACP,0,Temp,-1,Out,OutLen,0,0))
		StrToTcs(Out,In,OutLen);
#endif
}

void TcsToUTF8(char* Out,const tchar_t* In,int OutLen)
{
#ifdef UNICODE
	if (!WideCharToMultiByte(CP_UTF8,0,In,-1,Out,OutLen,0,0))
		TcsToStr(Out,In,OutLen);
#else
	WCHAR Temp[512];
	if (!MultiByteToWideChar(CP_ACP,0,In,-1,Temp,512) ||
		!WideCharToMultiByte(CP_UTF8,0,Temp,-1,Out,OutLen,0,0))
		StrToTcs(Out,In,OutLen);
#endif
}

void StrToTcs(tchar_t* Out,const char* In,int OutLen)
{
#ifdef UNICODE
	if (!MultiByteToWideChar(StringCodePage,MB_PRECOMPOSED,In,-1,Out,OutLen))
	{
		OutLen = mbstowcs(Out,In,--OutLen);
		if (OutLen<0) OutLen=0;
		Out[OutLen] = 0;
	}
#else
	int n = strlen(In);
	strncpy(Out,In,OutLen);
	if (n>=OutLen) 
		Out[OutLen-1] = 0;
#endif
}

static void StringLoad(const tchar_t* FileName, gzFile gf, int Size)
{
	int Lang;
	tchar_t s[512];
	tchar_t* p;
	tchar_t* q;
	uint16_t s16[512];
	char* s8 = (char*)s16;
	bool_t Unicode = 0;
	int No = 0;
	int CodePage = CP_OEMCP;
	bool_t CodePageFound = 0;

	// filter by file name if possible

	tcscpy(s,FileName);
	tcsupr(s);
	p = tcsstr(s,T("LANG_"));
	if (p)
	{
		q = tcschr(p,'.');
		if (q && tcsncmp(q,T(".TXT"),4)==0)
		{
			*q = 0;
			Lang = StringToFourCC(p+5,1);

			if (Lang != LANG_DEFAULT && Lang != StringLang && Lang != FOURCC('S','T','D','_'))
				return;
		}
	}

	Lang = 0;

	if (Size >= 2 && gzread(gf,s8,2) == 2)
	{
		Size -= 2;
		No = 2;
		Unicode = s8[0] == (char)-1 && s8[1] == (char)-2;
		if (Unicode)
		{
			CodePageFound = 1;
			No = 0;
		}
	}

	for (;;No=0)
	{
		if (Unicode)
		{
			if (Size<2)
				break;

			while (Size>=2 && gzread(gf,&s16[No],2)==2)
			{
				Size -= 2;
				if (s16[No] == 10 || s16[No] == 13)
					break;
				if (No<511) 
					++No;
			}
			s16[No] = 0;

			WcsToTcs(s,s16,sizeof(s)/sizeof(tchar_t));
		}
		else
		{
			if (Size<1)
				break;

			while (Size>=1 && gzread(gf,&s8[No],1)==1)
			{
				--Size;
				if (s8[No] == 10 || s8[No] == 13)
					break;
				if (No<511) 
					++No;
			}
			s8[No] = 0;

#ifdef UNICODE
			if (!MultiByteToWideChar(CodePage,MB_PRECOMPOSED,s8,-1,s,sizeof(s)/sizeof(tchar_t)))
			{
				mbstowcs(s,s8,511);
				s[511] = 0;
			}
#else
			StrToTcs(s,s8,sizeof(s)/sizeof(tchar_t));
#endif
		}
		if (gzeof(gf))
			break;

		p = s;
		while (*p==' ') ++p;
		if (*p==0)
			continue;

		if (*p==';')
		{
			if (stscanf(p,T(";CODEPAGE = %d"),&CodePage)==1)
			{
				CodePageFound = 1;
				//if (!IsValidCodePage(CodePage))
				//	break; // skip file
			}
			continue;
		}

		if (*p=='[')
		{
			q = tcschr(p,']');
			if (!q)
				break; // not a valid language file

			if (!CodePageFound)
				break; // not a valid language file

			*q = 0;

			Lang = StringToFourCC(p+1,1);

			if (Lang == FOURCC('D','E','F','A'))
				Lang = LANG_DEFAULT;

			if (Lang == StringLang && Lang != LANG_DEFAULT)
				StringCodePage = CodePage;
		}
		else
		{
			if (!Lang)
				break;

			if (Lang && Lang != LANG_DEFAULT && Lang != StringLang)
				continue;

			q = tcschr(p,'=');
			if (!q)
				break; // not a valid language file

			*q=0;
			++q;

			if (tcslen(p)>4)
				if (tcslen(p)<8)
					No = StringToFourCC(p+4,1);
				else
				{
					No = StringToInt(p+4,1);
					if (No >= 32768)
						No -= 65536;
				}
			else
				No = 0;

			FilterString(q);
			StringAdd(Lang,StringToFourCC(p,1),No,q);
		}
	}
}

static void StringLoadTXT(const tchar_t* FileName)
{
	gzFile gf = gzopen(FileName,"rb");
	if (gf)
	{
		StringLoad(FileName,gf,MAX_INT);
		gzclose(gf);
	}
}

typedef struct tarhead
{
	char Name[100];
	char mode[8];
	char uid[8];
	char gid[8];
	char size[12];
	char mtime[12];
	char chksum[8];
	char link;
	char linkname[100];
	char unused[512-100-8-8-8-12-12-8-1-100];

} tarhead;

static void StringLoadTGZ(const tchar_t* FileName)
{
	gzFile gf = gzopen(FileName,"rb");
	if (gf)
	{
		tchar_t Name[MAXPATH];
		tarhead Head;
		int Size,i;
		int Pos;

		while (!gzeof(gf) && gzread(gf,&Head,sizeof(Head))==sizeof(Head) && Head.size[0])
		{
			Pos = gztell(gf);
			// strtoul is not supported on wince2.11 and earlier
			Size = 0;
			for (i=0;i<12;++i)
				if (Head.size[i] >= '0' && Head.size[i] <= '7')
					Size = Size*8 + (Head.size[i] - '0');
				else
					break;

			StrToTcs(Name,Head.Name,MAXPATH);

			StringLoad(Name,gf,Size);

			gzseek(gf,Pos+((Size+511)&~511)-gztell(gf),SEEK_CUR);
		}
		gzclose(gf);
	}
}

static void PluginError(const tchar_t* FileName)
{
	tchar_t Name[MAXPATH];
	tchar_t Ext[MAXPATH];
	SplitURL(FileName,NULL,NULL,Name,Ext);
	tcsupr(Name);
	NodeError(NULL,ERR_ID,ERR_NOT_COMPATIBLE,Name);
}

static void PluginLoad(const tchar_t* FileName)
{
	HMODULE Dll;
#if !defined(_WIN32_WCE)
	int OldMode = SetErrorMode(SEM_NOOPENFILEERRORBOX| SEM_FAILCRITICALERRORS);
#endif
	Dll = LoadLibrary(FileName);

	DEBUG_MSG2(DEBUG_SYS,T("LOAD PLUGIN %s %d"),FileName,Dll);

#if !defined(_WIN32_WCE)
	SetErrorMode(OldMode);
#endif

	if (Dll)
	{
		int Result;
		int No;
		lib* Lib;

		for (No=0;No<NodeLibCount;++No)
			if (NodeLib[No].Dll == Dll)
			{
				FreeLibrary(Dll);
				return;
			}

		Lib = &NodeLib[NodeLibCount++];
		Lib->Min = Lib->Max = NULL;
		Lib->Dll = Dll;
		Lib->Func = GetProcAddress(Dll,TWIN("DLLRegister"));
		if (!Lib->Func)
			Lib->Func = GetProcAddress(Dll,TWIN("_DLLRegister@4"));
		if (Lib->Func)
		{
			Result = ((int(*)(int))Lib->Func)(NODE_VERSION);
			if (Result != ERR_NONE)
			{
				Lib->Func = NULL;
				if (Result == ERR_NOT_COMPATIBLE)
					PluginError(FileName);
			}
		}

	}
	else
	if (GetFileAttributes(FileName) == (DWORD)-1)
		PluginError(FileName);
}

static void FindFiles(const tchar_t* Path, const tchar_t* Mask, void(*Process)())
{
	WIN32_FIND_DATA FindData;
	tchar_t FindPath[MAXPATH];
	HANDLE Find;

	tcscpy(FindPath,Path);
	tcscat(FindPath,Mask);
	Find = FindFirstFile(FindPath,&FindData);

	if (Find != INVALID_HANDLE_VALUE)
	{
		do
		{
			tcscpy(FindPath,Path);
			tcscat(FindPath,FindData.cFileName);

			//DebugMessage(T("TIME:%d %s"),GetTime(),FindPath);

			DEBUG_MSG1(DEBUG_SYS,T("LOAD %s"),FindPath);
			Process(FindPath);
		}
		while (FindNextFile(Find,&FindData));

		FindClose(Find);
	}
}

static void PluginsFind(const tchar_t* Module, tchar_t* Path, const tchar_t* Skip )
{
	tchar_t *s;
	HMODULE Handle = NULL;

	if (Module)
		Handle = GetModuleHandle(Module);
	GetModuleFileName(Handle,Path,MAXPATH);
	s = tcsrchr(Path,'\\');
	if (s) *s = 0;
	tcscat(Path,T("\\"));

	if (Skip && tcsicmp(Path,Skip)==0)
		return;

	FindFiles(Path,T("*.tgz"),StringLoadTGZ);
	FindFiles(Path,T("*.txt"),StringLoadTXT);
	FindFiles(Path,T("*.plg"),PluginLoad);
}

void NodePlugins()
{
	tchar_t Path[MAXPATH];
	tchar_t Path2[MAXPATH];
	tchar_t Name[64];
	DWORD NameSize;
	DWORD RegSize;
	DWORD RegType;
	HKEY Key;
	int No;

	DMO_Init();

	PluginsFind(NULL,Path,NULL);
	PluginsFind(T("common.plg"),Path2,Path); // just in case somebody using the emulator without setting up pathes

	tcscpy(Path,T("SOFTWARE\\BetaPlugins"));

	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, Path, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		NameSize = sizeof(Name)/sizeof(tchar_t);
		RegSize = sizeof(Path2);
		for (No=0;RegEnumValue(Key,No,Name,&NameSize,NULL,&RegType,(LPBYTE)Path2,&RegSize)==ERROR_SUCCESS;++No)
		{
			if (RegType == REG_SZ)
				PluginLoad(Path2);

			NameSize = sizeof(Name)/sizeof(tchar_t);
			RegSize = sizeof(Path2);
		}
	}

	for (No=1;No<NodeLibCount;++No)
	{
		lib* Lib = &NodeLib[No];
		if (!Lib->Func)
		{
			FreeLibrary(Lib->Dll);
			Lib->Dll = NULL;
		}
	}
}

void Node_Done()
{
	int No;
	for (No=NodeLibCount-1;No>=1;--No)
	{
		lib* Lib = &NodeLib[No];
		if (Lib->Dll)
		{
			FARPROC UnRegister = GetProcAddress(Lib->Dll,TWIN("DLLUnRegister"));
			if (!UnRegister)
				UnRegister = GetProcAddress(Lib->Dll,TWIN("_DLLUnRegister@0"));

			if (UnRegister)
				((void(*)())UnRegister)();

			FreeLibrary(Lib->Dll);
		}
	}
	NodeLibCount = 0;

	DMO_Done();
	PCM_Done();
	Blit_Done();
	Advanced_Done();
	Platform_Done();
	String_Done();
}

bool_t NodeLocate( void* Ptr, tchar_t* OutName, int* OutBase )
{
	lib* Lib;
	int No;
	uint8_t* v = (uint8_t*)Ptr;

	for (No=0;No<NodeLibCount;++No)
	{
		Lib = &NodeLib[No];
		if (Lib->Dll)
		{
			if (!Lib->Min)
			{
				int PageSize;
				CodeFindPages(*(void**)&Lib->Func,&Lib->Min,&Lib->Max,&PageSize);
			}

			if (Lib->Min <= v && Lib->Max > v)
			{
				tchar_t Path[MAXPATH];
				GetModuleFileName(Lib->Dll,Path,MAXPATH);
				SplitURL(Path,NULL,NULL,OutName,OutName);
				*OutBase = v - Lib->Min;
				return 1;
			}
		}
	}
	return 0;
}

#endif
