/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: waveout_win32.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer Core
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../stdafx.h"

#if defined(_WIN32)

#ifndef STRICT
#define STRICT
#endif
#include <windows.h>

#define RAMPSIZE		15

#define BENCH_SIZE		32

struct waveout;

typedef struct buffer
{
	WAVEHDR Head;
	planes Planes;
	struct buffer* Next; //next free buffer
	tick_t RefTime;
	int Bytes;

	char Buffer[4];

} buffer;

typedef struct waveout
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;

	buffer* FreeFirst;
	buffer* FreeLast;

	tick_t BufferScaledTime;
	int BufferScaledAdjust;
	int BufferLength;

	int BytesPerSample;
	int TotalBytes; // source format
	int Bytes; // output format
	int FillPos;
	int Skip;
	buffer* FillFirst;
	buffer* FillLast;

	HWAVEOUT Handle;
	CRITICAL_SECTION Section;

	void* PCM;
	int PCMSpeed;
	int BufferLimit;
	tick_t Tick;
	int TimeRef;
	LONG Waiting;	// number of waiting buffers in WaveOut
	LONG Used;		// number of used buffers (waiting or in fill chain)

	int Play;
	fraction Speed;
	fraction SpeedTime;
	int AdjustedRate;

	bool_t Dither;
	bool_t Swap;
	int Stereo;
	int Quality;
	tick_t BufferTime;
	fraction Unit;

	bool_t MonoVol;
	bool_t Mute; // just for caching
	int Volume; // backup value when mute is turned on
	int VolumeRamp;

	WAVEFORMATEX Format;

	int BenchWait[BENCH_SIZE];
	int BenchSum[BENCH_SIZE];
	int BenchCurrSum;
	int BenchAvg;
	int BenchAdj;
	int BenchAvgLimit;
	int BenchSpeedAvg;
	int BenchWaitPos;
	int benchAvgRate;
	
} waveout;

static int Enum( waveout* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,TimerParams,FlowParams,OutputParams,AOutputParams,NULL);
}

static bool_t GetMute(waveout* p)
{
	DWORD Value = 0;
	waveOutGetVolume(NULL,&Value);
	if (p->MonoVol) Value &= 0xFFFF;
	p->Mute = Value==0;
	return p->Mute;
}

static int GetVolume(waveout* p)
{
	DWORD Value;
	if (waveOutGetVolume(NULL,&Value) == MMSYSERR_NOERROR)
	{
		if (p->MonoVol)
		{
			Value &= 0xFFFF;
			Value |= Value << 16;
		}
		p->Mute = Value==0;
		if (!p->Mute)
			p->Volume = ((LOWORD(Value)+HIWORD(Value)+600)*100) / (0xFFFF*2);
	}
	return p->Volume;
}

static int SetVolume(waveout* p)
{
	if (p->Mute)
		waveOutSetVolume(NULL,0);
	else
		waveOutSetVolume(NULL,0x10001 * ((0xFFFF * p->Volume) / 100));

	return ERR_NONE;
}

static int Get( waveout* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(WAVEOUT_ID,int); break;
	case FLOW_EMPTY: GETVALUE(p->Waiting<=0,bool_t); break;
	case AOUT_INPUT|PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case AOUT_OUTPUT|PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case AOUT_VOLUME: GETVALUE(GetVolume(p),int); break;
	case AOUT_MUTE: GETVALUE(GetMute(p),bool_t); break;
	case AOUT_STEREO: GETVALUE(p->Stereo,int); break; 
	case AOUT_SWAP: GETVALUE(p->Swap,bool_t); break;
	case AOUT_BUFFER: GETVALUE(p->BufferTime,tick_t); break;
	case AOUT_QUALITY: GETVALUE(p->Quality,int); break;
	case FLOW_TOTAL:GETVALUE(p->BytesPerSample ? (p->TotalBytes / p->BytesPerSample):0,int); break;
	case TIMER_PLAY: GETVALUE(p->Play,bool_t); break;
	case TIMER_SPEED: GETVALUE(p->Speed,fraction); break;
	case TIMER_UNIT: GETVALUE(p->Unit,fraction); break;
	case TIMER_TIME:
		if (Size == sizeof(tick_t))
		{
			tick_t t;
			if (p->Speed.Num==0)
				t = -1;
			else
			if (p->Play)
			{
				EnterCriticalSection(&p->Section);
				t = p->Tick + Scale(GetTime()-p->TimeRef,p->SpeedTime.Num,p->SpeedTime.Den);
				LeaveCriticalSection(&p->Section);
			}
			else
				t = p->Tick;

			*(tick_t*)Data = t;
			Result = ERR_NONE;
		}
		break;
	}
	return Result;
}

static void UpdateBenchAvg(waveout* p)
{
	int n;

	p->BenchAvg += 4;
	p->BenchAdj = (24*16) / p->BenchAvg;
	p->BenchSpeedAvg = SPEED_ONE/(p->BenchAvg*BENCH_SIZE);
	n = p->BytesPerSample * p->OutputFormat.SampleRate;
	if (n>0)
		p->BenchAvgLimit = (n * p->BenchAvg) / 160;
	else
		p->BenchAvgLimit = MAX_INT;

	p->BenchCurrSum += 2 * BENCH_SIZE;
	for (n=0;n<BENCH_SIZE;++n)
	{
		p->BenchSum[n] += 2 * BENCH_SIZE;
		p->BenchWait[n] += 2;
	}
}

static void ReleaseBuffer(waveout* p,buffer* Buffer,bool_t Update)
{
	EnterCriticalSection(&p->Section);

	if (Update)
	{
		tick_t Old;
		int Time = GetTime();

		Old = p->Tick + Scale(Time-p->TimeRef,p->SpeedTime.Num,p->SpeedTime.Den);

		if (Buffer->RefTime >= 0)
			p->Tick = Buffer->RefTime + p->BufferScaledTime;
		else
			p->Tick += p->BufferScaledTime;
		p->TimeRef = Time;

		// if difference is little then just adjust (because GetTime() has is more linear)
		if (abs(Old - p->Tick) < TICKSPERSEC/2)
			p->Tick = Old + ((p->Tick - Old) >> 2);
	}

	Buffer->Next = NULL;
	if (p->FreeLast)
		p->FreeLast->Next = Buffer;
	else
		p->FreeFirst = Buffer;
	p->FreeLast = Buffer;
	p->Used--;
	LeaveCriticalSection(&p->Section);
}

static void Reset(waveout* p)
{
	int n;

	// release buffers already sended to device
	if (p->Handle)
	{
		tick_t OldTick = p->Tick;
		waveOutReset(p->Handle);
		p->Tick = OldTick;
	}
	
	// release fill chain
	while (p->FillFirst)
	{
		buffer* Buffer = p->FillFirst;
		p->FillFirst = Buffer->Next;
		ReleaseBuffer(p,Buffer,0);
	}
	p->FillLast = NULL;
	p->FillPos = p->BufferLength;
	p->Skip = 0;
	p->Bytes = 0;

	p->BenchAvg = 16-4;
	UpdateBenchAvg(p);

	p->BenchWaitPos = 0;
	p->BenchCurrSum = p->BenchAvg * BENCH_SIZE;
	for (n=0;n<BENCH_SIZE;++n)
	{
		p->BenchSum[n] = p->BenchCurrSum;
		p->BenchWait[n] = p->BenchAvg;
	}

	PCMReset(p->PCM); // init dither and subsample position
}

static buffer* GetBuffer(waveout* p)
{
	buffer* Buffer;

	// try to find a free buffer

	EnterCriticalSection(&p->Section);
	Buffer = p->FreeFirst;
	if (Buffer)
	{
		p->FreeFirst = Buffer->Next;
		if (Buffer == p->FreeLast)
			p->FreeLast = NULL;
	}
	if (!Buffer)
	{
		// allocate new buffer
		Buffer = (buffer*) Alloc(sizeof(buffer) + p->BufferLength );
		if (Buffer)
		{
			Buffer->Planes[0] = Buffer->Buffer;
			Buffer->Next = NULL;

			memset(&Buffer->Head,0,sizeof(WAVEHDR));
			Buffer->Head.lpData = Buffer->Buffer;
			Buffer->Head.dwUser = (DWORD)Buffer;
			Buffer->Head.dwBufferLength = p->BufferLength;
			Buffer->Head.dwBytesRecorded = p->BufferLength;

			if (waveOutPrepareHeader(p->Handle, &Buffer->Head, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
			{
				Free(Buffer);
				Buffer = NULL;
			}
		}
	}

	if (Buffer)
	{
		p->Used++;
		Buffer->RefTime = -1;
		Buffer->Next =NULL;
	}

	LeaveCriticalSection(&p->Section);
	return Buffer;
}

static void Write(waveout* p, tick_t CurrTime)
{
	if (p->Play)
		while (p->FillFirst != p->FillLast)
		{
			buffer* Buffer = p->FillFirst;

			if (!p->Waiting && CurrTime >= 0 && Buffer->RefTime >= 0 && Buffer->RefTime > CurrTime + SHOWAHEAD)
				break;

			p->FillFirst = Buffer->Next;

			if (waveOutWrite( p->Handle, &Buffer->Head, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
				ReleaseBuffer(p,Buffer,0);
			else
				InterlockedIncrement(&p->Waiting);
		}
}

static int Send(waveout* p, const planes Planes, int Length, tick_t RefTime, tick_t CurrTime, int Speed )
{
	buffer* Buffer;
	int DstLength;
	int SrcLength;
	planes DstPlanes;
	planes SrcPlanes;

	p->TotalBytes += Length;

	SrcPlanes[0] = Planes[0];
	SrcPlanes[1] = Planes[1];

	if (p->Skip > 0)
	{
		SrcLength = p->Skip;
		if (SrcLength > Length)
			SrcLength = Length;

		SrcPlanes[0] = (uint8_t*)SrcPlanes[0] + SrcLength;
		SrcPlanes[1] = (uint8_t*)SrcPlanes[1] + SrcLength;
		Length -= SrcLength;
	}

	while (Length > 0)
	{
		if (p->FillPos >= p->BufferLength)
		{
			// allocate new buffer
			Buffer = GetBuffer(p);
			if (!Buffer)
				break;

			if (p->FillLast)
				p->FillLast->Next = Buffer;
			else
				p->FillFirst = Buffer;
			p->FillLast = Buffer;
			p->FillPos = 0;
			Buffer->Bytes = p->Bytes;
		}
		else
			Buffer = p->FillLast;

		if (RefTime >= 0)
		{
			Buffer->RefTime = RefTime - ((p->FillPos * p->BufferScaledAdjust) >> 12);
			RefTime = -1;
		}

		SrcLength = Length;
		DstLength = p->BufferLength - p->FillPos;
		DstPlanes[0] = Buffer->Buffer + p->FillPos;

		PCMConvert(p->PCM,DstPlanes,SrcPlanes,&DstLength,&SrcLength,Speed);

		if (p->VolumeRamp < (1<<RAMPSIZE))
		{
			uint8_t *u8,*u8e;
			int16_t *s16,*s16e;
			int Ramp = p->VolumeRamp;
			int Inc = (10*(1<<RAMPSIZE)) / (p->OutputFormat.Channels * p->OutputFormat.SampleRate);
			if (Inc<=0) Inc=1;

			switch (p->OutputFormat.Bits)
			{
			case 8:
				u8=DstPlanes[0];
				u8e=u8+DstLength;
				for (;u8!=u8e && Ramp<(1<<RAMPSIZE);++u8,Ramp+=Inc)
					*u8 = (uint8_t)((((*u8 - 0x80) * Ramp) >> RAMPSIZE) + 0x80);
				break;

			case 16:
				s16=DstPlanes[0];
				s16e=s16+(DstLength>>1);
				for (;s16!=s16e && Ramp<(1<<RAMPSIZE);++s16,Ramp+=Inc)
					*s16 = (int16_t)((*s16 * Ramp) >> RAMPSIZE);
				break;

			default:
				Ramp = RAMPSIZE;
				break;
			}

			p->VolumeRamp = Ramp;
		}

		p->Bytes += DstLength;
		p->FillPos += DstLength;

		SrcPlanes[0] = (uint8_t*)SrcPlanes[0] + SrcLength;
		SrcPlanes[1] = (uint8_t*)SrcPlanes[1] + SrcLength;
		Length -= SrcLength;

		if (!SrcLength)
			break;
	}

	if (Length && p->InputFormat.BlockAlign>0)
		p->Skip = p->InputFormat.BlockAlign - Length % p->InputFormat.BlockAlign;

	Write(p,CurrTime);
	return ERR_NONE;
}

static void CALLBACK WaveProc( HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, 
                               DWORD dwParam1, DWORD dwParam2 )
{
    if (uMsg == WOM_DONE)
    {
		buffer* Buffer = (buffer*)(((WAVEHDR*)dwParam1)->dwUser);
		waveout* p = (waveout*)dwInstance;
		ReleaseBuffer(p,Buffer,1);
		InterlockedDecrement(&p->Waiting);
		//DEBUG_MSG1(DEBUG_AUDIO,T("WAVEOUT Waiting %d"),p->Waiting);
    }
}

static void UpdatePCM(waveout* p,const audio* InputFormat)
{
	p->SpeedTime = p->Speed;
	p->SpeedTime.Num *= TICKSPERSEC;
	p->SpeedTime.Den *= 1000;
	p->BufferScaledTime = Scale(TICKSPERSEC,p->Speed.Num*p->Unit.Den,p->Speed.Den*p->Unit.Num);
	if (p->BufferLength)
		p->BufferScaledAdjust = (p->BufferScaledTime*4096) / p->BufferLength;
	else
		p->BufferScaledAdjust = 0;
	if (!p->Speed.Num)
		p->PCMSpeed = SPEED_ONE;
	else
		p->PCMSpeed = Scale(SPEED_ONE,p->Speed.Num,p->Speed.Den);
	p->AdjustedRate = Scale(p->OutputFormat.SampleRate,p->Speed.Den,p->Speed.Num);

	if (p->Swap)
		p->OutputFormat.Flags |= PCM_SWAPEDSTEREO;
	else
		p->OutputFormat.Flags &= ~PCM_SWAPEDSTEREO;

	PCMRelease(p->PCM);
	p->PCM = PCMCreate(&p->OutputFormat,InputFormat,p->Dither);
}

static int UpdateBufferTime(waveout* p)
{
	while (p->Unit.Num > p->Unit.Den * 32)
		p->Unit.Den *= 2;
	while (p->Unit.Num < p->Unit.Den * 15)
		p->Unit.Num *= 2;

	p->BufferLimit = Scale(p->BufferTime,p->Unit.Num,p->Unit.Den) / TICKSPERSEC;
	return ERR_NONE;
}

static int SetFormat(waveout* p, const audio* Format, bool_t KeepHandle )
{
	Reset(p);

	// free buffers
	while (p->FreeFirst)
	{
		buffer* Buffer = p->FreeFirst;
		p->FreeFirst = Buffer->Next;

		waveOutUnprepareHeader( p->Handle, &Buffer->Head, sizeof(WAVEHDR) );
		Free(Buffer);
	}
	p->FreeLast = NULL;

	if (p->Handle && !KeepHandle)
	{
		waveOutClose(p->Handle);
		p->Handle = NULL;
	}

	p->TotalBytes = 0;
	p->BytesPerSample = 1;

	memset(&p->InputFormat,0,sizeof(audio));

	if (Format)
	{
		if (!p->Handle || !KeepHandle)
		{
			MMRESULT MMResult;
			int Try;

			if (Format->Format != AUDIOFMT_PCM) 
				return ERR_INVALID_DATA;

			if (Format->Channels == 0)
				return ERR_NONE;

			p->OutputFormat = *Format;
			p->OutputFormat.Flags = 0;
			p->Dither = 0;

			if (p->Stereo>0)
			{
				p->OutputFormat.Channels = 1;
				p->OutputFormat.Flags |= (p->Stereo==1) ? PCM_ONLY_LEFT:PCM_ONLY_RIGHT;
			}

			switch (p->Quality)
			{
			case 0: // low quality for very poor devices
				p->OutputFormat.Bits = 8;
				p->OutputFormat.FracBits = 7;
				p->OutputFormat.Channels = 1;
				p->OutputFormat.SampleRate = 22050;
				break;

			case 1: // no dither and only standard samplerate
				p->OutputFormat.Bits = 16;
				p->OutputFormat.FracBits = 15;
				if (p->OutputFormat.SampleRate >= 44100)
					p->OutputFormat.SampleRate = 44100;
				else
					p->OutputFormat.SampleRate = 22050;
				break;

			default:
			case 2: // original samplerate 
				p->OutputFormat.Bits = 16;
				p->OutputFormat.FracBits = 15;
				p->Dither = 1;
				break;
			}

			Try = 0;
			do
			{
				if (p->OutputFormat.Bits <= 8)
					p->OutputFormat.Flags |= PCM_UNSIGNED;

				p->Format.wFormatTag = WAVE_FORMAT_PCM;
				p->Format.nChannels = (WORD)p->OutputFormat.Channels;
				p->Format.nSamplesPerSec = p->OutputFormat.SampleRate;
				p->Format.wBitsPerSample = (WORD)p->OutputFormat.Bits;
				p->Format.nBlockAlign = (WORD)((p->Format.nChannels * p->Format.wBitsPerSample) >> 3);
				p->Format.nAvgBytesPerSec = p->Format.nSamplesPerSec * p->Format.nBlockAlign;
				p->Format.cbSize = 0;

				MMResult = waveOutOpen( &p->Handle, WAVE_MAPPER, &p->Format,(DWORD)WaveProc,
										(DWORD)p,CALLBACK_FUNCTION);

				if (MMResult==WAVERR_BADFORMAT)
				{
					++Try;
					if (Try==1)
					{
						if (p->OutputFormat.SampleRate > 35000)
							p->OutputFormat.SampleRate = 44100;
						else
							++Try;
					}
					if (Try==2)
					{
						if (p->OutputFormat.SampleRate != 22050)
							p->OutputFormat.SampleRate = 22050;
						else
							++Try;
					}
					if (Try==3)
					{
						if (p->OutputFormat.Channels > 1)
							p->OutputFormat.Channels = 1;
						else
							++Try;
					}
					if (Try==4)
					{
						if (p->OutputFormat.Bits != 8)
						{
							p->OutputFormat.Bits = 8;
							p->OutputFormat.FracBits = 7;
						}
						else
							++Try;
					}
					if (Try==5)
						break;
				}
			}
			while (MMResult==WAVERR_BADFORMAT);
		}

		if (p->Handle)
		{
			UpdateBufferTime(p);

			p->TimeRef = GetTime();
			p->InputFormat = *Format;
			p->BufferLength = (Scale(p->Format.nAvgBytesPerSec,p->Unit.Den,p->Unit.Num)+3) & ~3;
			p->FillPos = p->BufferLength;
			p->VolumeRamp = 0;
			p->BytesPerSample = p->InputFormat.Bits >> 3;
			if (!(p->InputFormat.Flags & PCM_PLANES))
				p->BytesPerSample *= p->InputFormat.Channels;

			UpdatePCM(p,&p->InputFormat);
			Reset(p);
		}
		else
		{
			NodeError(&p->VMT,ERR_ID,ERR_DEVICE_ERROR);
			return ERR_DEVICE_ERROR;
		}
	}
		
	return ERR_NONE;
}

static int SetInput(waveout* p,const packet* Packet)
{
	if (!p->Handle)
	{
		// act as a null device
		if (Packet->CurrTime != -1 && (Packet->RefTime > Packet->CurrTime + SHOWAHEAD))
			return ERR_BUFFER_FULL;
		else
			return ERR_NONE;
	}

	if (p->Speed.Num==0) // benchmark mode (auto adjust speed)
	{
		int Pos = p->BenchWaitPos;
		int OldSum;
		int Speed;

		if (p->Play)
		{
			while (Packet->Length > p->BenchAvgLimit)
				UpdateBenchAvg(p);
	
			p->BenchCurrSum -= p->BenchWait[Pos];
			p->BenchWait[Pos] = p->Waiting;
			p->BenchCurrSum += p->BenchWait[Pos];

			OldSum = p->BenchSum[Pos];
			p->BenchSum[Pos] = p->BenchCurrSum;

			if (++Pos == BENCH_SIZE)
				Pos = 0;

			p->BenchWaitPos = Pos;

			if (p->BenchCurrSum < 2*BENCH_SIZE*p->BenchAvg)
				Speed = (p->BenchCurrSum+1) * p->BenchSpeedAvg;
			else
				Speed = 2*SPEED_ONE+(p->BenchCurrSum-2*BENCH_SIZE*p->BenchAvg+1) * 4*p->BenchSpeedAvg;

			Speed -= p->BenchAdj*(p->BenchCurrSum - OldSum);

			if (p->Waiting < 3)
				Speed -= SPEED_ONE/5;
		}
		else
			Speed = SPEED_ONE;

		DEBUG_MSG3(DEBUG_AUDIO,T("Audio speed:%d length:%d (wait:%d)"),Speed,Packet->Length,p->Waiting);

		return Send(p,Packet->Data,Packet->Length,Packet->RefTime,Packet->CurrTime,Speed);
	}

	if (Packet->DropLevel)
		return ERR_NONE;

	if (p->Used >= p->BufferLimit)
	{
		Write(p,Packet->CurrTime);
		return ERR_BUFFER_FULL;
	}

	DEBUG_MSG3(DEBUG_AUDIO,T("Waveout reftime:%d used:%d waiting:%d"),Packet->RefTime,p->Used,p->Waiting);
	return Send(p,Packet->Data,Packet->Length,Packet->RefTime,Packet->CurrTime,p->PCMSpeed);
}

static int Update(waveout* p)
{
	buffer *OldFirst;
	buffer *OldLast;
	audio OldFormat;
	buffer* OldBuffers;
	buffer* OldFill;
	int OldFillPos; 
	int OldUsed;
	buffer *Buffer,*Next;

	EnterCriticalSection(&p->Section);

	OldUsed = p->Used;
	OldFirst = p->FreeFirst;
	OldLast = p->FreeLast;
	OldFormat = p->OutputFormat;
	OldFormat.SampleRate = p->AdjustedRate;

	OldFill = p->FillFirst;
	OldFillPos = p->FillPos;

	p->FillFirst = NULL;
	p->FillLast = NULL;
	p->FillPos = p->BufferLength;
	p->FreeFirst = NULL;
	p->FreeLast = NULL;

	LeaveCriticalSection(&p->Section);

	Reset(p);

	OldBuffers = p->FreeFirst;
	if (p->FreeLast)
		p->FreeLast->Next = OldFill;
	else
		OldBuffers = OldFill;

	p->Used = OldUsed;
	p->FreeFirst = OldFirst;
	p->FreeLast = OldLast;

	// setup temporary format
	UpdatePCM(p,&OldFormat);

	for (Buffer=OldBuffers;Buffer;Buffer=Next)
	{
		Next = Buffer->Next;
		Send(p,Buffer->Planes,Next ? p->BufferLength:OldFillPos,Buffer->RefTime,-1,p->PCMSpeed);
		ReleaseBuffer(p,Buffer,0);
	}

	UpdatePCM(p,&p->InputFormat);
	return ERR_NONE;
}

static int UpdateFormat(waveout* p,bool_t KeepFormat)
{
	audio SaveFormat = p->InputFormat;
	return SetFormat(p,&SaveFormat,KeepFormat);
}

static int UpdatePlay(waveout* p)
{
	if (p->Play)
	{
		p->TimeRef = GetTime();
		if (p->Handle)
		{
			if (!p->Waiting)
				Write(p,p->Tick);
			waveOutRestart(p->Handle);
		}
	}
	else
	{
		if (!p->Waiting)
			p->Tick += Scale(GetTime()-p->TimeRef,p->SpeedTime.Num,p->SpeedTime.Den);
		else
		if (p->Handle)
			waveOutPause(p->Handle);
	}
	return ERR_NONE;
}

static int Set(waveout* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case AOUT_INPUT|PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data,0);
		}
		break;
	case AOUT_INPUT:
		if (Size == sizeof(packet))
			Result = SetInput(p,(const packet*)Data);
		break;

	case AOUT_VOLUME: SETVALUECMP(p->Volume,int,SetVolume(p),EqInt); break;
	case AOUT_MUTE: SETVALUECMP(p->Mute,bool_t,SetVolume(p),EqBool); break;
	case AOUT_SWAP: SETVALUECMP(p->Swap,bool_t,Update(p),EqBool); break;
	case AOUT_STEREO: SETVALUECMP(p->Stereo,int,UpdateFormat(p,0),EqInt); break; 
	case AOUT_QUALITY: SETVALUECMP(p->Quality,int,UpdateFormat(p,0),EqInt); break; 
	case AOUT_BUFFER: SETVALUE(p->BufferTime,tick_t,UpdateBufferTime(p)); break;

	case FLOW_TOTAL:
		if (Size == sizeof(int))
		{
			p->TotalBytes = *(const int*)Data * p->BytesPerSample;
			Result = ERR_NONE;
		}
		break;

	case FLOW_EMPTY:
		Reset(p);
		p->VolumeRamp = 0;
		Result = ERR_NONE;
		break;

	case TIMER_PLAY: SETVALUECMP(p->Play,bool_t,UpdatePlay(p),EqBool); break;
	case TIMER_UNIT: SETVALUECMP(p->Unit,fraction,UpdateFormat(p,1),EqFrac); break;

	case TIMER_TIME:
		if (Size == sizeof(tick_t))
		{
			EnterCriticalSection(&p->Section);
			p->Tick = *(tick_t*)Data;
			p->TimeRef = GetTime();
			LeaveCriticalSection(&p->Section);
			Result = ERR_NONE;
		}
		break;

	case TIMER_SPEED:
		if (Size == sizeof(fraction))
		{
			Result = ERR_NONE;
			if (!EqFrac( &p->Speed, (const fraction*)Data ))
			{
				p->Speed = *(const fraction*)Data;
				Result = Update(p);
			}
		}
		break;
	}
	return Result;
}

static waveout WaveOut;

void WaveOut_Init()
{
	WAVEOUTCAPS Caps;
	waveout* p = &WaveOut;

	NodeFill(&p->VMT,sizeof(waveout),Enum,Get,Set);
	p->Volume = 100;
	p->Unit.Num = 32;
	p->Unit.Den = 1;
	p->Volume = 70;
	p->Quality = 2;
	p->BufferTime = TICKSPERSEC*5;
	p->Speed.Den = p->Speed.Num = 1;
	p->SpeedTime.Num = TICKSPERSEC;
	p->SpeedTime.Den = 1000;
	p->PCM = NULL;
	waveOutGetDevCaps(WAVE_MAPPER,&Caps,sizeof(Caps));
	p->MonoVol = (Caps.dwSupport & WAVECAPS_LRVOLUME) == 0;
	InitializeCriticalSection(&p->Section);
	if (waveOutGetNumDevs())
		NodeRegister(&p->VMT,PRI_DEFAULT);
	GetVolume(p);
}

void WaveOut_Done()
{
	waveout* p = &WaveOut;
	PCMRelease(p->PCM);
	DeleteCriticalSection(&p->Section);
	NodeUnRegister(&p->VMT);
}

#endif
