/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: matroska.c 11 2004-07-15 13:46:20Z picard $
 *
 * BetaPlayer Matroska plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "matroska.h"
#include "MatroskaParser.h"

typedef struct matroska
{
	format_base Format;
	FileCache IO;
	MatroskaFile* File;

} matroska;

#define MK(p) ((matroska*)((char*)(p)-OFS(matroska,IO)))

static int Probe(matroska* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

	if (Data[0] == 0x1A && Data[1] == 0x45 &&
		Data[2] == 0xDF && Data[3] == 0xA3)
	{
		// ugly, but we don't want to parse EBML here...
		const char* i = (const char*) Ptr;
		for (;Length >= 8;++i,--Length)
			if (memcmp(i,"matroska",8)==0)
				return ERR_NONE;
	}
  
	if (CheckExts(URL,p->Format.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static void AudioTrackMS(format_stream* s,TrackInfo* Info)
{
	s->Fragmented = 1;
	Format_WaveFormatMem(s,Info->CodecPrivate,Info->CodecPrivateSize);
}

static void AudioTrack(format_stream* s,TrackInfo* Info,int Format)
{
	audio* Audio = Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
	if (Audio)
	{
		s->Fragmented = 1;
		Audio->Format = Format;
		Audio->Channels = Info->Audio.Channels;
		Audio->SampleRate = (int)Info->Audio.SamplingFreq;
		Audio->Bits = Info->Audio.BitDepth;

		if (Format_StreamExtra(s,Info->CodecPrivateSize))
			memcpy(s->Extra,Info->CodecPrivate,Info->CodecPrivateSize);

		Format_DefaultAudio(Audio);
  	}
}

static void VideoTrackMS(format_stream* s,TrackInfo* Info)
{
	Format_BitmapInfoMem(s,Info->CodecPrivate,Info->CodecPrivateSize);

	if (Info->DefaultDuration)
	{
		longlong v = Info->DefaultDuration;
		video* Video = (video*)s->Format;
		Video->FrameRate.Num = 1000000000;
		while (v > MAX_INT)
		{
			v >>= 1;
			Video->FrameRate.Num >>= 1;
		}
		Video->FrameRate.Den = (int)v;
  	}
}

static void VideoTrack(format_stream* s,TrackInfo* Info,int FourCC)
{
	video* Video = Format_StreamFormat(s,PACKET_VIDEO,sizeof(video));
	if (Video)
	{
		Video->Pixel.Flags = PF_FOURCC;
		Video->Pixel.FourCC = FourCC;
		Video->Width = Info->Video.PixelWidth;
		Video->Height = Info->Video.PixelHeight;
		Video->Aspect = ASPECT_ONE;
		if (Info->DefaultDuration)
		{
			longlong v = Info->DefaultDuration;
			Video->FrameRate.Num = 1000000000;
			while (v > MAX_INT)
			{
				v >>= 1;
				Video->FrameRate.Num >>= 1;
			}
			Video->FrameRate.Den = (int)v;
		}
		if (Format_StreamExtra(s,Info->CodecPrivateSize))
			memcpy(s->Extra,Info->CodecPrivate,Info->CodecPrivateSize);
  	}
}

static void AddChapter(pin* Comment,ulonglong Start,struct ChapterDisplay* p,int No)
{
	tchar_t Base[32];
	tchar_t s[256];
	int Hour,Min,Sec;

	stprintf(Base,T("CHAPTER%02d"),No);
	tcscpy(s,Base);
	tcscat(s,T("NAME="));
	UTF8ToTcs(s+tcslen(s),p->String,255-tcslen(s));
	Comment->Node->Set(Comment->Node,Comment->No,s,sizeof(s));

	Start /= 1000000000;
	Hour = (int)(Start / 3600);
	Start -= Hour * 3600;
	Min = (int)(Start / 60);
	Start -= Min * 60;
	Sec = (int)Start;

	stprintf(s,T("%s=%02d:%02d:%02d"),Base,Hour,Min,Sec);
	Comment->Node->Set(Comment->Node,Comment->No,s,sizeof(s));
}

static int AddChapters(pin* Comment,Chapter *Ch,int No) 
{
	uint32_t i,j;

	if (Ch)
		for (i=0;i<Ch->nChildren;++i) 
		{
			for (j=0;j<Ch->Children[i].nDisplay;++i)
				if (Ch->Children[i].Display[j].String)
				{
					AddChapter(Comment,Ch->Children[i].Start,
						&Ch->Children[i].Display[j],No++);
					break;
				}

			No = AddChapters(Comment,&Ch->Children[i],No);
		}

	return No;
}

static void AddComment(pin* Comment,tchar_t* Name,char* Value) 
{
	if (Comment->Node && Value)
	{
		tchar_t s[256];
		tcscpy(s,Name);
		tcscat(s,T("="));
		UTF8ToTcs(s+tcslen(s),Value,255-tcslen(s));
		Comment->Node->Set(Comment->Node,Comment->No,s,sizeof(s));
	}
}

static int Init(matroska* p)
{
	unsigned n,Count;
	SegmentInfo* Info;

	p->File = mkv_Open(&p->IO,NULL,0);
	if (!p->File)
		return ERR_INVALID_DATA;

	Info = mkv_GetFileInfo(p->File);
	if (Info)
		p->Format.Duration = (tick_t)((Info->Duration /** (longlong)Info->TimecodeScale*/ * TICKSPERSEC) / 1E9);

	Count = mkv_GetNumTracks(p->File);
	for (n=0;n<Count;++n)
	{
		format_stream* s = Format_AddStream(&p->Format,sizeof(format_stream));
		if (s)
		{
			TrackInfo *Info = mkv_GetTrackInfo(p->File,n);
			if (Info)
			{
				if (strcmp(Info->CodecID, "V_MS/VFW/FOURCC")==0)
					VideoTrackMS(s,Info);
				else if (strncmp(Info->CodecID, "V_MPEG4/ISO", 11)==0)
					VideoTrack(s,Info,FOURCC('M','P','4','V'));
				else if (strcmp(Info->CodecID, "V_MPEG4/MS/V3")==0)
					VideoTrack(s,Info,FOURCC('M','P','4','3'));
				else if (strcmp(Info->CodecID, "V_MPEG1")==0 || strcmp(Info->CodecID, "V_MPEG2")==0)
					VideoTrack(s,Info,FOURCC('M','P','E','G'));
				else if (strcmp(Info->CodecID, "A_MPEG/L3")==0)
					AudioTrack(s,Info,0x0055);
				else if (strcmp(Info->CodecID, "A_MPEG/L2")==0 || strcmp(Info->CodecID, "A_MPEG/L1")==0)
					AudioTrack(s,Info,0x0050);
				else if (strcmp(Info->CodecID, "A_PCM/INT/LIT")==0)
					AudioTrack(s,Info,0x0001);
				else if (strcmp(Info->CodecID, "A_AC3")==0 || strcmp(Info->CodecID, "A_DTS")==0)
					AudioTrack(s,Info,0x2000);
				else if (strcmp(Info->CodecID, "A_VORBIS")==0)
				{
					AudioTrack(s,Info,0x2674F);
					s->Fragmented = 0;
				}
				else if (strcmp(Info->CodecID, "A_MS/ACM")==0)
					AudioTrackMS(s,Info);
			}
			Format_PrepairStream(&p->Format,s);

			AddComment(&s->Comment,T("TITLE"),Info->Name);
			AddComment(&s->Comment,T("LANGUAGE"),Info->Language);
		}
	}

	if (p->Format.StreamCount)
	{
		pin* Comment = &p->Format.Streams[0]->Comment;
		Tag* Tags;
		Chapter* Ch;

		mkv_GetTags(p->File,&Tags,&Count);

		if (Tags && Comment->Node)
			for (n=0;n<Count;++n,++Tags)
			{
				uint32_t i;
				for (i=0;i<Tags->nSimpleTags;++i)
				{
					tchar_t	s[256];
					UTF8ToTcs(s,Tags->SimpleTags[i].Name,64);
					tcscat(s,T("="));
					UTF8ToTcs(s+tcslen(s),Tags->SimpleTags[i].Value,255-tcslen(s));
					Comment->Node->Set(Comment->Node,Comment->No,s,sizeof(s));
				}
			}

		mkv_GetChapters(p->File,&Ch,&Count);
		if (Count>0 && Info)
			AddChapters(Comment,Ch,1);
	}

	mkv_SetTrackMask(p->File,0);
	
	p->Format.HeaderLoaded = 1;
	return ERR_NONE;
}

static void Done(matroska* p)
{
	if (p->File)
	{
		mkv_Close(p->File);
		p->File = NULL;
	}
}

static int ReadPacket(matroska* p, format_packet* Packet)
{
	ulonglong Start,End;
	format_ref* Ref;
	ulonglong Pos;
	unsigned int Flags,Track,Size;
	int Result = mkv_ReadFrame(p->File,0,&Track,&Start,&End,&Pos,&Size,(void**)&Ref,&Flags);

	if (Result == EOF)
		return ERR_END_OF_FILE;

	if (Result != 0)
		return ERR_INVALID_DATA;

	Packet->Data = Ref;
	Packet->Key = (Flags & FRAME_NOTKF) == 0;
	Packet->Stream = p->Format.Streams[Track];
	Packet->RefTime = (tick_t)((Start * TICKSPERSEC) / 1000000000);

	return ERR_NONE;
}

static int Seek(matroska* p, tick_t Time, int FilePos )
{
	if (Time < 0 && FilePos >= 0 && p->Format.Duration > 0 && p->Format.FileSize > 0)
		Time = Scale( FilePos, p->Format.Duration, p->Format.FileSize );

	if (Time >= 0)
	{
		mkv_Seek(p->File,((ulonglong)Time * 1000000000) / TICKSPERSEC,0);
		return ERR_NONE;
	}
	return ERR_NOT_SUPPORTED;
}

static void IOClose(struct FileCache *IO)
{
}

static int IOReadCh(struct FileCache *IO)
{
	return Format_Read8(&MK(IO)->Format);
}

static int IORead(struct FileCache *IO,void *buffer,int count)
{
	return Format_Read(&MK(IO)->Format,buffer,count);
}
static int IOSeek(struct FileCache *IO,longlong where,int how)
{
	int pos = (int)where;

	assert(how != 1 || pos >= 0);
	if (how == 1 && pos >= 0 && pos < MK(IO)->Format.BufferSize*CHUNKSIZE)
	{
		Format_Skip(&MK(IO)->Format,pos);
	}
	else
	if (how != 0 || MK(IO)->Format.FilePos != pos)
	{
	    if (Format_Seek(&MK(IO)->Format,pos,how) != ERR_NONE)
			return -1;
	}

	return 0;
}
static longlong IOTell(struct FileCache *IO)
{
	return MK(IO)->Format.FilePos;
}
static void* IOMakeRef(struct FileCache *IO,int count)
{
	return Format_ReadAsRef(&MK(IO)->Format,count);
}
static void	IOReleaseRef(struct FileCache *IO,void* ref)
{
	Format_ReleaseRef(&MK(IO)->Format,(format_ref*)ref);
}

static matroska Matroska;

void Matroska_Init()
{
	matroska* p = &Matroska;

	p->Format = FormatBase;
	p->Format.VMT.Probe = Probe;
	p->Format.Init = Init;
	p->Format.Done = Done;
	p->Format.Seek = Seek;
	p->Format.ReadPacket = ReadPacket;
	p->Format.Id = MATROSKA_ID;
	p->Format.Exts = T("mkv:V;mka:A");

	p->IO.close = IOClose;
	p->IO.makeref = IOMakeRef;
	p->IO.releaseref = IOReleaseRef;
	p->IO.ioread = IORead;
	p->IO.ioreadch = IOReadCh;
	p->IO.ioseek = IOSeek;
	p->IO.iotell = IOTell;

	NodeRegister((node*)&p->Format.VMT,PRI_DEFAULT);
}

void Matroska_Done()
{
	NodeUnRegister((node*)&Matroska.Format.VMT);
}