/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: avi.c 117 2004-11-28 02:45:11Z picard $
 *
 * BetaPlayer MPEG4AAC plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "mpeg4aac.h"
#include "faad2/include/neaacdec.h"

#define FAAD_INPUT			0x100
#define FAAD_OUTPUT			0x101

#define MAX_CHANNELS		6
#define BUFFER_NEEDED		(FAAD_MIN_STREAMSIZE * MAX_CHANNELS)
#define HEADER_NEEDED		65536

typedef struct faad
{
	node VMT;

	int Id;
	audio InputFormat;
	audio OutputFormat;
	pin Output;
	merge Merge;

	// output data
	packet Packet;
	tick_t RefTime;
	bool_t Pending;
	bool_t Inited;
	bool_t Eof;
	bool_t PacketBased;
	int Total;

	NeAACDecHandle Decoder;

} faad;

static int SetFormat( faad* p, const audio* Format )
{
	if (p->Decoder)
	{
	    NeAACDecClose(p->Decoder);
		p->Decoder = NULL;
	}

	MergeDone(&p->Merge);
	memset(&p->InputFormat,0,sizeof(p->InputFormat));

	if (Format)
	{
	    NeAACDecConfigurationPtr Config;

		if (Format->Format != AUDIOFMT_AAC && Format->Format != AUDIOFMT_AAC_3VIX)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		p->OutputFormat.Bits = 16;
		p->OutputFormat.Flags = 0;
		p->OutputFormat.Format = AUDIOFMT_PCM;
		p->OutputFormat.FracBits = 15;
		p->OutputFormat.SampleRate = p->InputFormat.SampleRate;	// optional
		p->OutputFormat.Channels = p->InputFormat.Channels; // optional

		p->PacketBased = (Format->Flags & PCM_PACKET_BASED) != 0;
		p->Decoder = NeAACDecOpen();
		p->Pending = 0;
		p->RefTime = 0;
		p->Total = 0;
		p->Inited = 0;
		p->Eof = 0;

		Config = NeAACDecGetCurrentConfiguration(p->Decoder);
		Config->defObjectType = LC;
		Config->outputFormat = FAAD_FMT_16BIT;
		Config->downMatrix = 1;
		Config->dontUpSampleImplicitSBR = 1;
		NeAACDecSetConfiguration(p->Decoder, Config);

		if (Format->ExtraLength)
		{
			unsigned long SampleRate;
			uint8_t Channels;
	
		    if (NeAACDecInit2(p->Decoder, Format->Extra, Format->ExtraLength, &SampleRate, &Channels) < 0)
			{
				NeAACDecClose(p->Decoder);
				p->Decoder = NULL;
				return ERR_INVALID_DATA;
			}

			p->InputFormat.SampleRate = p->OutputFormat.SampleRate = SampleRate;
			p->InputFormat.Channels = p->OutputFormat.Channels = Channels;
			p->Inited = 1;
		}

		MergeInit(&p->Merge);
	}

	return ERR_NONE;
}

bool_t DecodeFrame( faad* p, bool_t First )
{
	while ((p->PacketBased && First) || p->Eof || p->Merge.WritePos - p->Merge.ReadPos >= BUFFER_NEEDED)
	{
	    NeAACDecFrameInfo FrameInfo;
	    void *Buffer;

        Buffer = NeAACDecDecode(p->Decoder, &FrameInfo,
            p->Merge.Data+p->Merge.ReadPos,p->Merge.WritePos-p->Merge.ReadPos);

        p->Merge.ReadPos += FrameInfo.bytesconsumed;

        if (!FrameInfo.error && Buffer && FrameInfo.samples>0)
		{
			p->Packet.RefTime = p->RefTime;
			p->Packet.Length = FrameInfo.samples * sizeof(int16_t);
			p->Packet.Data[0] = Buffer;
			p->Total += 1024;
			p->RefTime = -1;
			return 1;
		}

		if (!FrameInfo.bytesconsumed)
			p->Merge.ReadPos++;

		First = 0;
	}

	MergePack(&p->Merge,0);
	return 0;
}

static int Decode( faad* p, const packet* Packet )
{
	int Result;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;

	if (p->Pending)
	{
		do
		{
			if (p->Output.Node->Set(p->Output.Node,p->Output.No,
									&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
				return ERR_BUFFER_FULL;
		}
		while (DecodeFrame(p,0));

		p->Pending = 0;
	}

	// set new reftime
	if (Packet->RefTime >= 0)
		p->RefTime = Packet->RefTime;
	p->Eof = Packet->Length < 0;

	// add new packet to mergebuffer
	MergeWrite(&p->Merge,Packet->Data[0],Packet->Length);

	if (!p->Inited)
	{
		unsigned long SampleRate;
		uint8_t Channels;
		int Readed;
		uint8_t* Buffer = p->Merge.Data + p->Merge.ReadPos;
		int Length = p->Merge.WritePos - p->Merge.ReadPos;

		if (!p->Eof && Length < HEADER_NEEDED)
			return ERR_NEED_MORE_DATA;

		if (Buffer[0] == 0xFF && (Buffer[1] & 0xF6)==0xF0)
		{
			static int SampleRates[] = {96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000,7350,0,0,0};
			int Samples = 0;
			int Bytes = 0;

	        SampleRate = SampleRates[(Buffer[2] & 0x3C) >> 2];

			while (Length > 7)
			{
				int n;

				if (Buffer[0] != 0xFF || (Buffer[1] & 0xF6)!=0xF0)
					break;

				n = (((int)Buffer[3] & 3) << 11) | (((int)Buffer[4]) << 3) | (Buffer[5] >> 5);

				Bytes += n;
				Samples += 1024;

				Length -= n;
				Buffer += n;
			}

			p->InputFormat.ByteRate = Scale(Bytes,SampleRate,Samples);
 		}
		else
		if (memcmp(Buffer, "ADIF", 4) == 0) 
		{
			int BitRate;

			if (Buffer[4] & 0x80)
				Buffer += 9;

			BitRate = (((int)Buffer[4] & 0x0F) << 19) |
					  ((int)Buffer[5] << 11) |
					  ((int)Buffer[6] << 3) |
					  ((int)Buffer[7] & 0xE0);

			p->InputFormat.ByteRate = BitRate/8;
		}

		Readed = NeAACDecInit(p->Decoder, p->Merge.Data + p->Merge.ReadPos,
			p->Merge.WritePos - p->Merge.ReadPos, &SampleRate, &Channels);
		
		if (Readed < 0)
		{
			// this is a problem...
			MergeClear(&p->Merge);
			return ERR_INVALID_DATA;
		}

		p->Inited = 1;
		p->Merge.ReadPos += Readed;

		if (p->OutputFormat.SampleRate != (int)SampleRate ||
			p->OutputFormat.Channels != (int)Channels)
		{
			// set new output format
			p->InputFormat.SampleRate = p->OutputFormat.SampleRate = SampleRate;
			p->InputFormat.Channels = p->OutputFormat.Channels = Channels;
			
			if (p->Output.Node)
				p->Output.Node->Set(p->Output.Node,p->Output.No|PACKET_FORMAT,&p->OutputFormat,sizeof(audio));
		}
	}

	if (DecodeFrame(p,1))
	{
		Result = ERR_NONE;
		do
		{
			int i = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));
			if (i == ERR_BUFFER_FULL)
			{
				p->Pending = 1;
				//return last result
				break;
			}

			Result = i;
		}
		while (DecodeFrame(p,0));
	}
	else
		Result = ERR_NEED_MORE_DATA; // if no frame decoded: need more packet

	return Result;
}

static bool_t IsEmpty( faad* p )
{
	packet Zero;

	if (!p->InputFormat.Format) 
		return 1;

	Zero.CurrTime = -1;
	Zero.DropLevel = p->Packet.DropLevel;
	Zero.Length = 0;
	Zero.Data[0] = NULL;
	Zero.RefTime = -1;

	return Decode(p,&Zero) == ERR_NEED_MORE_DATA;
}

const datadef FAADParams[] = 
{
	{ FAAD_ID,		TYPE_CLASS },
	{ FAAD_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ FAAD_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },

	DATADEF_END
};

static int Enum( faad* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,FAADParams,NULL);
}

static int Get( faad* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case FAAD_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case FAAD_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case FAAD_OUTPUT: GETVALUE(p->Output,pin); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_EMPTY:GETVALUE(IsEmpty(p),bool_t); break;
	}
	return Result;
}

static int Set( faad* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case FAAD_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case FAAD_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case FAAD_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_EMPTY:

		// reset
		p->Pending = 0;
		p->Eof = 0;
		MergeClear(&p->Merge);

		if (p->Decoder)
			NeAACDecPostSeekReset(p->Decoder,-1);
		break;
	}
	return Result;
}

faad FAAD;

void FAAD_Init(int Id,int Pri)
{
	NodeFill(&FAAD.VMT,sizeof(FAAD),Enum,Get,Set);
	FAAD.Id = Id;
	NodeRegister(&FAAD.VMT,Pri);
}

void FAAD_Done()
{
	NodeUnRegister(&FAAD.VMT);
}

