/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: avi.c 117 2004-11-28 02:45:11Z picard $
 *
 * BetaPlayer MPEG4AAC plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 * used libavformat and faad2 as .mp4 format reference
 *
 ****************************************************************************/

#include "../common/common.h"
#include "mpeg4aac.h"

typedef struct mp4chunk
{
	int Offset;
	int Samples;

} mp4chunk;

typedef struct mp4sample
{
	int Size:31;
	int Key:1;
	int Duration;

} mp4sample;

typedef struct mp4stsc
{
	int FirstChunk;
	int Samples;

} mp4stsc;

typedef struct mp4stts
{
	int Count;
	int Delta;

} mp4stts;

typedef struct mp4stream
{
	format_stream Stream;

	int TimeScale;

	// playback
	mp4sample* CurrentSample;
	mp4chunk* CurrentChunk;
	int CurrentTime;
	int SamplesLeft;

	// chunk information
	mp4chunk* Chunk;
	mp4chunk* ChunkEnd;

	// sample information
	mp4sample* Sample;
	mp4sample* SampleEnd;
	bool_t HasKey;

	// stsc information (temporary)
	mp4stsc* STSC;
	mp4stsc* STSCEnd;
	mp4stts* STTS;
	mp4stts* STTSEnd;
	int32_t* STSZ;
	int32_t* STSZEnd;
	int STSZDef;
	
} mp4stream;

typedef struct mp4
{
	format_base Format;

	int Type;
	int CompressType;
	bool_t ErrorShowed;
	bool_t MOOVFound;
	bool_t MDATFound;
	int TimeScale;

	mp4stream* Current;
	tchar_t Meta[32];

} mp4;

typedef	void (*mp4atomfunc)(mp4* p, int EndOfAtom);

typedef	struct mp4atomtable
{
	int Atom;
	mp4atomfunc Func;

} mp4atomtable;

typedef struct mp4atom
{
	int Atom;
	int EndOfAtom;

} mp4atom;

void NotSupported(mp4* p)
{
	if (!p->ErrorShowed)
	{
		p->ErrorShowed = 1;
		NodeError((node*)&p->Format.VMT,MP4_ID,MP4_NOT_SUPPORTED);
	}
}

static int ReadVersion(mp4* p)
{
	int Ver = Format_Read8(&p->Format);
	Format_Skip(&p->Format,3); //flags;
	return Ver;
}

static bool_t ReadAtom(mp4* p, mp4atom* Atom)
{
	int Size;
	Atom->Atom = 0;
	Atom->EndOfAtom = p->Format.FilePos;
	Size = Format_ReadBE32(&p->Format);
	Format_Read(&p->Format,&Atom->Atom,4);
	if (Size == 1)
	{
		// 64bit size not supported
		Size = (int)Format_ReadBE64(&p->Format);
	}
	Atom->EndOfAtom += Size;
	return Size >= 8;
}

mp4atomtable AtomTable[];

static void ReadSubAtoms(mp4* p, int EndOfAtom)
{
	mp4atom Atom;
	mp4atomtable* i;

	while (p->Format.FilePos < EndOfAtom - 8)
	{
		if (!ReadAtom(p,&Atom))
			break;

		if (p->MOOVFound && Atom.Atom == FOURCC('m','d','a','t'))
			break;

		for (i=AtomTable;i->Atom;++i)
			if (i->Atom == Atom.Atom)
			{
				i->Func(p,Atom.EndOfAtom);
				break;
			}

		Format_FileSeek(&p->Format,Atom.EndOfAtom,SEEK_SET);
	}
}

static void ReadFTYP(mp4* p, int EndOfAtom)
{
	p->Type = 0;
	Format_Read(&p->Format,&p->Type,4);
}

static void ReadMOOV(mp4* p, int EndOfAtom)
{
	p->MOOVFound = 1;
	ReadSubAtoms(p,EndOfAtom);
}

static void ReadTRAK(mp4* p,int EndOfAtom)
{
	mp4stream* Stream;
	p->Current = (mp4stream*) Format_AddStream(&p->Format,sizeof(mp4stream));

	ReadSubAtoms(p,EndOfAtom);

	Stream = p->Current;
	if (Stream)
	{
		if (Stream->STTS) // fragmanted
		{
			int Count = Stream->ChunkEnd - Stream->Chunk;
			mp4sample* s = CAlloc(sizeof(mp4sample),Count);
			if (s)
			{
				mp4chunk* c = Stream->Chunk;
				mp4stts* STTS = Stream->STTS;
				int32_t* STSZ = Stream->STSZ;
			
				Stream->Sample = s;
				Stream->SampleEnd = s+Count;

				for (;s!=Stream->SampleEnd;++s,++c)
				{
					int n;
					for (n=0;n<c->Samples;++n)
					{
						if (STSZ && STSZ < Stream->STSZEnd)
							s->Size += *(STSZ++);
						else
							s->Size += Stream->STSZDef;

						if (STTS < Stream->STTSEnd)
						{
							s->Duration += STTS->Delta;
							if (--STTS->Count<=0)
								++STTS;
						}
					}

					c->Samples = 1;
				}
			}

			Free(Stream->STTS);
			Stream->STTS = NULL;
			Stream->STTSEnd = NULL;

			Free(Stream->STSZ);
			Stream->STSZ = NULL;
			Stream->STSZEnd = NULL;
		}

		Stream->CurrentTime = 0;
		Stream->CurrentSample = Stream->Sample;
		Stream->CurrentChunk = Stream->Chunk;

		Format_PrepairStream(&p->Format,&Stream->Stream);

		p->Current = NULL;
	}
}

static void ReadMETA(mp4* p,int EndOfAtom)
{
	ReadVersion(p);
	ReadSubAtoms(p,EndOfAtom);
}

static void ReadSTSS(mp4* p,int EndOfAtom)
{
	mp4stream* Stream = p->Current;
	int Count;

	if (!Stream || Stream->Stream.Fragmented) return;
	ReadVersion(p);

	Stream->HasKey = 1;
	Count = Format_ReadBE32(&p->Format);

	while (Count-->0)
	{
		mp4sample* s = Stream->Sample+Format_ReadBE32(&p->Format)-1;
		if (s >= Stream->Sample && s < Stream->SampleEnd)
			s->Key = 1;
	}
}

static void ReadSTSC(mp4* p,int EndOfAtom)
{
	mp4stream* Stream = p->Current;
	int Count;
	mp4stsc* s;

	if (!Stream) return;
	ReadVersion(p);

	Count = Format_ReadBE32(&p->Format);
	s = (mp4stsc*) Alloc(sizeof(mp4stsc)*(Count+1)); // plus one delimiter

	if (s)
	{
		Stream->STSC = s;
		Stream->STSCEnd = s+Count;

		for (;s!=Stream->STSCEnd;++s)
		{
			s->FirstChunk = Format_ReadBE32(&p->Format)-1; 
			s->Samples = Format_ReadBE32(&p->Format);
			Format_ReadBE32(&p->Format); // desc
		}
	}
}

static void ReadSTCO(mp4* p,int EndOfAtom)
{
	mp4stream* Stream = p->Current;
	int Count;
	mp4chunk* c;
	mp4stsc* s;
	int n;

	if (!Stream) return;
	ReadVersion(p);

	Count = Format_ReadBE32(&p->Format);
	c = (mp4chunk*) Alloc(sizeof(mp4chunk)*Count);

	if (c)
	{
		Stream->Chunk = c;
		Stream->ChunkEnd = c+Count;

		for (;c!=Stream->ChunkEnd;++c)
			c->Offset = Format_ReadBE32(&p->Format);

		if (Stream->STSC)
		{
			Stream->STSCEnd->FirstChunk = Count; // delimiter
			for (s=Stream->STSC;s!=Stream->STSCEnd;++s)
				for (n=s[0].FirstChunk;n<s[1].FirstChunk;++n)
					Stream->Chunk[n].Samples = s->Samples;

			Free(Stream->STSC);
			Stream->STSC = NULL;
			Stream->STSCEnd = NULL;
		}
	}
}

static void ReadCTTS(mp4* p,int EndOfAtom)
{
/*	mp4stream* Stream = p->Current;
	mp4sample* s;
	int Count;

	if (!Stream) return;
	ReadVersion(p);

	s = Stream->Sample;
	Count = Format_ReadBE32(&p->Format);
	while (Count-->0 && s<Stream->SampleEnd)
	{
		int Samples = Format_ReadBE32(&p->Format);
		int Offset = Format_ReadBE32(&p->Format);

 		if (Offset)
			while (Samples-->0 && s<Stream->SampleEnd)
			{
				//present time offset
				++s;
			}
		else
			s += Samples;
	}
*/
}

static void ReadSTSZ(mp4* p,int EndOfAtom)
{
	mp4stream* Stream = p->Current;
	mp4sample* s;
	mp4sample* se;
	int Count;
	int DefSize;

	if (!Stream) return;
	ReadVersion(p);

	DefSize = Format_ReadBE32(&p->Format);
	Count = Format_ReadBE32(&p->Format);

	if (Stream->Stream.Fragmented)
	{
		Stream->STSZDef = DefSize;
		if (!DefSize)
		{
			int32_t* s = (int32_t*)Alloc(sizeof(int32_t)*Count);
			if (s)
			{
				Stream->STSZ = s;
				Stream->STSZEnd = s+Count;

				for (;s!=Stream->STSZEnd;++s)
					*s = Format_ReadBE32(&p->Format);
			}
		}
	}
	else
	if (DefSize)
	{
		se = Stream->SampleEnd;
		for (s=Stream->Sample;s!=se;++s)
			s->Size = DefSize;
	}
	else
	{
		se = Stream->Sample+Count;
		if (se > Stream->SampleEnd)
			se = Stream->SampleEnd;

		for (s=Stream->Sample;s!=se;++s)
			s->Size = Format_ReadBE32(&p->Format);
	}
}

static void ReadSTTS(mp4* p,int EndOfAtom)
{
	mp4stream* Stream = p->Current;
	mp4stts* s;
	int Sum = 0;
	int Count;
	int i;

	if (!Stream) return;
	ReadVersion(p);

	Count = Format_ReadBE32(&p->Format);
	s = (mp4stts*) Alloc(sizeof(mp4stts)*Count);
	if (s)
	{
		Stream->STTS = s;
		Stream->STTSEnd = s+Count;

		for (;s!=Stream->STTSEnd;++s)
		{
			s->Count = Format_ReadBE32(&p->Format); 
			s->Delta = Format_ReadBE32(&p->Format);
			Sum += s->Count;
		}

		if (!Stream->Stream.Fragmented)
		{
			mp4sample* Sample = (mp4sample*) CAlloc(sizeof(mp4sample),Sum);
			if (Sample)
			{
				Stream->Sample = Sample;
				Stream->SampleEnd = Sample+Sum;

				for (s=Stream->STTS;s!=Stream->STTSEnd;++s)
					for (i=0;i<s->Count;++i,++Sample)
						Sample->Duration = s->Delta;
			}

			Free(Stream->STTS);
			Stream->STTS = NULL;
			Stream->STTSEnd = NULL;
		}
	}
}

static void ReadExtra(mp4* p,int EndOfAtom)
{
	int Len = EndOfAtom - p->Format.FilePos;
	void* Extra; 

	if (!p->Current || Len<=0) return;

	Extra = Format_StreamExtra(&p->Current->Stream,Len);
	if (Extra)
		Format_Read(&p->Format,Extra,Len);
}

static int DescLen(mp4* p)
{
	int Len = 0;
	int i;
	for (i=0;i<4;++i)
	{
		int v = Format_Read8(&p->Format);
		Len = (Len << 7) | (v & 0x7F);
		if (!(v & 0x80))
			break;
	}
	return Len;
}

static void ReadESDS(mp4* p,int EndOfAtom)
{
	int Tag,Len;

	if (!p->Current) return;

	ReadVersion(p);

	Tag = Format_Read8(&p->Format);
	Len = DescLen(p);

	Format_Skip(&p->Format,2); //ID
	if (Tag == 3)
		Format_Skip(&p->Format,1); //priority

	Tag = Format_Read8(&p->Format);
	Len = DescLen(p);

	if (Tag == 4) //config descr
	{
		Format_Skip(&p->Format,13); //obj_type_id,stream_type,buffer_size,max_bitrate,min_bitrate
		
		Tag = Format_Read8(&p->Format);
		Len = DescLen(p);

		if (Tag == 5)
		{
			void* Extra = Format_StreamExtra(&p->Current->Stream,Len);
			if (Extra)
				Format_Read(&p->Format,Extra,Len);
		}
	}
}

static void SetAudio(mp4* p,int Format,int SampleRate,int Bits,int Channels)
{
	audio* Audio = (audio*) Format_StreamFormat(&p->Current->Stream,PACKET_AUDIO,sizeof(audio));
	if (Audio)
	{
		Audio->Format = Format;
		Audio->Flags = PCM_PACKET_BASED;
		Audio->SampleRate = SampleRate;
		Audio->Bits = Bits;
		Audio->Channels = Channels;
		
		Format_DefaultAudio(Audio);
	}
}

static void ReadAudio(mp4* p,int Format,int Atom,int EndOfAtom)
{
	audio* Audio = (audio*) Format_StreamFormat(&p->Current->Stream,PACKET_AUDIO,sizeof(audio));
	if (Audio)
	{
		Audio->Format = Format;
		Audio->Channels = 1;
		Audio->Flags = PCM_PACKET_BASED;

		if (Atom == FOURCC('r','a','w',' '))
			Audio->Bits = 8;
		else
			Audio->Bits = 16;

		if (EndOfAtom - p->Format.FilePos >= 20)
		{
			Format_Skip(&p->Format,2+2+4); // version,revision, vendor

			Audio->Channels = Format_ReadBE16(&p->Format); // channel count
			Audio->Bits = Format_ReadBE16(&p->Format); // sample size

			Format_Skip(&p->Format,4); // reserved

			Audio->SampleRate = Format_ReadBE16(&p->Format); // sample rate
		
			Format_Skip(&p->Format,2); // reserved

			ReadSubAtoms(p,EndOfAtom);
		}

		Format_DefaultAudio(Audio);
	}
}

static void ReadVideo(mp4* p,int FourCC,int EndOfAtom)
{
	video* Video = (video*) Format_StreamFormat(&p->Current->Stream,PACKET_VIDEO,sizeof(video));
	if (Video)
	{
		Format_Skip(&p->Format,16); // reserved

		Video->Pixel.FourCC = FourCC;
		Video->Width = Format_ReadBE16(&p->Format);
		Video->Height = Format_ReadBE16(&p->Format); 

		if (FourCC == FOURCC('m','p','4','v'))
		{
			// use mpeg4 vol information instead 
			Video->Width = 0;
			Video->Height = 0;
		}

		Format_Skip(&p->Format,14+32); // reserved,compressor

		Video->Pixel.BitCount = Format_ReadBE16(&p->Format);

		Format_Skip(&p->Format,2); // colortable
    
		ReadSubAtoms(p,EndOfAtom);

		Format_DefaultVideo(Video);
	}
}

static void ReadSTSD(mp4* p,int EndOfAtom)
{
	mp4stream* Stream = p->Current;
	int DescCount;
	mp4atom Atom;

	if (!Stream) return;

	ReadVersion(p);

	DescCount = Format_ReadBE32(&p->Format);
	if (DescCount && ReadAtom(p,&Atom))
	{
		Format_Skip(&p->Format,6+2); // reserved,reference index

		Stream->Stream.Fragmented = 0;

		// fourcc codes from libavformat

		switch (Atom.Atom)
		{
		case FOURCC('a','v','c','1'):
		case FOURCC('j','p','e','g'):
		case FOURCC('m','j','p','a'):
		case FOURCC('m','j','p','b'):
		case FOURCC('S','V','Q','1'):
		case FOURCC('S','V','Q','3'):
		case FOURCC('m','p','4','v'):
		case FOURCC('D','I','V','X'):
		case FOURCC('V','P','3','1'):
		case FOURCC('h','2','6','3'):
			ReadVideo(p,Atom.Atom,Atom.EndOfAtom);
			break;

		case FOURCC('r','a','w',' '):
			Stream->Stream.Fragmented = 1;
			ReadAudio(p,AUDIOFMT_PCM,Atom.Atom,Atom.EndOfAtom);
			break;

		case FOURCC('u','l','a','w'):
			Stream->Stream.Fragmented = 1;
			ReadAudio(p,AUDIOFMT_MULAW,Atom.Atom,Atom.EndOfAtom);
			break;

		case FOURCC('a','l','a','w'):
			Stream->Stream.Fragmented = 1;
			ReadAudio(p,AUDIOFMT_ALAW,Atom.Atom,Atom.EndOfAtom);
			break;

		case FOURCC('.','m','p','3'):
		case 0x6D730055:
		case 0x5500736D:
			ReadAudio(p,AUDIOFMT_MP3,Atom.Atom,Atom.EndOfAtom);
			break;

		case FOURCC('m','p','4','a'):
			ReadAudio(p,AUDIOFMT_AAC,Atom.Atom,Atom.EndOfAtom);
			break;

		case FOURCC('s','a','m','r'):
			SetAudio(p,AUDIOFMT_AMR_NR,8000,16,1);
			break;
		}
	}

	if (!Stream->Stream.FormatType)
	{
		Format_RemoveStream(&p->Format);
		p->Current = NULL;
	}
}

static void ReadMVHD(mp4* p,int EndOfAtom)
{
	int Duration;

	ReadVersion(p);
	Format_ReadBE32(&p->Format); // creation time
	Format_ReadBE32(&p->Format); // modification time
	
	p->TimeScale = Format_ReadBE32(&p->Format); // timescale
	Duration = Format_ReadBE32(&p->Format); // duration

	if (Duration>0)
		p->Format.Duration = Scale(Duration,TICKSPERSEC,p->TimeScale);
}

static void ReadMDHD(mp4* p,int EndOfAtom)
{
	int Duration = -1;

	if (!p->Current) return;

	switch (ReadVersion(p))
	{
	case 0:
		Format_ReadBE32(&p->Format); // creation time
		Format_ReadBE32(&p->Format); // modification time
		p->Current->TimeScale = Format_ReadBE32(&p->Format); // timescale
		Duration = Format_ReadBE32(&p->Format); // duration
		break;

	case 1:
		Format_ReadBE64(&p->Format); // creation time
		Format_ReadBE64(&p->Format); // modification time
		p->Current->TimeScale = Format_ReadBE32(&p->Format); // timescale
		Duration = (int)Format_ReadBE64(&p->Format); // duration
		break;
	}

	if (Duration>0)
	{
		Duration = Scale(Duration,TICKSPERSEC,p->Current->TimeScale);
		if (p->Format.Duration < Duration)
			p->Format.Duration = Duration;
	}
}

static void ReadCMOV(mp4* p,int EndOfAtom)
{
	ReadSubAtoms(p,EndOfAtom);
	p->CompressType = 0;
}

static void ReadDCOM(mp4* p,int EndOfAtom)
{
	p->CompressType = 0;
	Format_Read(&p->Format,&p->CompressType,4);
}

static void ReadCMVD(mp4* p,int EndOfAtom)
{
	if (p->CompressType = FOURCC('z','l','i','b'))
	{
		//!!!
		//zlib uncompress
	}
}

static void AddMeta(mp4* p,const tchar_t* Meta,int Size)
{
	format_stream* Stream = (format_stream*)p->Current;
	if (!Stream && p->Format.StreamCount)
		Stream = p->Format.Streams[0];
	if (Stream && Stream->Comment.Node)
		Stream->Comment.Node->Set(Stream->Comment.Node,Stream->Comment.No,Meta,Size);
}

static void ReadMetaString(mp4* p,int EndOfAtom)
{
	tchar_t Meta[512];
	char UTF8[512];
	int Len = EndOfAtom - p->Format.FilePos;
	if (Len>=512) Len=511;
	Len = Format_Read(&p->Format,UTF8,Len);
	if (Len>0)
	{
		UTF8[Len] = 0;
		tcscpy(Meta,p->Meta);
		UTF8ToTcs(Meta+tcslen(p->Meta),UTF8,512-tcslen(p->Meta));
		AddMeta(p,Meta,sizeof(Meta));
	}
}

static void ReadDATA(mp4* p,int EndOfAtom)
{
	ReadVersion(p);
	Format_Skip(&p->Format,4); //reserved;
	if (tcscmp(p->Meta,T("GENRE="))==0)
	{
		tchar_t Meta[256];
		tcscpy(Meta,p->Meta);
		if (Id3Genre(Format_ReadBE16(&p->Format)-1,Meta+tcslen(Meta),256-tcslen(Meta)))
			AddMeta(p,Meta,sizeof(Meta));
	}
	else
	if (p->Meta[0])
		ReadMetaString(p,EndOfAtom);
}

static void ReadNAME(mp4* p,int EndOfAtom)
{
	ReadVersion(p);
	if (p->Meta[0])
		ReadMetaString(p,EndOfAtom);
}

static void ReadMeta(mp4* p,int EndOfAtom, const tchar_t* Meta)
{
	tcscpy(p->Meta,Meta);
	ReadSubAtoms(p,EndOfAtom);
	p->Meta[0] = 0;
}

static void ReadMetaName(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("TITLE=")); }
static void ReadMetaArtist(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("ARTIST=")); }
static void ReadMetaWriter(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("WRITER=")); }
static void ReadMetaAlbum(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("ALBUM=")); }
static void ReadMetaDate(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("DATE=")); }
static void ReadMetaTool(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("TOOL=")); }
static void ReadMetaComment(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("COMMENT=")); }
static void ReadMetaGenre(mp4* p,int EndOfAtom) { ReadMeta(p,EndOfAtom,T("GENRE=")); }

mp4atomtable AtomTable[] =
{
	{ FOURCC('f','t','y','p'), ReadFTYP },
	{ FOURCC('m','o','o','v'), ReadMOOV },
	{ FOURCC('t','r','a','k'), ReadTRAK },
	{ FOURCC('m','v','h','d'), ReadMVHD },
	{ FOURCC('m','d','h','d'), ReadMDHD },
	{ FOURCC('s','t','b','l'), ReadSubAtoms },
	{ FOURCC('m','i','n','f'), ReadSubAtoms },
	{ FOURCC('e','d','t','s'), ReadSubAtoms },
	{ FOURCC('m','d','i','a'), ReadSubAtoms },
	{ FOURCC('u','d','t','a'), ReadSubAtoms },
	{ FOURCC('i','l','s','t'), ReadSubAtoms },
	{ FOURCC('m','e','t','a'), ReadMETA },
	{ FOURCC('s','t','s','d'), ReadSTSD },
	{ FOURCC('s','t','s','s'), ReadSTSS },
	{ FOURCC('s','t','s','c'), ReadSTSC },
	{ FOURCC('s','t','c','o'), ReadSTCO },
	{ FOURCC('c','t','t','s'), ReadCTTS },
	{ FOURCC('s','t','s','z'), ReadSTSZ },
	{ FOURCC('s','t','t','s'), ReadSTTS },
	{ FOURCC('c','m','o','v'), ReadCMOV },
	{ FOURCC('d','c','o','m'), ReadDCOM },
	{ FOURCC('c','m','v','d'), ReadCMVD },
	{ FOURCC('e','s','d','s'), ReadESDS },
	{ FOURCC('S','M','I',' '), ReadExtra },
	{ FOURCC('d','a','t','a'), ReadDATA },
	{ FOURCC('n','a','m','e'), ReadNAME },
	{ FOURCC('\251','n','a','m'), ReadMetaName },
	{ FOURCC('\251','A','R','T'), ReadMetaArtist },
	{ FOURCC('\251','w','r','t'), ReadMetaWriter },
	{ FOURCC('\251','a','l','b'), ReadMetaAlbum },
	{ FOURCC('\251','d','a','y'), ReadMetaDate },
	{ FOURCC('\251','t','o','o'), ReadMetaTool },
	{ FOURCC('\251','c','m','t'), ReadMetaComment },
	{ FOURCC('\251','g','e','n'), ReadMetaGenre },
	{ FOURCC('g','n','r','e'), ReadMetaGenre },

	{ 0, NULL }
};

static mp4stream* FindNextChunk(mp4* p)
{
	mp4stream* Stream = NULL;
	int Min = MAX_INT;
	int No;
	for (No=0;No<p->Format.StreamCount;++No)
	{
		mp4stream* s = (mp4stream*)p->Format.Streams[No];
		if (s->CurrentSample != s->SampleEnd && 
			s->CurrentChunk != s->ChunkEnd && 
			s->CurrentChunk->Offset < Min)
		{
			Stream = s;
			Min = s->CurrentChunk->Offset;
		}
	}

	if (Stream)
	{
		if (Format_FileSeek(&p->Format,Min,SEEK_SET) == ERR_NONE)
		{
			p->Current = Stream;
			Stream->SamplesLeft = Stream->CurrentChunk->Samples;
		}
		else
			Stream = NULL;
	}

	return Stream;
}

static int Init(mp4* p)
{
	p->MOOVFound = 0;
	p->Type = FOURCC('m','o','o','v'); //quicktime .mov
	p->TimeScale = 0;
	p->ErrorShowed = 0;
	p->Format.HeaderLoaded = 1;
	p->Meta[0] = 0;

	ReadSubAtoms(p,MAX_INT);

	if (!p->MOOVFound)
		return ERR_INVALID_DATA;

	FindNextChunk(p);
	return ERR_NONE;
}

static int Probe(mp4* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

    if (Data[4]=='f' &&	Data[5]=='t' &&	Data[6]=='y' &&	Data[7]=='p')
		return ERR_NONE;

    if (Data[4]=='m' &&	Data[5]=='o' &&	Data[6]=='o' &&	Data[7]=='v')
		return ERR_NONE;

	if (CheckExts(URL,p->Format.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static int ReadPacket(mp4* p, format_packet* Packet)
{
	mp4stream* Stream = p->Current;
	mp4sample* Sample;

	if (!Stream)
	{
		Stream = FindNextChunk(p);
		if (!Stream)
			return ERR_END_OF_FILE;
	}

	Sample = Stream->CurrentSample++;

	Packet->Stream = &Stream->Stream;
	Packet->RefTime = Scale(Stream->CurrentTime,TICKSPERSEC,Stream->TimeScale);
	Packet->Key = Sample->Key != 0;
	Packet->Data = Format_ReadAsRef(&p->Format,Sample->Size);
	
	Stream->CurrentTime += Sample->Duration;
	if (--Stream->SamplesLeft==0 || Stream->CurrentSample==Stream->SampleEnd)
	{
		p->Current = NULL;
		if (Stream->CurrentChunk != Stream->ChunkEnd)
			++Stream->CurrentChunk;
	}

	return ERR_NONE;
}

static int Seek(mp4* p, tick_t DstTime, int DstPos )
{
	bool_t NeedKey;
	mp4sample* Sample;
	mp4chunk* Chunk;
	int SamplesLeft;
	int Pos,No;
	int Time;
	mp4stream* Stream = (mp4stream*) Format_DefSyncStream(&p->Format);

	if (!Stream || Stream->Chunk == Stream->ChunkEnd)
		return ERR_NOT_SUPPORTED;

	if (DstTime >= 0)
	{
		DstTime = Scale(DstTime,Stream->TimeScale,TICKSPERSEC);
		DstPos = -1;
	}

	NeedKey = Stream->HasKey && Stream->Stream.FormatType == PACKET_VIDEO;

	// find sample 
	Time = 0;
	Sample = Stream->Sample;
	for (Chunk=Stream->Chunk;Chunk<Stream->ChunkEnd;++Chunk)
	{
		Pos = Chunk->Offset;
		SamplesLeft = Chunk->Samples;

		while (SamplesLeft>0 && Sample<Stream->SampleEnd)
		{
			if (DstTime >= 0 && Time >= DstTime && (!NeedKey || Sample->Key))
				goto found;

			if (DstPos >= 0 && Pos >= DstPos)
				goto found;

			Pos += Sample->Size;
			Time += Sample->Duration;
			++Sample;
			--SamplesLeft;
		}
	}
found:

	if (Format_Seek(&p->Format,Pos,SEEK_SET) != ERR_NONE)
		return ERR_NOT_SUPPORTED;

	p->Current = SamplesLeft ? Stream:0;
	Stream->SamplesLeft = SamplesLeft;
	Stream->CurrentSample = Sample;
	Stream->CurrentChunk = Chunk;
	Stream->CurrentTime = Time;

	// set current positions for other streams
	for (No=0;No<p->Format.StreamCount;++No)
	{
		mp4stream* s = (mp4stream*) p->Format.Streams[No];
		if (s != Stream)
		{
			Time = 0;
			Sample = s->Sample;
			for (Chunk=s->Chunk;Chunk<s->ChunkEnd;++Chunk)
			{
				if (Chunk->Offset >= Pos)
					break;

				SamplesLeft = Chunk->Samples;
				while (SamplesLeft>0 && Sample<s->SampleEnd)
				{
					Time += Sample->Duration;
					++Sample;
					--SamplesLeft;
				}
			}

			s->CurrentTime = Time;
			s->CurrentChunk = Chunk;
			s->CurrentSample = Sample;
		}
	}

	return ERR_NONE;
}

static void FreeStream(mp4* p, mp4stream* Stream)
{
	Free(Stream->Chunk);
	Free(Stream->Sample);
	Free(Stream->STSC);
	Free(Stream->STTS);
	Free(Stream->STSZ);
}

static mp4 MP4;

void MP4_Init()
{
	MP4.Format = FormatBase;
	MP4.Format.VMT.Probe = Probe;
	MP4.Format.Init = Init;
	MP4.Format.Seek = Seek;
	MP4.Format.ReadPacket = ReadPacket;
	MP4.Format.FreeStream = FreeStream;
	MP4.Format.Id = MP4_ID;
	MP4.Format.Exts = T("mp4:V;3gp:V;m4a:A;m4v:V;mov:V");

	NodeRegister((node*)&MP4.Format.VMT,PRI_DEFAULT);
}

void MP4_Done()
{
	NodeUnRegister((node*)&MP4.Format.VMT);
}

