/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: libmad.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer MPEGAudio plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "..\common\common.h"
#include "mpegaudio.h"
#include "libmad\frame.h"
#include "libmad\synth.h"
#include <math.h>

#define LIBMAD_INPUT		0x100
#define LIBMAD_OUTPUT		0x101
#define LIBMAD_EQ			0x102
#define LIBMAD_COMMENTS		0x103

typedef struct libmad
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;
	pin Output;
	pin Comments;

	equalizer Eq;

	merge Merge;
	struct mad_stream Stream;
	struct mad_frame Frame;
	struct mad_synth Synth;

	// output data
	packet Packet;
	tick_t RefTime;
	int AdjustBytes; // adjust (rewind) RefTime
	bool_t Skip; // skip next frame
	bool_t Pending;
	int AdjustTime;
	int FormatSet;
	int Total;

	mad_fixed_t Scale;
	mad_fixed_t EqTable[32];

	mad_fixed_t EqScale[32];

} libmad;

const datadef LibMadParams[] = 
{
	{ LIBMAD_ID,		TYPE_CLASS },
	{ LIBMAD_INPUT,		TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ LIBMAD_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },
	{ LIBMAD_COMMENTS,	TYPE_PIN, DF_OUTPUT, PACKET_COMMENT },
	{ LIBMAD_EQ,		TYPE_EQUALIZER, DF_HIDDEN },

	DATADEF_END
};

static void UpdateScale( libmad* p )
{
	int i;
	for (i=0;i<32;++i)
		p->EqScale[i] = mad_f_mul(p->EqTable[i],p->Scale);
}

static int UpdateEq( libmad* p )
{
	static const uint8_t Map[32] = 
	{
		0, 1, 2, 3, 4, 5, 6, 6, 
		7, 7, 7, 7, 8, 8, 8, 8,
		8, 8, 8, 8, 9, 9, 9, 9, 
		9, 9, 9, 9, 9, 9, 9, 9
	};

	int i;
	for (i=0;i<32;++i)
	{
		int db = p->Eq.Amplify + p->Eq.Eq[Map[i]];
		if (db>18)
			db=18;
		p->EqTable[i] = mad_f_tofixed(pow(10,db/20.f));
	}

	p->Scale=MAD_F_ONE;
	UpdateScale(p);
	return ERR_NONE;
}

static inline void UpdateAdjustTime( libmad* p )
{
	if (p->InputFormat.ByteRate)
		p->AdjustTime = (TICKSPERSEC * 4096) / p->InputFormat.ByteRate;
	else
		p->AdjustTime = 0;
}

static int SetFormat( libmad* p, const audio* Format )
{
	if (p->InputFormat.Format)
	{
		memset(&p->InputFormat,0,sizeof(p->InputFormat));
		MergeDone(&p->Merge);
		mad_frame_finish(&p->Frame);
		mad_stream_finish(&p->Stream);
		mad_synth_finish(&p->Synth);
	}

	if (Format)
	{
		// accept only mp2 and mp3 
		if (Format->Format != AUDIOFMT_MP2 && Format->Format != AUDIOFMT_MP3)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		p->OutputFormat.Bits = sizeof(mad_fixed_t)*8;
		p->OutputFormat.Flags = PCM_PLANES;
		p->OutputFormat.Format = AUDIOFMT_PCM;
		p->OutputFormat.FracBits = MAD_F_FRACBITS;
		p->OutputFormat.SampleRate = p->InputFormat.SampleRate;	// optional
		p->OutputFormat.Channels = p->InputFormat.Channels; // optional

		p->Skip = 0;
		p->Pending = 0;
		p->FormatSet = 0;
		p->RefTime = 0;
		p->AdjustBytes = 0;
		p->Total = 0;

		UpdateAdjustTime(p);
		UpdateEq(p);

		MergeInit(&p->Merge);
		mad_stream_init(&p->Stream);
		mad_frame_init(&p->Frame);
		mad_synth_init(&p->Synth);
	}

	return ERR_NONE;
}

void Rescale( libmad* p, mad_fixed_t v, int ramp )
{
	mad_fixed_t dv;
	int ch;

	if (p->Synth.pcm.length < 4) return;
	while ((1<<ramp) > p->Synth.pcm.length) --ramp;
	dv = (v-MAD_F_ONE) >> ramp;

	for (ch=0;ch<p->Synth.pcm.channels;++ch)
	{
		mad_fixed_t* i=p->Synth.pcm.samples[ch];
		mad_fixed_t* ie=i+(1<<ramp);

		v = MAD_F_ONE;
		for (;i!=ie;++i,v+=dv)
			i[0] = mad_f_mul(i[0],v);

		ie += p->Synth.pcm.length-(1<<ramp);
		for (;i!=ie;i+=4)
		{
			i[0] = mad_f_mul(i[0],v);
			i[1] = mad_f_mul(i[1],v);
			i[2] = mad_f_mul(i[2],v);
			i[3] = mad_f_mul(i[3],v);
		}
	}
}

bool_t DecodeFrame( libmad* p )
{
	for (;;)
	{
		while (mad_frame_decode(&p->Frame, &p->Stream) == -1)
		{
			if (p->Stream.error == MAD_ERROR_LOSTSYNC && p->Comments.Node)
			{
				int TagSize = Id3TagQuery(p->Stream.this_frame,p->Stream.bufend - p->Stream.this_frame);
				if (TagSize > 0) 
				{
					if (TagSize > p->Stream.bufend - p->Stream.this_frame)
					{
						MergePack(&p->Merge,p->Stream.this_frame - p->Stream.buffer);
						return 0;
					}

					Id3TagParse(p->Stream.this_frame, TagSize, &p->Comments);
				}
			}
			else if (p->Stream.error == MAD_ERROR_BUFLEN || p->Stream.error == MAD_ERROR_BUFPTR)
			{
				MergePack(&p->Merge,p->Stream.next_frame - p->Stream.buffer);
				return 0;
			}
		}

		if (p->Eq.Enabled)
		{
			int n = MAD_NCHANNELS(&p->Frame.header);
			int m = MAD_NSBSAMPLES(&p->Frame.header);
			int ch;

			for (ch=0;ch<n;++ch)
			{
				mad_fixed_t* i;
				mad_fixed_t* ie;
				int eq=0;

				ie=p->Frame.sbsample[ch][m];
				for (i=p->Frame.sbsample[ch][0];i!=ie;i+=4,eq+=4)
				{
					eq &= 31;
					i[0] = mad_f_mul(i[0],p->EqScale[eq+0]);
					i[1] = mad_f_mul(i[1],p->EqScale[eq+1]);
					i[2] = mad_f_mul(i[2],p->EqScale[eq+2]);
					i[3] = mad_f_mul(i[3],p->EqScale[eq+3]);
				}
			}
		}

		//DEBUG_MSG4(-1,T("MP3 %8d %d %d %d"),p->RefTime,p->Frame.sbsample[0][0][0],p->Frame.sbsample[0][4][0],p->Frame.sbsample[0][8][0]);

		if (p->Skip)
		{
			// the first frame may be corrupted
			p->Skip--;
			continue;
		}

		mad_synth_frame(&p->Synth,&p->Frame);

		// handle stereo streams with random mono frames (is this really a good mp3?)

		if (p->OutputFormat.Channels==2 && p->Synth.pcm.channels==1)
		{
			p->Synth.pcm.channels = 2;
			memcpy(p->Synth.pcm.samples[1],p->Synth.pcm.samples[0],
				p->Synth.pcm.length * sizeof(mad_fixed_t));
		}

		if (p->InputFormat.ByteRate != (int)(p->Frame.header.bitrate >> 3))
		{
			p->InputFormat.ByteRate = (p->Frame.header.bitrate >> 3);
			UpdateAdjustTime(p);
		}

		// adjust RefTime with AdjustBytes (now that we know the bitrate)
		if (p->RefTime >= 0)
			p->RefTime -= (p->AdjustBytes * p->AdjustTime) >> 12;

		// output format setup needed?

		if (p->OutputFormat.SampleRate != (int)p->Synth.pcm.samplerate ||
			p->OutputFormat.Channels != p->Synth.pcm.channels)
		{
			if (p->FormatSet)
			{
				p->FormatSet--;
				continue; // probably a bad frame, drop it
			}

			// set new output format
			p->InputFormat.SampleRate = p->OutputFormat.SampleRate = p->Synth.pcm.samplerate;
			p->InputFormat.Channels = p->OutputFormat.Channels = p->Synth.pcm.channels;
			
			if (p->Output.Node)
				p->Output.Node->Set(p->Output.Node,p->Output.No|PACKET_FORMAT,&p->OutputFormat,sizeof(audio));
		}
		p->FormatSet = 16; // reset countdown

		if (p->Eq.Enabled)
		{
			mad_fixed_t Max = 0;
			int ch;
			for (ch=0;ch<p->Synth.pcm.channels;++ch)
			{
				mad_fixed_t* i=p->Synth.pcm.samples[ch];
				mad_fixed_t* ie=i+p->Synth.pcm.length;
				for (;i!=ie;++i)
				{
					mad_fixed_t v = *i;
					if (v<0) v=-v;
					if (v>Max) Max=v;
				}
			}

			if (Max > MAD_F_ONE)
			{
				Max = (mad_fixed_t) (((int64_t)MAD_F_ONE * (int64_t)MAD_F_ONE *15/16) / Max);
				Rescale(p,Max,5);
				
				p->Scale = mad_f_mul(p->Scale,Max);
				UpdateScale(p);
			}
			else
			if (Max < ((MAD_F_ONE*6)/8) && p->Scale < (MAD_F_ONE - MAD_F_ONE/256))
			{
				Max = (MAD_F_ONE/31)*32;
				if (mad_f_mul(p->Scale,Max) > MAD_F_ONE)
					Max = (mad_fixed_t) (((int64_t)MAD_F_ONE * (int64_t)MAD_F_ONE) / p->Scale);
				Rescale(p,Max,10);
			
				p->Scale = mad_f_mul(p->Scale,Max);
				UpdateScale(p);
			}
			
			//DEBUG_MSG1(-1,T("Attenaute %08x"),p->Scale);
		}

		// build output packet
		p->Packet.RefTime = p->RefTime;
		p->Packet.Length = p->Synth.pcm.length * sizeof(mad_fixed_t);
		p->Packet.Data[0] = p->Synth.pcm.samples[0];
		p->Packet.Data[1] = p->Synth.pcm.samples[1];
		p->Total += p->Synth.pcm.length;

		// update output info
		p->RefTime = -1;
		return 1;
	}
}

static int Decode( libmad* p, const packet* Packet )
{
	int Result;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;

	if (p->Pending)
	{
		do
		{
			if (p->Output.Node->Set(p->Output.Node,p->Output.No,
									&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
				return ERR_BUFFER_FULL;
		}
		while (DecodeFrame(p));

		p->Pending = 0;
	}

	// set new reftime
	if (Packet->RefTime >= 0)
	{
		p->RefTime = Packet->RefTime;
		p->AdjustBytes = p->Merge.WritePos - p->Merge.ReadPos;
	}

	// add new packet to mergebuffer
	MergeWrite(&p->Merge,Packet->Data[0],Packet->Length);

	mad_stream_buffer(&p->Stream, p->Merge.Data+p->Merge.ReadPos, p->Merge.WritePos-p->Merge.ReadPos);

	if (DecodeFrame(p))
	{
		Result = ERR_NONE;
		do
		{
			int i = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));
			if (i == ERR_BUFFER_FULL)
			{
				p->Pending = 1;
				//return last result
				break;
			}

			Result = i;
		}
		while (DecodeFrame(p));
	}
	else
		Result = ERR_NEED_MORE_DATA; // if no frame decoded: need more packet

	return Result;
}

static bool_t IsEmpty( libmad* p )
{
	packet Zero;

	if (!p->InputFormat.Format) 
		return 1;

	Zero.CurrTime = -1;
	Zero.DropLevel = p->Packet.DropLevel;
	Zero.Length = 0;
	Zero.Data[0] = NULL;
	Zero.RefTime = -1;

	return Decode(p,&Zero) == ERR_NEED_MORE_DATA;
}

static int Enum( libmad* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,LibMadParams,NULL);
}

static int Get( libmad* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(LIBMAD_ID,int); break;
	case LIBMAD_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case LIBMAD_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case LIBMAD_OUTPUT: GETVALUE(p->Output,pin); break;
	case LIBMAD_COMMENTS: GETVALUE(p->Comments,pin); break;
	case LIBMAD_EQ: GETVALUE(p->Eq,equalizer); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_EMPTY:GETVALUE(IsEmpty(p),bool_t); break;
	}
	return Result;
}

static int Set( libmad* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case LIBMAD_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case LIBMAD_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case LIBMAD_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case LIBMAD_COMMENTS: SETVALUE(p->Comments,pin,ERR_NONE); break;
	case LIBMAD_EQ: SETVALUE(p->Eq,equalizer,UpdateEq(p)); break;
	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_EMPTY:

		// reset
		MergeClear(&p->Merge);
		p->Pending = 0;
		p->Skip = 1;

		mad_frame_finish(&p->Frame);
		mad_synth_finish(&p->Synth);
		mad_stream_finish(&p->Stream);

		mad_frame_init(&p->Frame);
		mad_synth_init(&p->Synth);
		mad_stream_init(&p->Stream);

		break;
	}
	return Result;
}

libmad LibMad;

extern struct fixedfloat {
  unsigned long mantissa : 27;
  unsigned long exponent :  5;
} rq_table[8207];

void LibMad_Init()
{
	// init rq_table
	int x;
	struct fixedfloat* p=rq_table;
	for (x=0;x<8207;++x,++p)
	{
		int exp;
		double v = frexp(pow(x,4./3.),&exp);
		if (exp) ++exp;
		p->exponent = (unsigned short)exp;
		p->mantissa = (int)(0x10000000 * (v*0.5));
	}

	NodeFill(&LibMad.VMT,sizeof(LibMad),Enum,Get,Set);
	NodeRegister(&LibMad.VMT,PRI_DEFAULT);
}

void LibMad_Done()
{
	NodeUnRegister(&LibMad.VMT);
}



