/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: mpeg.c 11 2004-07-15 13:46:20Z picard $
 *
 * based on MPEG1/2 mux/demux
 * Copyright (c) 2000, 2001, 2002 Fabrice Bellard.
 *
 * modified for BetaPlayer by Gabor Kovacs Copyright (c) 2004 
 *
 ****************************************************************************/

#include "stdafx.h"

#define AUDIO_ID					0xC0
#define VIDEO_ID					0xE0

typedef format_base mpeg;

static int Init(mpeg* p)
{
	p->StartTime = 0;
	return ERR_NONE;
}

static int Probe(mpeg* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	const uint8_t* Data = (const uint8_t*)Ptr;
	int Code=-1;
    int i;

	for (i=0;i<20 && i<Length;i++) 
	{
		Code = (Code << 8) | Data[i];
		if ((Code & ~0xFF) == 0x100) 
		{
			if (Code == PACK_START_CODE ||
				Code == SYSTEM_HEADER_START_CODE ||
				(Code >= 0x1E0 && Code <= 0x1Ef) ||
				(Code >= 0x1C0 && Code <= 0x1Df) ||
				Code == PRIVATE_STREAM_2 ||
				Code == PROGRAM_STREAM_MAP ||
				Code == PRIVATE_STREAM_1 ||
				Code == PADDING_STREAM)
				return ERR_NONE;

			break;
		}
	}

	if (CheckExts(URL,p->Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static bool_t GetStartCode(mpeg* p)
{
	int Size;
	int StartCode = -1;

    for (Size=0;Size<CHUNKSIZE;++Size)
	{
		bool_t Found = StartCode == 0x000001;
        int v = Format_Read8(p);
		if (v<0)
			break;

		StartCode = ((StartCode << 8) | v) & 0xFFFFFF;

		if (Found)
			return StartCode;
    }
    return -1;
}

static int64_t GetPTS(mpeg* p,int i)
{
    int64_t PTS;

    if (i < 0)
        i = Format_Read8(p);

    PTS = (int64_t)((i >> 1) & 0x07) << 30;
    PTS |= (Format_ReadBE16(p) >> 1) << 15;
    PTS |= (Format_ReadBE16(p) >> 1);

    return PTS;
}

static int ReadPacket(mpeg* p, format_packet* Packet)
{
	int StartCode = GetStartCode(p);
	int Len,i;
	int PCM = 0;
	int64_t PTS;

	if (StartCode == -1 ||
		StartCode == PACK_START_CODE ||
		StartCode == SYSTEM_HEADER_START_CODE)
		return ERR_NONE;

    if (StartCode == PADDING_STREAM ||
        StartCode == PRIVATE_STREAM_2) 
	{
		Len = Format_ReadBE16(p);
		Format_Skip(p,Len);
		return ERR_NONE;
    }

    if (!((StartCode >= 0x1C0 && StartCode <= 0x1DF) ||
          (StartCode >= 0x1E0 && StartCode <= 0x1EF) ||
          (StartCode == 0x1BD)))
        return ERR_NONE;

    Len = Format_ReadBE16(p);
    PTS = -1;
 
    do
	{
        i = Format_Read8(p);
        --Len;
	}
	while (i == 0xFF);
   
    if ((i & 0xC0) == 0x40) 
	{
		Format_Read8(p);
        i = Format_Read8(p);
        Len -= 2;
    }
    if ((i & 0xF0) == 0x20) 
	{
        PTS = GetPTS(p,i);
        Len -= 4;
    } 
	else 
	if ((i & 0xF0) == 0x30) 
	{
        GetPTS(p,i);
        PTS = GetPTS(p,-1);
        Len -= 9;
    } 
	else 
	if ((i & 0xC0) == 0x80) 
	{
		int HeadLen;

		if ((i & 0x30) != 0) 
		{
			NodeError((node*)p,MPEG_ID,MPEG_ENCRYPTED);
	        return ERR_END_OF_FILE;
		}

        i = Format_Read8(p);
        HeadLen = Format_Read8(p);
        Len -= 2;

        if (HeadLen > Len)
			return ERR_NONE;

        if ((i & 0xC0) == 0x80) 
		{
            PTS = GetPTS(p,-1);
            HeadLen -= 5;
            Len -= 5;
        }
		else
		if ((i & 0xC0) == 0xC0) 
		{
            GetPTS(p,-1);
            PTS = GetPTS(p,-1);
            HeadLen -= 10;
            Len -= 10;
        }

		Format_Skip(p,HeadLen);
        Len -= HeadLen;
    }

    if (StartCode == 0x1BD)
	{
        StartCode = Format_Read8(p);
        Len--;
        if (StartCode >= 0x80 && StartCode <= 0xBF) 
		{
			Format_Skip(p,3);
            Len -= 3;
        }
		if (StartCode >= 0xA0 && StartCode <= 0xBF) 
		{
			Format_Read8(p);
			PCM = Format_Read8(p);
			Format_Read8(p); 
			Len -= 3;
		}
    }

	if (PTS >= 0)
	{
		Packet->RefTime = (tick_t)((PTS * TICKSPERSEC) / 90000) - p->StartTime;
		if (Packet->RefTime < 0)
			Packet->RefTime = 0;
	}
	else
		Packet->RefTime = -1;

	for (i=0;i<p->StreamCount;++i)
		if (p->Streams[i]->Id == StartCode)
			break;

	if (i==p->StreamCount)
	{
		format_stream* s = NULL;

		if (!p->StartTime)
		{
			p->StartTime = Packet->RefTime;
			Packet->RefTime = 0;
		}

	    if (StartCode >= 0x1E0 && StartCode <= 0x1EF) 
		{
			s = Format_AddStream(p,sizeof(format_stream));
			if (s)
			{
				video* Video = (video*) Format_StreamFormat(s,PACKET_VIDEO,sizeof(video));
				s->Id = StartCode;
				if (Video)
				{
					Video->Pixel.Flags = PF_FOURCC | PF_FRAGMENTED;
					Video->Pixel.FourCC = FOURCC('M','P','E','G');
				}
				Format_PrepairStream(p,s);
			}
		}
		else 
		if (StartCode >= 0x1C0 && StartCode <= 0x1DF) 
		{
			s = Format_AddStream(p,sizeof(format_stream));
			if (s)
			{
				audio* Audio = (audio*) Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
				s->Id = StartCode;
				if (Audio)
					Audio->Format = AUDIOFMT_MP2;
				Format_PrepairStream(p,s);
			}
		}
		else 
		if (StartCode >= 0xA0 && StartCode <= 0xBF) 
		{
			s = Format_AddStream(p,sizeof(format_stream));
			if (s)
			{
				static const int Freq[4] = { 48000, 96000, 44100, 32000 };
				audio* Audio = (audio*) Format_StreamFormat(s,PACKET_AUDIO,sizeof(audio));
				s->Id = StartCode;
				if (Audio)
				{
					Audio->Format = AUDIOFMT_PCM;
					Audio->SampleRate = Freq[(PCM >> 4) & 3];
					Audio->Channels = 1 + (PCM & 7);
					Audio->Bits = 16;
					Audio->FracBits = 15;
					Audio->ByteRate = Audio->SampleRate * Audio->Channels * 2;
				}
				Format_PrepairStream(p,s);
			}
		} 
		
		if (!s)
		{
			Format_Skip(p,Len);
			return ERR_NONE;
		}
	}

	Packet->Data = Format_ReadAsRef(p,Len);
	Packet->Stream = p->Streams[i];

	//if (i==0) DEBUG_MSG3(-1,T("MPEG %d %d %d"),Packet->RefTime,(int)PTS,(int)DTS);
	return ERR_NONE;
}

static mpeg MPEG;

void MPEG_Init()
{
	MPEG = FormatBase;
	MPEG.VMT.Probe = (fmtprobe)Probe;
	MPEG.Init = (fmtfunc)Init;
	MPEG.Seek = (fmtseek)Format_SeekByPacket;
	MPEG.ReadPacket = (fmtreadpacket)ReadPacket;
	MPEG.Id = MPEG_ID;
	MPEG.Exts = T("mpeg:V;mpg:V;dat:V;mpv:V");
	MPEG.TimeStamps = 1;

	NodeRegister((node*)&MPEG.VMT,PRI_DEFAULT);
}

void MPEG_Done()
{
	NodeUnRegister((node*)&MPEG.VMT);
}

