/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: ogg.c 144 2004-12-12 15:28:39Z picard $
 *
 * BetaPlayer Vorbis plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "vorbis.h"
#include "tremor/ogg.h"
#include "tremor/ivorbiscodec.h"

// OGG and OGM file format

#define PACKET_TYPE_HEADER		0x01
#define PACKET_TYPE_COMMENT		0x03
#define PACKET_TYPE_CODEBOOK	0x05
#define PACKET_TYPE_BITS		0x07

#define	PACKET_IS_SYNCPOINT		0x08
#define PACKET_LEN_BITS01		0xc0
#define PACKET_LEN_BITS2		0x02

const char MT_Video[]	= "video";
const char MT_Audio[]	= "audio";
const char MT_Text[]	= "text";

typedef struct oggstream
{
	format_stream p;

	bool_t Invalid; // not vorbis and not ogm
	bool_t Vorbis;	// sending direct oggpackets (not the merged binary data)
	int PacketNo;

	int64_t MediaTime; // in units
	int64_t MediaRateNum; // in units/tick
	int64_t MediaRateDen;
	int DefaultLen;

	bool_t NeedMorePage;

	ogg_stream_state* OggStream;
	ogg_packet       OggPacket;

	vorbis_info      Info;
	vorbis_comment   Comment;

} oggstream;

typedef struct ogg
{
	format_base p;

	ogg_sync_state*	OggSync;
	ogg_page        OggPage;

} ogg;

typedef struct ogm_header_video
{
	int32_t	width;
	int32_t	height;

} ogm_header_video;
	
typedef struct ogm_header_audio
{
	int16_t	channels;
	int16_t	blockalign;
	int32_t	avgbytespersec;
} ogm_header_audio;

typedef struct ogm_header
{
	char	streamtype[8];
	char	subtype[4];

	int32_t	size;

	int64_t	time_unit;
	int64_t	samples_per_unit;
	int32_t default_len;

	int32_t buffersize;
	int16_t	bits_per_sample;

	union
	{
		ogm_header_video video;
		ogm_header_audio audio;
	};

} ogm_header;

static int Probe(ogg* p, const void* Ptr, int Length, const tchar_t* URL, const tchar_t* ContentType )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

	if (tcsicmp(ContentType,T("application/ogg"))==0)
		return ERR_NONE;

    if (Data[0]=='O' &&	Data[1]=='g' &&
		Data[2]=='g' &&	Data[3]=='S')
		return ERR_NONE;

	if (CheckExts(URL,p->p.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

static void FreeStream(ogg* p,oggstream* Stream)
{
	vorbis_comment_clear(&Stream->Comment);
	vorbis_info_clear(&Stream->Info);

	ogg_packet_release(&Stream->OggPacket);
	if (Stream->OggStream)
	{
		ogg_stream_destroy(Stream->OggStream);
		Stream->OggStream = NULL;
	}
}

static void Done(ogg* p)
{
	ogg_page_release(&p->OggPage);
	if (p->OggSync)
	{
		ogg_sync_destroy(p->OggSync);
		p->OggSync = NULL;
	}
}

static int Init(ogg* p)
{
	p->OggSync = ogg_sync_create();
	if (!p->OggSync)
		return ERR_OUT_OF_MEMORY;

	return ERR_NONE;
}

static void AfterSeek(ogg* p)
{
	ogg_sync_reset(p->OggSync);
}

static void ReleaseStream(ogg* p,oggstream* Stream)
{
	ogg_stream_reset_serialno(Stream->OggStream,Stream->p.Id);
	Stream->NeedMorePage = 1;
	Stream->MediaTime = p->p.FilePos>0 && Stream->PacketNo>3 ? -1:0;
}

int ReadPacketTime( ogg* p, format_packet* Packet )
{
	// only called by CalcDuration

	format_buffer* Buffer;
	int Bytes;
	void* Ptr;
	ogg_sync_state* Sync = ogg_sync_create();
	ogg_page Page;

	if (!Sync)
		return ERR_OUT_OF_MEMORY;

	while ((Buffer = Format_BufferRemove(&p->p))!=NULL)
	{
		Ptr = ogg_sync_bufferin(Sync,Buffer->Length);
		if (Ptr)
		{
			memcpy(Ptr,Buffer->Buffer,Buffer->Length);
			ogg_sync_wrote(Sync,Buffer->Length);
		}
		Format_BufferRelease(&p->p,Buffer);
	}

	memset(&Page,0,sizeof(Page));

	while ((Bytes = ogg_sync_pageseek(Sync,&Page)) != 0)
	{
		if (Bytes > 0)
		{
			int64_t MediaTime = ogg_page_granulepos(&Page);
			int	Id = ogg_page_serialno(&Page);
			int i;

			if (MediaTime != -1)
				for (i=0;i<p->p.StreamCount;++i)
					if (p->p.Streams[i]->Id == Id)
					{
						oggstream* s = (oggstream*) p->p.Streams[i];

						if (s->MediaRateNum)
						{
							tick_t RefTime = (tick_t)(MediaTime * s->MediaRateDen / s->MediaRateNum);
							
							if (RefTime > Packet->RefTime)
							{
								Packet->Stream = &s->p;
								Packet->RefTime = RefTime;
							}
						}
						break;
					}
		}
	}

	ogg_page_release(&Page);
	ogg_sync_destroy(Sync);

	return ERR_NONE;
}

static void FrameRate( fraction* f, int64_t Num, int64_t Den )
{
	while (Num > MAX_INT || Den > MAX_INT)
	{
		Num >>= 1;
		Den >>= 1;
	}
	f->Num = (int)Num;
	f->Den = (int)Den;
}

bool_t OGMHeader(ogg* p, oggstream* s, char* Data, int Length )
{
	if ((*Data & PACKET_TYPE_BITS) != PACKET_TYPE_HEADER)
		return 0;

	if (strncmp(Data+1, "Direct Show Samples embedded in Ogg", 35) == 0)
	{
		// old header

		if (INT32LE(*(int32_t*)(Data+96)) == 0x05589F80 &&
			Format_StreamFormat(&s->p,PACKET_VIDEO,sizeof(video)))
		{
			video* Format = (video*) s->p.Format;

			Format->Width = INT32LE(*(int32_t*)(Data+176));
			Format->Height = INT32LE(*(int32_t*)(Data+180));
			Format->Pixel.Flags = PF_FOURCC;
			Format->Pixel.FourCC = INT32LE(*(int32_t*)(Data+68));
			Format->Pixel.BitCount = INT16LE(*(int16_t*)(Data+182));
			DefaultPitch(Format);

			s->MediaRateDen = INT64LE(*(int32_t*)(Data+164))*TICKSPERSEC;
			s->MediaRateNum = 10000000;
			s->DefaultLen = 1;

			FrameRate( &Format->FrameRate, s->MediaRateNum,s->MediaRateDen/TICKSPERSEC );
			return 1;
		}

		if (INT32LE(*(int32_t*)(Data+96)) == 0x05589F81 &&
			Format_StreamFormat(&s->p,PACKET_AUDIO,sizeof(audio)))
		{
			int i;
			audio* Format = (audio*) s->p.Format;

			Format->Format = INT16LE(*(int16_t*)(Data+124));
			Format->Channels = INT16LE(*(int16_t*)(Data+126));
			Format->BlockAlign = INT16LE(*(int16_t*)(Data+136));
			Format->Bits = INT16LE(*(int16_t*)(Data+138));
			Format->SampleRate = INT32LE(*(int32_t*)(Data+128));
			Format->ByteRate = INT32LE(*(int32_t*)(p+132));

			i = INT16LE(*(int16_t*)(Data+136));	// extrasize
			if (i && Format_StreamExtra(&s->p,i))
				memcpy(s->p.Extra,Data+142,i); // extradata

			s->MediaRateDen = TICKSPERSEC;
			s->MediaRateNum = INT32LE(*(int32_t*)(Data+128));
			s->DefaultLen = 1;
			return 1;
		}
	}
	else
	if (Length >= sizeof(ogm_header)+1)
	{
		ogm_header Head;
		memcpy(&Head,Data+1,sizeof(Head));

		// new header
		if (strncmp(Head.streamtype, MT_Video, strlen(MT_Video)) == 0 && 
			Format_StreamFormat(&s->p,PACKET_VIDEO,sizeof(video)))
		{
			video* Format = (video*) s->p.Format;

			Format->Width = INT32LE(Head.video.width);
			Format->Height = INT32LE(Head.video.height);
			Format->Pixel.Flags = PF_FOURCC;
			Format->Pixel.FourCC = INT32LE(*(int32_t*)Head.subtype);
			Format->Pixel.BitCount = INT16LE(Head.bits_per_sample);
			DefaultPitch(Format);

			s->MediaRateDen = INT64LE(Head.time_unit)*TICKSPERSEC;
			s->MediaRateNum = INT64LE(Head.samples_per_unit) * 10000000;
			s->DefaultLen = INT32LE(Head.default_len);

			FrameRate( &Format->FrameRate, s->MediaRateNum,s->MediaRateDen/TICKSPERSEC );
			return 1;
		}
		
		if (strncmp(Head.streamtype, MT_Audio, strlen(MT_Audio)) == 0 &&
			Format_StreamFormat(&s->p,PACKET_AUDIO,sizeof(audio)))
		{
			int i;
			audio* Format = (audio*) s->p.Format;

			Format->Format = 0;
			for (i=0;i<4;++i)
				Format->Format = Format->Format*16+
					toupper(Head.subtype[i])-(Head.subtype[i]>'9'?'A':'0');
				
			Format->Channels = INT16LE(Head.audio.channels);
			Format->Bits = INT16LE(Head.bits_per_sample);
			Format->BlockAlign = INT16LE(Head.audio.blockalign);
			Format->ByteRate = INT32LE(Head.audio.avgbytespersec);
			Format->SampleRate = (int)INT64LE(Head.samples_per_unit);

			s->MediaRateDen = INT64LE(Head.time_unit)*TICKSPERSEC;
			s->MediaRateNum = INT64LE(Head.samples_per_unit) * 10000000;
			s->DefaultLen = INT32LE(Head.default_len);

			i = Length - sizeof(ogm_header)+1;
			if (i && Format_StreamExtra(&s->p,i))
				memcpy(s->p.Extra,Data+1+sizeof(ogm_header),i);

			return 1;
		}

		if (strncmp(Data+1, MT_Text,  strlen(MT_Text)) == 0)
		{
		}
	}

	return 0;
}

static void SendComments(oggstream* Stream)
{
	tchar_t	s[256];
	if (Stream->p.Comment.Node)
	{
		int No;
		for (No=0;No<Stream->Comment.comments;++No)
		{
			StrToTcs(s,Stream->Comment.user_comments[No],256);
			Stream->p.Comment.Node->Set(Stream->p.Comment.Node,Stream->p.Comment.No,s,sizeof(s));
		}
	}
}

bool_t VorbisHeader(ogg* p,oggstream* s)
{
	bool_t			Result;

	Result = vorbis_synthesis_headerin(&s->Info,&s->Comment,&s->OggPacket) >= 0;

	if (Result && !Format_StreamFormat(&s->p,PACKET_AUDIO,sizeof(audio)))
		Result = 0;

	if (Result)
	{
		audio* Format = (audio*) s->p.Format;

		s->Vorbis = 1;

		Format->Format = AUDIOFMT_VORBIS_INTERNAL;
		Format->Channels = s->Info.channels;
		Format->SampleRate = s->Info.rate;
		Format->ByteRate = s->Info.bitrate_nominal >> 3;
		Format->Extra = NULL; // set after all the three header packets are processed
		Format->ExtraLength = -1;

		s->MediaRateDen = TICKSPERSEC;
		s->MediaRateNum = s->Info.rate;
	}

	return Result;
}

int FillQueue(ogg* p)
{
	for (;;)
	{
		int Bytes = ogg_sync_pageseek(p->OggSync,&p->OggPage);

		if (Bytes == 0) // need more data
		{
			void* Ptr;
			format_buffer* Buffer;

			if (!p->p.BufferAvailable)
				return ERR_NEED_MORE_DATA;

			Buffer = Format_BufferRemove(&p->p);
			if (!Buffer)
				return ERR_NEED_MORE_DATA;
			
			if (Buffer->Length)
			{
				Ptr = ogg_sync_bufferin(p->OggSync,Buffer->Length);
				if (!Ptr)
				{
					Format_BufferRelease(&p->p,Buffer);
					return ERR_OUT_OF_MEMORY;
				}

				memcpy(Ptr,Buffer->Buffer,Buffer->Length);
				ogg_sync_wrote(p->OggSync,Buffer->Length);
			}

			Format_BufferRelease(&p->p,Buffer);
		}
		else		
		if (Bytes < 0)
			p->p.FilePos -= Bytes;
		else
		if (Bytes > 0)
		{
			int i;
			oggstream* s;
			int Id;
			
			p->p.FilePos += Bytes;

			Id = ogg_page_serialno(&p->OggPage);
		
			DEBUG_MSG4(DEBUG_FORMAT,T("OGG Page id:%d size:%d gran:%d filepos:%d"),Id,p->OggPage.body_len,(int)ogg_page_granulepos(&p->OggPage),p->p.FilePos - Bytes);

			for (i=0;i<p->p.StreamCount;++i)
				if (p->p.Streams[i]->Id == Id)
					break;

			if (i==p->p.StreamCount)
			{
				s = (oggstream*) Format_AddStream(&p->p,sizeof(oggstream));
				if (!s)	continue;

				// init stream
				s->p.Id = Id;
				s->OggStream = ogg_stream_create(Id);
				s->NeedMorePage = 1;
				s->MediaTime = 0;
				s->Invalid = 0;
				s->Vorbis = 0;
				s->PacketNo = 0;

				vorbis_info_init(&s->Info);
				vorbis_comment_init(&s->Comment);
			}

			s = (oggstream*) p->p.Streams[i];

			if (s->Invalid) // drop invalid streams
				continue;

			if (s->PacketNo>=3)
			{
				if (!s->p.Pin.Node) // drop unused streams
					continue;

				if (p->p.InSeek)
				{
					// reftime needed for SeekByPacket
					if ((s->MediaTime = ogg_page_granulepos(&p->OggPage)) != -1)
					{
						s->p.LastTime = (tick_t)(s->MediaTime * s->MediaRateDen / s->MediaRateNum);
						if (s->p.FormatType == PACKET_AUDIO)
						{
							s->p.LastTime += p->p.AVOffset;
							if (s->p.LastTime < 0)
								s->p.LastTime = 0;
						}
					}
				}
			}

			// add page to stream
			if (ogg_stream_pagein(s->OggStream,&p->OggPage) >= 0)
			{
				if (s->PacketNo<3) // header packet needed?
				{
					i = ogg_stream_packetout(s->OggStream,&s->OggPacket);
					if (i == 0) // stream needs more pages
						continue;
					
					if (++s->PacketNo==1) // first packet?
					{
						ogg_reference* Ref;

						if (i < 0)
						{
							// first header packet is a must have
							s->Invalid = 1;
							continue;
						}

						MergeClear(&s->p.Merge);
						for (Ref=s->OggPacket.packet;Ref;Ref=Ref->next)
							MergeWrite(&s->p.Merge,Ref->buffer->data + Ref->begin, Ref->length);

						if (OGMHeader(p,s,(char*)s->p.Merge.Data,s->p.Merge.WritePos))
							s->PacketNo = 3; // no more headers
						else
						if (!VorbisHeader(p,s))
						{
							s->Invalid = 1;
							continue;
						}

						while (s->MediaRateNum > (1<<30))
						{
							s->MediaRateDen >>= 1;
							s->MediaRateNum >>= 1;
						}

						Format_PrepairStream(&p->p,&s->p);
						continue;
					}
					else
					{
						assert(s->Vorbis);

						// error in second or third header packet will not cause fatal error
						vorbis_synthesis_headerin(&s->Info,&s->Comment,&s->OggPacket); 

						if (s->PacketNo == 3)
						{
							audio* Format = (audio*) s->p.Format;

							// got the three header packets: reinit codec with vorbis_info
							Format->Extra = &s->Info;
							if (s->p.Pin.Node)
								s->p.Pin.Node->Set(s->p.Pin.Node,s->p.Pin.No|PACKET_FORMAT,Format,sizeof(audio));

							SendComments(s);
						}
						continue;
					}
				}

				s->NeedMorePage = 0;
				break;
			}
		}
	}
	return ERR_NONE;
}

int Process(ogg* p,oggstream* Stream)
{
	int Result;
	int No;
	int Burst = 1;

	if (Stream->Invalid)
		return ERR_NONE;
	
	if (Stream->p.Pending)
	{
		Result = Format_Send(&p->p,&Stream->p);

		if (Result == ERR_BUFFER_FULL || Result == ERR_SYNCED)
			return Result;
	}

	// process a limited number of packets at once (other streams need cpu time too)

	Burst = Stream->p.PacketBurst;

	for (No=0;No<Burst;++No)
	{
		while (Stream->NeedMorePage)
		{
			Result = FillQueue(p);

			if (Result != ERR_NONE || (p->p.Bench && Stream->NeedMorePage))
				return Result;
		}

		// sync: only SyncStream is processed
		if (p->p.SyncMode && Stream != (oggstream*)p->p.SyncStream)
			return ERR_NEED_MORE_DATA;

		Result = ogg_stream_packetout(Stream->OggStream,&Stream->OggPacket);
		if (Result == 0)
		{
			// stream needs more pages
			Stream->NeedMorePage = 1;
			if (p->p.Bench) break; // benchmark mode: no additional pages are added in this loop
			--No;
			continue;
		}
		if (Result < 0) // packet hole or span
			continue;
	
		DEBUG_MSG3(DEBUG_FORMAT,T("OGG Packet id:%d size:%d gran:%d"),Stream->p.Id,Stream->OggPacket.bytes,(int)Stream->OggPacket.granulepos);

		if (!Stream->Vorbis && Stream->PacketNo==3)
		{
			// non vorbis: second packet can be a comment (after first header)
			++Stream->PacketNo;

			Stream->Info.rate=1;
			if (vorbis_synthesis_headerin(&Stream->Info,&Stream->Comment,&Stream->OggPacket) >= 0)
			{
				SendComments(Stream);
				continue;
			}
		}

		// build output packet

		if (Stream->OggPacket.granulepos >= 0)
			Stream->MediaTime = Stream->OggPacket.granulepos;

		if (Stream->MediaTime >= 0)
		{
			tick_t t = (tick_t)(Stream->MediaTime * Stream->MediaRateDen / Stream->MediaRateNum);
			if (Stream->p.FormatType == PACKET_AUDIO)
			{
				t += p->p.AVOffset;
				if (t<0) t=0;
			}
			Stream->p.LastTime = Stream->p.Packet.RefTime = t;
		}
		else
		{
			if (Stream->p.LastTime < 0)
				continue;
			Stream->p.Packet.RefTime = -1;
		}

		if (!Stream->Vorbis)
		{
			ogg_reference* Ref;
			uint8_t* Data;
			int Len;
			int PacketTime;

			// merge buffers

			MergeClear(&Stream->p.Merge);
			for (Ref=Stream->OggPacket.packet;Ref;Ref=Ref->next)
				MergeWrite(&Stream->p.Merge,Ref->buffer->data + Ref->begin, Ref->length);
		
			// process OGM information

			Data = (uint8_t*)Stream->p.Merge.Data;

			if (*Data & PACKET_TYPE_HEADER)
				continue; // not data packet

			Len = (*Data & PACKET_LEN_BITS01) >> 6;
			Len |= (*Data & PACKET_LEN_BITS2) << 1;

			Stream->p.Packet.Key = (*Data & PACKET_IS_SYNCPOINT) != 0;
			Stream->p.Packet.Data[0] = Data + 1 + Len;
			Stream->p.Packet.Length = Stream->p.Merge.WritePos - 1 - Len;

			if (!Len)
				PacketTime = Stream->DefaultLen;
			else
				for (PacketTime=0;Len;--Len)
				{
					PacketTime <<= 8;
					PacketTime |= Data[Len];
				}

			Stream->MediaTime += PacketTime;
		}
		else
		{
			// send oggpacket directly 

			Stream->p.Packet.Data[0] = &Stream->OggPacket;
			Stream->p.Packet.Length = sizeof(Stream->OggPacket);
			Stream->MediaTime = -1;
		}
		Stream->p.Pending = 1;

		Result = Format_Send(&p->p,&Stream->p);

		if (Result == ERR_BUFFER_FULL || Result == ERR_SYNCED)
			break;

		if (Stream->p.Packet.DropLevel==2)
			Burst = Stream->p.PacketBurst*2; // try to catch up
	}

	// buffer full: there was possibly some processing so don't allow sleeping
	// need data: burst count exceeded, but it doesn't mean there is no data left in buffers
	if (Result == ERR_BUFFER_FULL || Result == ERR_NEED_MORE_DATA)
		Result = ERR_NONE;

	return Result;
}

static ogg OGG;

void OGG_Init()
{
	OGG.p = FormatBase;
	OGG.p.VMT.Probe = Probe;
	OGG.p.Init = Init;
	OGG.p.Done = Done;
	OGG.p.Seek = Format_SeekByPacket;
	OGG.p.FillQueue = FillQueue;
	OGG.p.ReadPacket = ReadPacketTime;
	OGG.p.ReleaseStream = ReleaseStream;
	OGG.p.AfterSeek = AfterSeek;
	OGG.p.FreeStream = FreeStream;
	OGG.p.Process = Process;
	OGG.p.Sended = NULL;
	OGG.p.Id = OGG_ID;
	OGG.p.Exts = T("ogg:A;ogm:V");
	OGG.p.TimeStamps = 1;

	NodeRegister((node*)&OGG.p.VMT,PRI_DEFAULT);
}

void OGG_Done()
{
	NodeUnRegister((node*)&OGG.p.VMT);
}
