/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: oggembedded.c 131 2004-12-04 20:36:04Z picard $
 *
 * BetaPlayer Vorbis plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "vorbis.h"
#include "tremor/ogg.h"
#include "tremor/ivorbiscodec.h"

// Ogg file format embedded inside an audio stream

#define OGGEMBEDDED_INPUT		0x100
#define OGGEMBEDDED_OUTPUT		0x101
#define OGGEMBEDDED_COMMENT		0x102

typedef struct oggembedded
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;
	pin Output;
	pin Comments;

	int OggStreamNo;
	ogg_sync_state*	 OggSync;
	ogg_stream_state* OggStream;
	ogg_page         OggPage;
	ogg_packet       OggPacket;

	vorbis_info      Info;
	vorbis_comment   Comment;

	// output data
	packet Packet;
	tick_t RefTime;
	bool_t Reset;
	bool_t Pending;
	bool_t NeedMorePage;
	int Bytes;   // bytes in Sync buffer
	int PageNo;

} oggembedded;

const datadef OGGEmbeddedParams[] = 
{
	{ OGGEMBEDDED_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ OGGEMBEDDED_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },
	{ OGGEMBEDDED_COMMENT,	TYPE_PIN, DF_OUTPUT, PACKET_COMMENT },

	DATADEF_END
};

static int SetFormat( oggembedded* p, const audio* Format )
{
	if (p->InputFormat.Format)
	{
		memset(&p->InputFormat,0,sizeof(p->InputFormat));

		vorbis_comment_clear(&p->Comment);
		vorbis_info_clear(&p->Info);

		ogg_packet_release(&p->OggPacket);
		ogg_page_release(&p->OggPage);

		if (p->OggStream)
			ogg_stream_destroy(p->OggStream);

		if (p->OggSync)
			ogg_sync_destroy(p->OggSync);

		p->OggSync = NULL;
		p->OggStream = NULL;
	}

	if (Format)
	{
		// accept only vorbis
		if (Format->Format != AUDIOFMT_VORBIS_MODE1 && 
			Format->Format != AUDIOFMT_VORBIS_MODE1P && 
			Format->Format != AUDIOFMT_VORBIS_MODE2 && 
			Format->Format != AUDIOFMT_VORBIS_MODE2P && 
			Format->Format != AUDIOFMT_VORBIS_MODE3 && 
			Format->Format != AUDIOFMT_VORBIS_MODE3P)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		memset(&p->OutputFormat,0,sizeof(p->OutputFormat));
		p->OutputFormat.Format = AUDIOFMT_VORBIS_INTERNAL;

		p->Packet.Data[0] = &p->OggPacket;
		p->Packet.Length = sizeof(p->OggPacket);
		p->Pending = 0;
		p->RefTime = -1;
		p->Bytes = 0;
		p->OggStreamNo = 0;
		p->NeedMorePage = 1;
		p->PageNo = 0;

		p->OggSync = ogg_sync_create();

		vorbis_info_init(&p->Info);
		vorbis_comment_init(&p->Comment);
	}

	return ERR_NONE;
}

bool_t DecodePacket( oggembedded* p )
{
	for (;;)
	{
		while (p->NeedMorePage)
		{
			int StreamNo;
			do
			{
				int Bytes = ogg_sync_pageseek(p->OggSync,&p->OggPage);
				if (Bytes == 0)
					return 0; // missing data

				if (Bytes < 0) 
				{
					// corrupt data
					p->Bytes += Bytes; //sub skipped bytes 
					StreamNo = -1;
				}
				else
				{
					p->Bytes -= Bytes;
					StreamNo = ogg_page_serialno(&p->OggPage);
					if (!p->OggStream)
					{
						//add stream
						p->OggStream = ogg_stream_create(StreamNo);
						p->OggStreamNo = StreamNo;
					}
				}
			}
			while (StreamNo != p->OggStreamNo);

			if (ogg_stream_pagein(p->OggStream,&p->OggPage) >= 0)
				p->NeedMorePage = 0;
		}

		while (!p->NeedMorePage)
		{
			int Result = ogg_stream_packetout(p->OggStream,&p->OggPacket);

			if (Result == 0)
				p->NeedMorePage = 1; // set need more page
			else
			if (Result > 0) 
			{
				// valid packet
				if (p->PageNo < 3) // need 3 header packets
				{
					// first page is vital
					if (vorbis_synthesis_headerin(&p->Info,&p->Comment,&p->OggPacket)>=0 || p->PageNo>0)
					{
						if (++p->PageNo == 3)
						{
							tchar_t Comment[256];
							int No;

							if (p->Comments.Node)
								for (No=0;No<p->Comment.comments;++No)
								{
									StrToTcs(Comment,p->Comment.user_comments[No],256);
									p->Comments.Node->Set(p->Comments.Node,p->Comments.No,Comment,sizeof(Comment));
								}

							p->InputFormat.ByteRate = p->OutputFormat.ByteRate = p->Info.bitrate_nominal >> 3;
							p->InputFormat.SampleRate = p->OutputFormat.SampleRate = p->Info.rate;
							p->InputFormat.Channels = p->OutputFormat.Channels = p->Info.channels;

							p->OutputFormat.Extra = &p->Info;
							p->OutputFormat.ExtraLength = -1;

							if (p->Output.Node)
								p->Output.Node->Set(p->Output.Node,p->Output.No|PACKET_FORMAT,&p->OutputFormat,sizeof(audio));
						}
					}
				}
				else
				{
					p->Packet.RefTime = p->RefTime;

					if (p->RefTime >= 0)
						if (p->OutputFormat.ByteRate)
							p->RefTime += Scale(p->OggPacket.bytes,TICKSPERSEC,p->OutputFormat.ByteRate);
						else
							p->RefTime = -1;
					return 1; 
				}
			}
		}
	}
	return 0;
}

static int Decode( oggembedded* p, const packet* Packet )
{
	void* Ptr;
	int Result;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;

	if (p->Pending)
	{
		do
		{
			if (p->Output.Node->Set(p->Output.Node,p->Output.No,
									&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
				return ERR_BUFFER_FULL;
		}
		while (DecodePacket(p));

		p->Pending = 0;
	}

	// set new reftime (rewind with bytes already buffered)
	if (Packet->RefTime >= 0)
		p->RefTime = Packet->RefTime - Scale(p->Bytes,TICKSPERSEC,p->OutputFormat.ByteRate);

	// add new packet to oggsync
	Ptr = ogg_sync_bufferin(p->OggSync,Packet->Length);
	if (Ptr)
	{
		p->Bytes += Packet->Length;
		memcpy(Ptr,Packet->Data[0],Packet->Length);
		ogg_sync_wrote(p->OggSync,Packet->Length);
	}

	p->NeedMorePage = 1;

	if (DecodePacket(p))
	{
		Result = ERR_NONE;
		do
		{
			int i = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));

			if (i == ERR_BUFFER_FULL)
			{
				p->Pending = 1;
				//return last result
				break;
			}

			Result = i;
		}
		while (DecodePacket(p));
	}
	else
		Result = ERR_NEED_MORE_DATA; // if no packet decoded

	return Result;
}

static int Enum( oggembedded* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,OGGEmbeddedParams,NULL);
}

static int Get( oggembedded* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(OGGEMBEDDED_ID,int); break;
	case OGGEMBEDDED_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case OGGEMBEDDED_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case OGGEMBEDDED_OUTPUT: GETVALUE(p->Output,pin); break;
	case OGGEMBEDDED_COMMENT: GETVALUE(,p->Comments,pin); break;
	}
	return Result;
}

static int Set( oggembedded* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case OGGEMBEDDED_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case OGGEMBEDDED_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case OGGEMBEDDED_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case OGGEMBEDDED_COMMENT: SETVALUE(p->Comments,pin,ERR_NONE); break;

	case FLOW_EMPTY:
		ogg_sync_reset(p->OggSync);

		if (p->OggStream)
			ogg_stream_reset_serialno(p->OggStream,p->OggStreamNo);

		p->Pending = 0;
		p->Bytes = 0;
		break;
	}
	return Result;
}

static oggembedded OGGEmbedded = 
{
	Enum,
	Get,
	Set,
};

void OGGEmbedded_Init()
{
	NodeRegister((node*)&OGGEmbedded.VMT,PRI_DEFAULT);
}

void OGGEmbedded_Done()
{
	NodeUnRegister((node*)&OGGEmbedded.VMT);
}
