/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: oggpacket.c 3 2004-07-13 11:26:13Z picard $
 *
 * BetaPlayer Vorbis plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "../common/common.h"
#include "vorbis.h"
#include "tremor/ogg.h"
#include "tremor/ivorbiscodec.h"

#define OGGPACKET_INPUT			0x100
#define OGGPACKET_OUTPUT		0x101
#define OGGPACKET_COMMENT		0x102

typedef struct oggpacket
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;
	pin Output;
	pin Comments;

	vorbis_info     Info;
	vorbis_comment  Comment;

	ogg_reference	OggRef;
	ogg_buffer		OggBuffer;
	ogg_packet      OggPacket;

	packet			Packet;

} oggpacket;

const datadef OGGPacketParams[] = 
{
	{ OGGPACKET_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ OGGPACKET_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },
	{ OGGPACKET_COMMENT,	TYPE_PIN, DF_OUTPUT, PACKET_COMMENT },

	DATADEF_END
};

static bool_t SynthesisHeader(oggpacket* p)
{
	++p->OggPacket.packetno;
	p->OggPacket.b_o_s = p->OggPacket.packetno == 0;
	p->OggPacket.bytes = p->OggPacket.packet->length;

	return vorbis_synthesis_headerin(&p->Info,&p->Comment,&p->OggPacket) >= 0;
}

static int SetFormat( oggpacket* p, const audio* Format )
{
	if (p->InputFormat.Format)
	{
		memset(&p->InputFormat,0,sizeof(p->InputFormat));

		vorbis_comment_clear(&p->Comment);
		vorbis_info_clear(&p->Info);
	}

	if (Format)
	{
		int	i,j,n;

		if (Format->Format != AUDIOFMT_VORBIS_PACKET ||
			Format->ExtraLength < 1 || Format->Extra==NULL)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		memset(&p->OutputFormat,0,sizeof(p->OutputFormat));
		p->OutputFormat.Format = AUDIOFMT_VORBIS_INTERNAL;

		p->Packet.Data[0] = &p->OggPacket;
		p->Packet.Length = sizeof(p->OggPacket);

		vorbis_info_init(&p->Info);
		vorbis_comment_init(&p->Comment);

		memset(&p->OggPacket,0,sizeof(ogg_packet));

		p->OggBuffer.data = (uint8_t*) Format->Extra;
		p->OggBuffer.size = Format->ExtraLength;
		p->OggBuffer.refcount = 1;

		p->OggRef.buffer = &p->OggBuffer;
		p->OggRef.next = NULL;

		p->OggPacket.packet = &p->OggRef;
		p->OggPacket.packetno = -1; 

		n = p->OggBuffer.data[0];
		i = 1+n;
		j = 1;

		while (p->OggPacket.packetno < 3 && n>=j)
		{
			p->OggRef.begin = i;
			p->OggRef.length = 0;
			do
			{
				p->OggRef.length += p->OggBuffer.data[j];
			}
			while (p->OggBuffer.data[j++] == 255 && n>=j);
			i += p->OggRef.length;

			if (i > p->OggBuffer.size)
				return ERR_INVALID_DATA;

			if (!SynthesisHeader(p) && p->OggPacket.packetno==0)
				return ERR_INVALID_DATA;
		}
		
		if (p->OggPacket.packetno < 3)
		{
			p->OggRef.begin = i;
			p->OggRef.length = p->OggBuffer.size - i; 

			SynthesisHeader(p);
		}

		if (p->Comments.Node)
			for (i=0;i<p->Comment.comments;++i)
			{
				tchar_t Comment[256];
				StrToTcs(Comment,p->Comment.user_comments[i],256);
				p->Comments.Node->Set(p->Comments.Node,p->Comments.No,Comment,sizeof(Comment));
			}

		p->InputFormat.ByteRate = p->OutputFormat.ByteRate = p->Info.bitrate_nominal >> 3;
		p->InputFormat.SampleRate = p->OutputFormat.SampleRate = p->Info.rate;
		p->InputFormat.Channels = p->OutputFormat.Channels = p->Info.channels;
		p->OutputFormat.Extra = &p->Info;
		p->OutputFormat.ExtraLength = -1;

		if (p->Output.Node)
			p->Output.Node->Set(p->Output.Node,p->Output.No|PACKET_FORMAT,&p->OutputFormat,sizeof(audio));

		p->OggPacket.b_o_s = 0;
		p->OggPacket.packetno = 0;
		p->OggRef.begin = 0;
	}

	return ERR_NONE;
}

static int Decode( oggpacket* p, const packet* Packet )
{
	p->OggBuffer.data = Packet->Data[0];
	p->OggPacket.bytes = p->OggRef.length = p->OggBuffer.size = Packet->Length;
	++p->OggPacket.packetno;

	p->Packet.RefTime = Packet->RefTime;
	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;
	p->Packet.Key = Packet->Key;

	return p->Output.Node->Set(p->Output.Node,p->Output.No,&p->Packet,sizeof(packet));
}

static int Enum( oggpacket* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,OGGPacketParams,NULL);
}

static int Get( oggpacket* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(OGGPACKET_ID,int); break;
	case OGGPACKET_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case OGGPACKET_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case OGGPACKET_OUTPUT: GETVALUE(p->Output,pin); break;
	case OGGPACKET_COMMENT: GETVALUE(,p->Comments,pin); break;
	}
	return Result;
}

static int Set( oggpacket* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case OGGPACKET_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case OGGPACKET_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case OGGPACKET_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case OGGPACKET_COMMENT: SETVALUE(p->Comments,pin,ERR_NONE); break;
	}
	return Result;
}

static oggpacket OGGPacket = 
{
	Enum,
	Get,
	Set,
};

void OGGPacket_Init()
{
	NodeRegister((node*)&OGGPacket.VMT,PRI_DEFAULT);
}

void OGGPacket_Done()
{
	NodeUnRegister((node*)&OGGPacket.VMT);
}


